@@ -19,7 +19,6 @@
  */
 
 #include <bts.h>
-#include <poll_controller.h>
 #include <tbf.h>
 #include <tbf_ul.h>
 #include <encoding.h>
@@ -33,6 +32,7 @@
 #include <pdch.h>
 #include <gprs_ms_storage.h>
 #include <sba.h>
+#include <bts_pch_timer.h>
 
 extern "C" {
 	#include <osmocom/core/talloc.h>
@@ -77,11 +77,25 @@ void bts_trx_free_all_tbf(struct gprs_rlcmac_trx *trx)
 }
 
 static struct osmo_tdef T_defs_bts[] = {
-	{ .T=3142, .default_val=20,  .unit=OSMO_TDEF_S,  .desc="timer (s)", .val=0 },
-	{ .T=3169, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of USF and TFI(s) after the MS uplink TBF assignment is invalid (s)", .val=0 },
-	{ .T=3191, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of TFI(s) after sending (1) last RLC Data Block on TBF(s), or (2) PACKET TBF RELEASE for an MBMS radio bearer (s)", .val=0 },
-	{ .T=3193, .default_val=100, .unit=OSMO_TDEF_MS, .desc="Reuse of TFI(s) after reception of final PACKET DOWNLINK ACK/NACK from MS for TBF (ms)", .val=0 },
-	{ .T=3195, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of TFI(s) upon no response from the MS (radio failure or cell change) for TBF/MBMS radio bearer (s)", .val=0 },
+	{ .T=3142, .default_val=20,  .unit=OSMO_TDEF_S,  .desc="Wait Indication used in Imm Ass Reject during TBF Establishment (CCCH)", .val=0, .min_val = 0, .max_val = 255 }, /* TS 44.018 10.5.2.43, TS 44.060 7.1.3.2.1 (T3172) */
+	{ .T=3169, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of USF and TFI(s) after the MS uplink TBF assignment is invalid", .val=0 },
+	{ .T=3191, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of TFI(s) after sending (1) last RLC Data Block on TBF(s), or (2) PACKET TBF RELEASE for an MBMS radio bearer", .val=0 },
+	{ .T=3193, .default_val=100, .unit=OSMO_TDEF_MS, .desc="Reuse of TFI(s) after reception of final PACKET DOWNLINK ACK/NACK from MS for TBF", .val=0 },
+	{ .T=3195, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of TFI(s) upon no response from the MS (radio failure or cell change) for TBF/MBMS radio bearer", .val=0 },
+	{ .T = -16, .default_val = 1000, .unit = OSMO_TDEF_MS,
+		.desc = "Granularity for *:all_allocated rate counters: amount of milliseconds that one counter increment"
+			" represents. See also X17, X18" },
+	{ .T = -17, .default_val = 0, .unit = OSMO_TDEF_MS,
+		.desc = "Rounding threshold for *:all_allocated rate counters: round up to the next counter increment"
+			" after this many milliseconds. If set to half of X16 (or 0), employ the usual round() behavior:"
+			" round up after half of a granularity period. If set to 1, behave like ceil(): already"
+			" increment the counter immediately when all channels are allocated. If set >= X16, behave like"
+			" floor(): only increment after a full X16 period of all channels being occupied."
+			" See also X16, X18" },
+	{ .T = -18, .default_val = 60000, .unit = OSMO_TDEF_MS,
+		.desc = "Forget-sum period for *:all_allocated rate counters:"
+			" after this amount of idle time, forget internally cumulated time remainders. Zero to always"
+			" keep remainders. See also X16, X17." },
 	{ .T=0, .default_val=0, .unit=OSMO_TDEF_S, .desc=NULL, .val=0 } /* empty item at the end */
 };
 
@@ -91,6 +105,7 @@ static struct osmo_tdef T_defs_bts[] = {
  * the code below.
  */
 static const struct rate_ctr_desc bts_ctr_description[] = {
+	{ "pdch:all_allocated",		"Cumulative counter of seconds where all enabled PDCH resources were allocated"},
 	{ "tbf:dl:alloc",		"TBF DL Allocated     "},
 	{ "tbf:dl:freed",		"TBF DL Freed         "},
 	{ "tbf:dl:aborted",		"TBF DL Aborted       "},
@@ -100,6 +115,11 @@ static const struct rate_ctr_desc bts_ctr_description[] = {
 	{ "tbf:reused",			"TBF Reused           "},
 	{ "tbf:alloc:algo-a",		"TBF Alloc Algo A     "},
 	{ "tbf:alloc:algo-b",		"TBF Alloc Algo B     "},
+	{ "tbf:alloc:failed",		"TBF Alloc Failure (any reason)"},
+	{ "tbf:alloc:failed:no_tfi",	"TBF Alloc Failure (TFIs exhausted)"},
+	{ "tbf:alloc:failed:no_usf",	"TBF Alloc Failure (USFs exhausted)"},
+	{ "tbf:alloc:failed:no_slot_combi", "TBF Alloc Failure (No valid UL/DL slot combination found)"},
+	{ "tbf:alloc:failed:no_slot_avail", "TBF Alloc Failure (No slot available)"},
 	{ "rlc:sent",			"RLC Sent             "},
 	{ "rlc:resent",			"RLC Resent           "},
 	{ "rlc:restarted",		"RLC Restarted        "},
@@ -127,13 +147,22 @@ static const struct rate_ctr_desc bts_ctr_description[] = {
 	{ "llc:scheduled",		"Scheduled Frames     "},
 	{ "llc:dl_bytes",               "RLC encapsulated PDUs"},
 	{ "llc:ul_bytes",               "full PDUs received   "},
-	{ "rach:requests",		"RACH requests        "},
-	{ "11bit_rach:requests",	"11BIT_RACH requests  "},
+	{ "pch:requests",		"PCH requests sent    "},
+	{ "pch:requests:already",	"PCH requests on subscriber already being paged"},
+	{ "pch:requests:timeout",	"PCH requests timeout "},
+	{ "rach:requests",		"RACH requests received"},
+	{ "rach:requests:11bit",	"11BIT_RACH requests received"},
+	{ "rach:requests:one_phase",	"One phase packet access with request for single TS UL"}, /* TS 52.402 B.2.1.49 */
+	{ "rach:requests:two_phase",	"Single block packet request for two phase packet access"}, /* TS 52.402 B.2.1.49 */
+	{ "rach:requests:unexpected",	"RACH Request with unexpected content received"},
 	{ "spb:uplink_first_segment",   "First seg of UL SPB  "},
 	{ "spb:uplink_second_segment",  "Second seg of UL SPB "},
 	{ "spb:downlink_first_segment", "First seg of DL SPB  "},
 	{ "spb:downlink_second_segment","Second seg of DL SPB "},
 	{ "immediate:assignment_UL",	"Immediate Assign UL  "},
+	{ "immediate:assignment_ul:one_phase",	"Immediate Assign UL (one phase packet access)"}, /* TS 52.402 B.2.1.50 */
+	{ "immediate:assignment_ul:two_phase",	"Immediate Assign UL (two phase packet access)"}, /* TS 52.402 B.2.1.50 */
+	{ "immediate:assignment_ul:contention_resolution_success", "First RLC Block (PDU) on the PDTCH from the MS received"}, /* TS 52.402 B.2.1.51 */
 	{ "immediate:assignment_rej",   "Immediate Assign Rej "},
 	{ "immediate:assignment_DL",	"Immediate Assign DL  "},
 	{ "channel:request_description","Channel Request Desc "},
@@ -191,6 +220,14 @@ static const struct rate_ctr_group_desc bts_ctrg_desc = {
 static const struct osmo_stat_item_desc bts_stat_item_description[] = {
 	{ "ms.present",		"MS Present           ",
 		OSMO_STAT_ITEM_NO_UNIT, 4, 0},
+	{ "pdch.available",	"PDCH available       ",
+		OSMO_STAT_ITEM_NO_UNIT, 50, 0},
+	{ "pdch.occupied",	"PDCH occupied (all)  ",
+		OSMO_STAT_ITEM_NO_UNIT, 50, 0},
+	{ "pdch.occupied.gprs",	"PDCH occupied (GPRS) ",
+		OSMO_STAT_ITEM_NO_UNIT, 50, 0},
+	{ "pdch.occupied.egprs","PDCH occupied (EGPRS)",
+		OSMO_STAT_ITEM_NO_UNIT, 50, 0},
 };
 
 static const struct osmo_stat_item_group_desc bts_statg_desc = {
@@ -207,8 +244,8 @@ static int bts_talloc_destructor(struct gprs_rlcmac_bts* bts)
 	 * m_ms_store's destructor */
 	bts->ms_store->cleanup();
 	delete bts->ms_store;
-	delete bts->sba;
-	delete bts->pollController;
+
+	osmo_time_cc_cleanup(&bts->all_allocated_pdch);
 
 	if (bts->ratectrs) {
 		rate_ctr_group_free(bts->ratectrs);
@@ -225,6 +262,8 @@ static int bts_talloc_destructor(struct gprs_rlcmac_bts* bts)
 		bts->app_info = NULL;
 	}
 
+	bts_pch_timer_stop_all(bts);
+
 	llist_del(&bts->list);
 	return 0;
 }
@@ -240,11 +279,9 @@ struct gprs_rlcmac_bts* bts_alloc(struct gprs_pcu *pcu, uint8_t bts_nr)
 	bts->pcu = pcu;
 	bts->nr = bts_nr;
 
-	bts->pollController = new PollController(*bts);
-	bts->sba = new SBAController(*bts);
 	bts->ms_store = new GprsMsStorage(bts);
 
-	bts->cur_fn = 0;
+	bts->cur_fn = FN_UNSET;
 	bts->cur_blk_fn = -1;
 	bts->max_cs_dl = MAX_GPRS_CS;
 	bts->max_cs_ul = MAX_GPRS_CS;
@@ -262,22 +299,9 @@ struct gprs_rlcmac_bts* bts_alloc(struct gprs_pcu *pcu, uint8_t bts_nr)
 	bts->T_defs_bts = T_defs_bts;
 	osmo_tdefs_reset(bts->T_defs_bts);
 
-	INIT_LLIST_HEAD(&bts->ul_tbfs);
-	INIT_LLIST_HEAD(&bts->dl_tbfs);
-
 	/* initialize back pointers */
-	for (size_t trx_no = 0; trx_no < ARRAY_SIZE(bts->trx); ++trx_no) {
-		struct gprs_rlcmac_trx *trx = &bts->trx[trx_no];
-		trx->trx_no = trx_no;
-		trx->bts = bts;
-
-		for (size_t ts_no = 0; ts_no < ARRAY_SIZE(trx->pdch); ++ts_no) {
-			struct gprs_rlcmac_pdch *pdch = &trx->pdch[ts_no];
-			pdch->init_ptcch_msg();
-			pdch->ts_no = ts_no;
-			pdch->trx = trx;
-		}
-	}
+	for (size_t trx_no = 0; trx_no < ARRAY_SIZE(bts->trx); ++trx_no)
+		bts_trx_init(&bts->trx[trx_no], bts, trx_no);
 
 	/* The static allocator might have already registered the counter group.
 	   If this happens and we still called explicitly (in tests/ for example)
@@ -292,26 +316,34 @@ struct gprs_rlcmac_bts* bts_alloc(struct gprs_pcu *pcu, uint8_t bts_nr)
 	bts->statg = osmo_stat_item_group_alloc(tall_pcu_ctx, &bts_statg_desc, 0);
 	OSMO_ASSERT(bts->statg);
 
+	osmo_time_cc_init(&bts->all_allocated_pdch);
+	struct osmo_time_cc_cfg *cc_cfg = &bts->all_allocated_pdch.cfg;
+	cc_cfg->gran_usec = 1*1000000,
+	cc_cfg->forget_sum_usec = 60*1000000,
+	cc_cfg->rate_ctr = rate_ctr_group_get_ctr(bts->ratectrs, CTR_PDCH_ALL_ALLOCATED),
+	cc_cfg->T_gran = -16,
+	cc_cfg->T_round_threshold = -17,
+	cc_cfg->T_forget_sum = -18,
+	cc_cfg->T_defs = T_defs_bts,
+
 	llist_add_tail(&bts->list, &pcu->bts_list);
 
+	INIT_LLIST_HEAD(&bts->pch_timer);
+
 	return bts;
 }
 
-void bts_set_current_frame_number(struct gprs_rlcmac_bts *bts, int fn)
+void bts_set_current_frame_number(struct gprs_rlcmac_bts *bts, uint32_t fn)
 {
 	/* The UL frame numbers lag 3 behind the DL frames and the data
 	 * indication is only sent after all 4 frames of the block have been
 	 * received. Sometimes there is an idle frame between the end of one
-	 * and start of another frame (every 3 blocks).  So the timeout should
-	 * definitely be there if we're more than 8 frames past poll_fn. Let's
-	 * stay on the safe side and say 13 or more. An additional delay can
-	 * happen due to the block processing time in the DSP, so the delay of
-	 * decoded blocks relative to the timing clock can be much larger.
-	 * Values up to 50 frames have been observed under load. */
-	const static int max_delay = 60;
-
+	 * and start of another frame (every 3 blocks). */
+	if (fn != bts->cur_fn && bts->cur_fn != FN_UNSET && fn != fn_next_block(bts->cur_fn)) {
+		LOGP(DRLCMAC, LOGL_NOTICE,
+			"Detected FN jump! %u -> %u\n", bts->cur_fn, fn);
+	}
 	bts->cur_fn = fn;
-	bts->pollController->expireTimedout(bts->cur_fn, max_delay);
 }
 
 static inline int delta_fn(int fn, int to)
@@ -319,7 +351,7 @@ static inline int delta_fn(int fn, int to)
 	return (fn + GSM_MAX_FN * 3 / 2 - to) % GSM_MAX_FN - GSM_MAX_FN/2;
 }
 
-void bts_set_current_block_frame_number(struct gprs_rlcmac_bts *bts, int fn, unsigned max_delay)
+void bts_set_current_block_frame_number(struct gprs_rlcmac_bts *bts, int fn)
 {
 	int delay = 0;
 	const int late_block_delay_thresh = 13;
@@ -346,40 +378,88 @@ void bts_set_current_block_frame_number(struct gprs_rlcmac_bts *bts, int fn, uns
 
 	bts->cur_blk_fn = fn;
 	if (delay < fn_update_ok_min_delay || delay > fn_update_ok_max_delay ||
-		bts_current_frame_number(bts) == 0)
-		bts->cur_fn = fn;
-
-	bts->pollController->expireTimedout(fn, max_delay);
+	    bts_current_frame_number(bts) == FN_UNSET)
+		bts_set_current_frame_number(bts, fn);
 }
 
-int bts_add_paging(struct gprs_rlcmac_bts *bts, uint8_t chan_needed, const struct osmo_mobile_identity *mi)
+/* Helper used by bts_add_paging() whenever the target MS is known */
+static void bts_add_paging_known_ms(struct GprsMs *ms, const struct osmo_mobile_identity *mi, uint8_t chan_needed)
+{
+	uint8_t ts;
+
+	if (ms->ul_tbf) {
+		for (ts = 0; ts < ARRAY_SIZE(ms->ul_tbf->pdch); ts++) {
+			if (ms->ul_tbf->pdch[ts]) {
+				LOGPDCH(ms->ul_tbf->pdch[ts], DRLCMAC, LOGL_INFO,
+					"Paging on PACCH for %s\n", tbf_name(ms->ul_tbf));
+				if (!ms->ul_tbf->pdch[ts]->add_paging(chan_needed, mi))
+					continue;
+				return;
+			}
+		}
+	}
+	if (ms->dl_tbf) {
+		for (ts = 0; ts < ARRAY_SIZE(ms->dl_tbf->pdch); ts++) {
+			if (ms->dl_tbf->pdch[ts]) {
+				LOGPDCH(ms->dl_tbf->pdch[ts], DRLCMAC, LOGL_INFO,
+					"Paging on PACCH for %s\n", tbf_name(ms->ul_tbf));
+				if (!ms->dl_tbf->pdch[ts]->add_paging(chan_needed, mi))
+					continue;
+				return;
+			}
+		}
+	}
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Unable to page on PACCH, no available TBFs\n");
+	return;
+}
+
+/* ms is NULL if no specific taget was found */
+int bts_add_paging(struct gprs_rlcmac_bts *bts, const struct paging_req_cs *req, struct GprsMs *ms)
 {
 	uint8_t l, trx, ts, any_tbf = 0;
 	struct gprs_rlcmac_tbf *tbf;
-	struct llist_item *pos;
+	struct llist_head *tmp;
+	const struct osmo_mobile_identity *mi;
 	uint8_t slot_mask[8];
 	int8_t first_ts; /* must be signed */
 
-	struct llist_head *tbfs_lists[] = {
-		&bts->ul_tbfs,
-		&bts->dl_tbfs,
-		NULL
-	};
+	/* First, build the MI used to page on PDCH from available subscriber info: */
+	if (req->mi_tmsi_present) {
+		mi = &req->mi_tmsi;
+	} else if (req->mi_imsi_present) {
+		mi = &req->mi_imsi;
+	} else {
+		LOGPMS(ms, DRLCMAC, LOGL_ERROR, "Unable to page on PACCH, no TMSI nor IMSI in request\n");
+		return -EINVAL;
+	}
 
 	if (log_check_level(DRLCMAC, LOGL_INFO)) {
 		char str[64];
 		osmo_mobile_identity_to_str_buf(str, sizeof(str), mi);
-		LOGP(DRLCMAC, LOGL_INFO, "Add RR paging: chan-needed=%d MI=%s\n", chan_needed, str);
+		LOGP(DRLCMAC, LOGL_INFO, "Add RR paging: chan-needed=%d MI=%s\n", req->chan_needed, str);
 	}
 
-	/* collect slots to page
-	 * Mark slots for every TBF, but only mark one of it.
+	/* We known the target MS for the paging req, send the req only on PDCH
+	 * were that target MS is listening (first slot is enough), and we are done. */
+	if (ms) {
+		bts_add_paging_known_ms(ms, mi, req->chan_needed);
+		return 0;
+	}
+
+	/* We don't know the target MS.
+	 * collect slots to page
+	 * Mark up to one slot attached to each of the TBF of the MS.
 	 * Mark only the first slot found.
 	 * Don't mark, if TBF uses a different slot that is already marked. */
 	memset(slot_mask, 0, sizeof(slot_mask));
-	for (l = 0; tbfs_lists[l]; l++) {
-		llist_for_each_entry(pos, tbfs_lists[l], list) {
-			tbf = (struct gprs_rlcmac_tbf *)pos->entry;
+
+	llist_for_each(tmp, bts_ms_store(bts)->ms_list()) {
+		ms = llist_entry(tmp, typeof(*ms), list);
+		struct gprs_rlcmac_tbf *tbfs[] = { ms->ul_tbf, ms->dl_tbf };
+		for (l = 0; l < ARRAY_SIZE(tbfs); l++) {
+			tbf = (struct gprs_rlcmac_tbf *)tbfs[l];
+			if (!tbf)
+				continue;
 			first_ts = -1;
 			for (ts = 0; ts < 8; ts++) {
 				if (tbf->pdch[ts]) {
@@ -414,10 +494,10 @@ int bts_add_paging(struct gprs_rlcmac_bts *bts, uint8_t chan_needed, const struc
 		for (ts = 0; ts < 8; ts++) {
 			if ((slot_mask[trx] & (1 << ts))) {
 				/* schedule */
-				if (!bts->trx[trx].pdch[ts].add_paging(chan_needed, mi))
+				if (!bts->trx[trx].pdch[ts].add_paging(req->chan_needed, mi))
 					return -ENOMEM;
 
-				LOGP(DRLCMAC, LOGL_INFO, "Paging on PACCH of TRX=%d TS=%d\n", trx, ts);
+				LOGPDCH(&bts->trx[trx].pdch[ts], DRLCMAC, LOGL_INFO, "Paging on PACCH\n");
 				any_tbf = 1;
 			}
 		}
@@ -483,45 +563,6 @@ void bts_send_gsmtap_meas(struct gprs_rlcmac_bts *bts,
 		    meas->rssi, meas->link_qual, data, len);
 }
 
-static inline bool tbf_check(gprs_rlcmac_tbf *tbf, uint32_t fn, uint8_t trx_no, uint8_t ts)
-{
-	if (tbf->state_is_not(GPRS_RLCMAC_RELEASING) && tbf->poll_scheduled()
-	    && tbf->poll_fn == fn && tbf->trx->trx_no == trx_no && tbf->poll_ts == ts)
-		return true;
-
-	return false;
-}
-
-struct gprs_rlcmac_dl_tbf *bts_dl_tbf_by_poll_fn(struct gprs_rlcmac_bts *bts, uint32_t fn, uint8_t trx, uint8_t ts)
-{
-	struct llist_item *pos;
-	struct gprs_rlcmac_tbf *tbf;
-
-	/* only one TBF can poll on specific TS/FN, because scheduler can only
-	 * schedule one downlink control block (with polling) at a FN per TS */
-	llist_for_each_entry(pos, &bts->dl_tbfs, list) {
-		tbf = (struct gprs_rlcmac_tbf *)pos->entry;
-		if (tbf_check(tbf, fn, trx, ts))
-			return as_dl_tbf(tbf);
-	}
-	return NULL;
-}
-
-struct gprs_rlcmac_ul_tbf *bts_ul_tbf_by_poll_fn(struct gprs_rlcmac_bts *bts, uint32_t fn, uint8_t trx, uint8_t ts)
-{
-	struct llist_item *pos;
-	struct gprs_rlcmac_tbf *tbf;
-
-	/* only one TBF can poll on specific TS/FN, because scheduler can only
-	 * schedule one downlink control block (with polling) at a FN per TS */
-	llist_for_each_entry(pos, &bts->ul_tbfs, list) {
-		tbf = (struct gprs_rlcmac_tbf *)pos->entry;
-		if (tbf_check(tbf, fn, trx, ts))
-			return as_ul_tbf(tbf);
-	}
-	return NULL;
-}
-
 /* lookup downlink TBF Entity (by TFI) */
 struct gprs_rlcmac_dl_tbf *bts_dl_tbf_by_tfi(struct gprs_rlcmac_bts *bts, uint8_t tfi, uint8_t trx, uint8_t ts)
 {
@@ -608,6 +649,7 @@ int bts_tfi_find_free(const struct gprs_rlcmac_bts *bts, enum gprs_rlcmac_tbf_di
 
 	if (best_trx_nr == 0xff || best_cnt == 0) {
 		LOGP(DRLCMAC, LOGL_NOTICE, "No TFI available (suggested TRX: %d).\n", use_trx);
+		bts_do_rate_ctr_inc(bts, CTR_TBF_ALLOC_FAIL_NO_TFI);
 		return -EBUSY;
 	}
 
@@ -654,9 +696,7 @@ int bts_rcv_imm_ass_cnf(struct gprs_rlcmac_bts *bts, const uint8_t *data, uint32
 	}
 
 	LOGP(DRLCMAC, LOGL_DEBUG, "Got IMM.ASS confirm for TLLI=%08x\n", tlli);
-
-	if (dl_tbf->m_wait_confirm)
-		T_START(dl_tbf, T0, -2002, "assignment (AGCH)", true);
+	osmo_fsm_inst_dispatch(dl_tbf->state_fsm.fi, TBF_EV_ASSIGN_PCUIF_CNF, NULL);
 
 	return 0;
 }
@@ -823,20 +863,57 @@ static int parse_rach_ind(const struct rach_ind_params *rip,
 	return 0;
 }
 
+struct gprs_rlcmac_sba *bts_alloc_sba(struct gprs_rlcmac_bts *bts, uint8_t ta)
+{
+	struct gprs_rlcmac_pdch *pdch;
+	struct gprs_rlcmac_sba *sba = NULL;
+	int8_t trx, ts;
+
+	if (!gsm48_ta_is_valid(ta))
+		return NULL;
+
+	for (trx = 0; trx < 8; trx++) {
+		for (ts = 7; ts >= 0; ts--) {
+			pdch = &bts->trx[trx].pdch[ts];
+			if (!pdch->is_enabled())
+				continue;
+			break;
+		}
+		if (ts >= 0)
+			break;
+	}
+	if (trx == 8) {
+		LOGP(DRLCMAC, LOGL_NOTICE, "No PDCH available.\n");
+		return NULL;
+	}
+
+	sba = sba_alloc(bts, pdch, ta);
+	if (!sba)
+		return NULL;
+
+	bts_do_rate_ctr_inc(bts, CTR_SBA_ALLOCATED);
+	return sba;
+}
+
 int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 {
 	struct chan_req_params chan_req = { 0 };
 	struct gprs_rlcmac_ul_tbf *tbf = NULL;
-	uint8_t trx_no, ts_no;
+	struct gprs_rlcmac_sba *sba;
+	struct gprs_rlcmac_pdch *pdch;
+	struct gprs_rlcmac_trx *trx;
 	uint32_t sb_fn = 0;
 	uint8_t usf = 7;
-	uint8_t tsc = 0;
 	int plen, rc;
 
+	/* Allocate a bit-vector for RR Immediate Assignment [Reject] */
+	struct bitvec *bv = bitvec_alloc(22, tall_pcu_ctx); /* without plen */
+	bitvec_unhex(bv, DUMMY_VEC); /* standard '2B'O padding */
+
 	bts_do_rate_ctr_inc(bts, CTR_RACH_REQUESTS);
 
 	if (rip->is_11bit)
-		bts_do_rate_ctr_inc(bts, CTR_11BIT_RACH_REQUESTS);
+		bts_do_rate_ctr_inc(bts, CTR_RACH_REQUESTS_11BIT);
 
 	/* Determine full frame number */
 	uint32_t Fn = bts_rfn_to_fn(bts, rip->rfn);
@@ -850,15 +927,24 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 
 	/* Parse [EGPRS Packet] Channel Request from RACH.ind */
 	rc = parse_rach_ind(rip, &chan_req);
-	if (rc) /* Send RR Immediate Assignment Reject */
+	if (rc) {
+		bts_do_rate_ctr_inc(bts, CTR_RACH_REQUESTS_UNEXPECTED);
+		/* Send RR Immediate Assignment Reject */
 		goto send_imm_ass_rej;
+	}
 
-	if (chan_req.single_block)
-		LOGP(DRLCMAC, LOGL_DEBUG, "MS requests single block allocation\n");
-	else if (bts->pcu->vty.force_two_phase) {
-		LOGP(DRLCMAC, LOGL_DEBUG, "MS requests single block allocation, "
-		     "but we force two phase access\n");
-		chan_req.single_block = true;
+	if (chan_req.single_block) {
+		bts_do_rate_ctr_inc(bts, CTR_RACH_REQUESTS_TWO_PHASE);
+		LOGP(DRLCMAC, LOGL_DEBUG, "MS requests single block allocation "
+		     "(two phase packet access)\n");
+	} else {
+		bts_do_rate_ctr_inc(bts, CTR_RACH_REQUESTS_ONE_PHASE);
+		LOGP(DRLCMAC, LOGL_DEBUG, "MS requests single TS uplink transmission "
+		     "(one phase packet access)\n");
+		if (bts->pcu->vty.force_two_phase) {
+			LOGP(DRLCMAC, LOGL_DEBUG, "Forcing two phase access\n");
+			chan_req.single_block = true;
+		}
 	}
 
 	/* TODO: handle Radio Priority (see 3GPP TS 44.060, table 11.2.5a.5) */
@@ -868,68 +954,63 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 
 	/* Should we allocate a single block or an Uplink TBF? */
 	if (chan_req.single_block) {
-		rc = bts_sba(bts)->alloc(&trx_no, &ts_no, &sb_fn, ta);
-		if (rc < 0) {
+		sba = bts_alloc_sba(bts, ta);
+		if (!sba) {
 			LOGP(DRLCMAC, LOGL_NOTICE, "No PDCH resource for "
-			     "single block allocation: rc=%d\n", rc);
+			     "single block allocation\n");
+			rc = -EBUSY;
 			/* Send RR Immediate Assignment Reject */
 			goto send_imm_ass_rej;
 		}
 
-		tsc = bts->trx[trx_no].pdch[ts_no].tsc;
+		pdch = sba->pdch;
+		sb_fn = sba->fn;
 		LOGP(DRLCMAC, LOGL_DEBUG, "Allocated a single block at "
-		     "SBFn=%u TRX=%u TS=%u\n", sb_fn, trx_no, ts_no);
+		     "SBFn=%u TRX=%u TS=%u\n", sb_fn, pdch->trx->trx_no, pdch->ts_no);
+		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF_TWO_PHASE);
 	} else {
 		GprsMs *ms = bts_alloc_ms(bts, 0, chan_req.egprs_mslot_class);
-		tbf = tbf_alloc_ul_tbf(bts, ms, -1, true);
+		tbf = tbf_alloc_ul_ccch(bts, ms);
 		if (!tbf) {
-			LOGP(DRLCMAC, LOGL_NOTICE, "No PDCH resource for Uplink TBF\n");
 			/* Send RR Immediate Assignment Reject */
 			rc = -EBUSY;
 			goto send_imm_ass_rej;
 		}
-
-		/* FIXME: Copy and paste with other routines.. */
 		tbf->set_ta(ta);
-		TBF_SET_STATE(tbf, GPRS_RLCMAC_FLOW);
-		TBF_ASS_TYPE_SET(tbf, GPRS_RLCMAC_FLAG_CCCH);
-		T_START(tbf, T3169, 3169, "RACH (new UL-TBF)", true);
-		trx_no = tbf->trx->trx_no;
-		ts_no = tbf->first_ts;
-		usf = tbf->m_usf[ts_no];
-		tsc = tbf->tsc();
+		pdch = &tbf->trx->pdch[tbf->first_ts];
+		usf = tbf->m_usf[pdch->ts_no];
+		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF_ONE_PHASE);
 	}
+	trx = pdch->trx;
+
+	LOGP(DRLCMAC, LOGL_DEBUG, "Tx Immediate Assignment on AGCH: "
+	     "TRX=%u (ARFCN %u) TS=%u TA=%u TSC=%u TFI=%d USF=%d\n",
+	     trx->trx_no, trx->arfcn & ~ARFCN_FLAG_MASK,
+	     pdch->ts_no, ta, pdch->tsc, tbf ? tbf->tfi() : -1, usf);
+	plen = Encoding::write_immediate_assignment(pdch, tbf, bv,
+		false, rip->ra, Fn, ta, usf, false, sb_fn,
+		bts_get_ms_pwr_alpha(bts), bts->pcu->vty.gamma, -1,
+		rip->burst_type);
+	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF);
+	if (plen >= 0) {
+		pcu_l1if_tx_agch(bts, bv, plen);
+		rc = 0;
+	} else {
+		rc = plen;
+	}
+	bitvec_free(bv);
+	return rc;
 
 send_imm_ass_rej:
-	/* Allocate a bit-vector for RR Immediate Assignment [Reject] */
-	struct bitvec *bv = bitvec_alloc(22, tall_pcu_ctx); /* without plen */
-	bitvec_unhex(bv, DUMMY_VEC); /* standard '2B'O padding */
-
-	if (rc != 0) {
-		LOGP(DRLCMAC, LOGL_DEBUG, "Tx Immediate Assignment Reject on AGCH\n");
-		plen = Encoding::write_immediate_assignment_reject(
-			bv, rip->ra, Fn, rip->burst_type);
-		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_REJ);
-	} else {
-		LOGP(DRLCMAC, LOGL_DEBUG, "Tx Immediate Assignment on AGCH: "
-		     "TRX=%u (ARFCN %u) TS=%u TA=%u TSC=%u TFI=%d USF=%d\n",
-		     trx_no, bts->trx[trx_no].arfcn & ~ARFCN_FLAG_MASK,
-		     ts_no, ta, tsc, tbf ? tbf->tfi() : -1, usf);
-		plen = Encoding::write_immediate_assignment(
-			&bts->trx[trx_no].pdch[ts_no], tbf, bv,
-			false, rip->ra, Fn, ta, usf, false, sb_fn,
-			bts_get_ms_pwr_alpha(bts), bts->pcu->vty.gamma, -1,
-			rip->burst_type);
-		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF);
-	}
-
+	LOGP(DRLCMAC, LOGL_DEBUG, "Tx Immediate Assignment Reject on AGCH\n");
+	plen = Encoding::write_immediate_assignment_reject(
+		bv, rip->ra, Fn, rip->burst_type,
+		(uint8_t)osmo_tdef_get(bts->T_defs_bts, 3142, OSMO_TDEF_S, -1));
+	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_REJ);
 	if (plen >= 0)
 		pcu_l1if_tx_agch(bts, bv, plen);
-	else
-		rc = plen;
-
 	bitvec_free(bv);
-
+	/* rc was already properly set before goto */
 	return rc;
 }
 
@@ -958,7 +1039,7 @@ int bts_rcv_ptcch_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params
 
 	/* Make sure PDCH time-slot is enabled */
 	pdch = &bts->trx[rip->trx_nr].pdch[rip->ts_nr];
-	if (!pdch->m_is_enabled) {
+	if (!pdch->is_enabled()) {
 		LOGP(DRLCMAC, LOGL_NOTICE, "(TRX=%u TS=%u RFN=%u) Rx RACH.ind (PTCCH/U) "
 		     "for inactive PDCH\n", rip->trx_nr, rip->ts_nr, rip->rfn);
 		return -EAGAIN;
@@ -982,10 +1063,11 @@ int bts_rcv_ptcch_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params
 	return 0;
 }
 
-void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf, bool poll, uint16_t pgroup)
+void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf)
 {
 	uint8_t trx_no = tbf->trx->trx_no;
 	uint8_t ts_no = tbf->first_ts;
+	uint16_t pgroup = ms_paging_group(tbf_ms(tbf));
 	int plen;
 
 	LOGPTBF(tbf, LOGL_INFO, "TX: START Immediate Assignment Downlink (PCH)\n");
@@ -993,12 +1075,12 @@ void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf, bo
 	bitvec_unhex(immediate_assignment, DUMMY_VEC); /* standard '2B'O padding */
 	/* use request reference that has maximum distance to current time,
 	 * so the assignment will not conflict with possible RACH requests. */
-	LOGP(DRLCMAC, LOGL_DEBUG, " - TRX=%d (%d) TS=%d TA=%d pollFN=%d\n",
-		trx_no, tbf->trx->arfcn, ts_no, tbf->ta(), poll ? tbf->poll_fn : -1);
+	LOGP(DRLCMAC, LOGL_DEBUG, " - TRX=%d (%d) TS=%d TA=%d\n",
+		trx_no, tbf->trx->arfcn, ts_no, tbf->ta());
 	plen = Encoding::write_immediate_assignment(&bts->trx[trx_no].pdch[ts_no],
 						    tbf, immediate_assignment, true, 125,
 						    (tbf->pdch[ts_no]->last_rts_fn + 21216) % GSM_MAX_FN,
-						    tbf->ta(), 7, poll, tbf->poll_fn,
+						    tbf->ta(), 7, false, 0,
 						    bts_get_ms_pwr_alpha(bts), bts->pcu->vty.gamma, -1,
 						    GSM_L1_BURST_TYPE_ACCESS_0);
 	if (plen >= 0) {
@@ -1080,12 +1162,7 @@ GprsMs *bts_alloc_ms(struct gprs_rlcmac_bts* bts, uint8_t ms_class, uint8_t egpr
 	return ms;
 }
 
-struct SBAController *bts_sba(struct gprs_rlcmac_bts *bts)
-{
-	return bts->sba;
-}
-
-struct GprsMsStorage *bts_ms_store(struct gprs_rlcmac_bts *bts)
+struct GprsMsStorage *bts_ms_store(const struct gprs_rlcmac_bts *bts)
 {
 	return bts->ms_store;
 }
@@ -1108,10 +1185,8 @@ void update_tbf_ta(struct gprs_rlcmac_ul_tbf *tbf, int8_t ta_delta)
 		/* limit target TA in range 0..63 bits */
 		ta_target = ta_limit(ta_adj);
 
-		LOGP(DL1IF, LOGL_INFO, "PH-DATA-IND is updating TLLI=0x%08x: TA %u -> %u on "
-				"TRX = %d, TS = %d, FN = %d\n",
-				tbf->tlli(), tbf->ta(), ta_target,
-				tbf->trx->trx_no , tbf->poll_ts, tbf->poll_fn);
+		LOGP(DL1IF, LOGL_INFO, "PH-DATA-IND is updating %s: TA %u -> %u on "
+		     "TRX = %d\n", tbf_name(tbf), tbf->ta(), ta_target, tbf->trx->trx_no);
 		tbf->set_ta(ta_target);
 	}
 }
@@ -1125,10 +1200,8 @@ void set_tbf_ta(struct gprs_rlcmac_ul_tbf *tbf, uint8_t ta)
 		/* limit target TA in range 0..63 bits */
 		ta_target = ta_limit(ta);
 
-		LOGP(DL1IF, LOGL_INFO, "PH-RA-IND is updating TLLI=0x%08x: TA %u -> %u on "
-				"TRX = %d, TS = %d, FN = %d\n",
-				tbf->tlli(), tbf->ta(), ta_target,
-				tbf->trx->trx_no , tbf->poll_ts, tbf->poll_fn);
+		LOGP(DL1IF, LOGL_INFO, "PH-RA-IND is updating %s: TA %u -> %u on "
+		     "TRX = %d\n", tbf_name(tbf), tbf->ta(), ta_target, tbf->trx->trx_no);
 		tbf->set_ta(ta_target);
 	}
 }
@@ -1136,23 +1209,44 @@ void set_tbf_ta(struct gprs_rlcmac_ul_tbf *tbf, uint8_t ta)
 void bts_update_tbf_ta(struct gprs_rlcmac_bts *bts, const char *p, uint32_t fn,
 		       uint8_t trx_no, uint8_t ts, int8_t ta, bool is_rach)
 {
-	struct gprs_rlcmac_ul_tbf *tbf =
-		bts_ul_tbf_by_poll_fn(bts, fn, trx_no, ts);
-	if (!tbf)
-		LOGP(DL1IF, LOGL_DEBUG, "[%s] update TA = %u ignored due to "
-		     "unknown UL TBF on TRX = %d, TS = %d, FN = %d\n",
-		     p, ta, trx_no, ts, fn);
-	else {
-		/* we need to distinguish TA information provided by L1
-		 * from PH-DATA-IND and PHY-RA-IND so that we can properly
-		 * update TA for given TBF
-		 */
-		if (is_rach)
-			set_tbf_ta(tbf, (uint8_t)ta);
-		else
-			update_tbf_ta(tbf, ta);
+	struct gprs_rlcmac_pdch *pdch = &bts->trx[trx_no].pdch[ts];
+	struct pdch_ulc_node *poll;
+	struct gprs_rlcmac_ul_tbf *tbf;
+	if (!pdch->is_enabled())
+		goto no_tbf;
 
-	}
+	poll = pdch_ulc_get_node(pdch->ulc, fn);
+	if (!poll || poll->type != PDCH_ULC_NODE_TBF_POLL ||
+	    poll->tbf_poll.poll_tbf->direction != GPRS_RLCMAC_UL_TBF)
+		goto no_tbf;
+
+	tbf = as_ul_tbf(poll->tbf_poll.poll_tbf);
+	/* we need to distinguish TA information provided by L1
+	 * from PH-DATA-IND and PHY-RA-IND so that we can properly
+	 * update TA for given TBF
+	 */
+	if (is_rach)
+		set_tbf_ta(tbf, (uint8_t)ta);
+	else
+		update_tbf_ta(tbf, ta);
+	return;
+
+no_tbf:
+	LOGP(DL1IF, LOGL_DEBUG, "[%s] update TA = %u ignored due to "
+	     "unknown UL TBF on TRX = %d, TS = %d, FN = %d\n",
+	     p, ta, trx_no, ts, fn);
+}
+
+void bts_trx_init(struct gprs_rlcmac_trx *trx, struct gprs_rlcmac_bts *bts, uint8_t trx_no)
+{
+	trx->trx_no = trx_no;
+	trx->bts = bts;
+
+	INIT_LLIST_HEAD(&trx->ul_tbfs);
+	INIT_LLIST_HEAD(&trx->dl_tbfs);
+
+	for (size_t ts_no = 0; ts_no < ARRAY_SIZE(trx->pdch); ts_no++)
+		pdch_init(&trx->pdch[ts_no], trx, ts_no);
 }
 
 void bts_trx_reserve_slots(struct gprs_rlcmac_trx *trx, enum gprs_rlcmac_tbf_direction dir,
@@ -1298,8 +1392,25 @@ uint8_t bts_get_ms_pwr_alpha(const struct gprs_rlcmac_bts *bts)
 	if (bts->pcu->vty.force_alpha != (uint8_t)-1)
 		return bts->pcu->vty.force_alpha;
 	if (bts->si13_is_set)
-		return bts->si31_ro_decoded.pwr_ctrl_pars.alpha;
+		return bts->si13_ro_decoded.pwr_ctrl_pars.alpha;
 	/* default if no SI13 is received yet: closed loop control, TS 44.060
 	 * B.2 Closed loop control */
 	return 0;
 }
+
+/* Used by counter availablePDCHAllocatedTime, TS 52.402 B.2.1.45 "All available PDCH allocated time" */
+bool bts_all_pdch_allocated(const struct gprs_rlcmac_bts *bts)
+{
+	unsigned trx_no, ts_no;
+	for (trx_no = 0; trx_no < ARRAY_SIZE(bts->trx); trx_no++) {
+		const struct gprs_rlcmac_trx *trx = &bts->trx[trx_no];
+		for (ts_no = 0; ts_no < ARRAY_SIZE(trx->pdch); ts_no++) {
+			const struct gprs_rlcmac_pdch *pdch = &trx->pdch[ts_no];
+			if (!pdch_is_enabled(pdch))
+				continue;
+			if(!pdch_is_full(pdch))
+				return false;
+		}
+	}
+	return true;
+}
