@@ -32,6 +32,7 @@ extern "C" {
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/stat_item.h>
 #include <osmocom/core/tdef.h>
+#include <osmocom/core/time_cc.h>
 #include <osmocom/gprs/gprs_ns2.h>
 #include <osmocom/gsm/l1sap.h>
 #include <osmocom/gsm/protocol/gsm_04_08.h>
@@ -59,12 +60,18 @@ struct gprs_rlcmac_trx {
 	struct gprs_rlcmac_bts *bts;
 	uint8_t trx_no;
 
+	/* list of uplink TBFs */
+	struct llist_head ul_tbfs; /* list of gprs_rlcmac_tbf */
+	/* list of downlink TBFs */
+	struct llist_head dl_tbfs; /* list of gprs_rlcmac_tbf */
+
 };
 
 
 #ifdef __cplusplus
 extern "C" {
 #endif
+void bts_trx_init(struct gprs_rlcmac_trx *trx, struct gprs_rlcmac_bts *bts, uint8_t trx_no);
 void bts_trx_reserve_slots(struct gprs_rlcmac_trx *trx, enum gprs_rlcmac_tbf_direction dir, uint8_t slots);
 void bts_trx_unreserve_slots(struct gprs_rlcmac_trx *trx, enum gprs_rlcmac_tbf_direction dir, uint8_t slots);
 void bts_trx_free_all_tbf(struct gprs_rlcmac_trx *trx);
@@ -78,6 +85,7 @@ void bts_update_tbf_ta(struct gprs_rlcmac_bts *bts, const char *p, uint32_t fn,
 
 
 enum {
+	CTR_PDCH_ALL_ALLOCATED,
 	CTR_TBF_DL_ALLOCATED,
 	CTR_TBF_DL_FREED,
 	CTR_TBF_DL_ABORTED,
@@ -87,6 +95,11 @@ enum {
 	CTR_TBF_REUSED,
 	CTR_TBF_ALLOC_ALGO_A,
 	CTR_TBF_ALLOC_ALGO_B,
+	CTR_TBF_ALLOC_FAIL,
+	CTR_TBF_ALLOC_FAIL_NO_TFI,
+	CTR_TBF_ALLOC_FAIL_NO_USF,
+	CTR_TBF_ALLOC_FAIL_NO_SLOT_COMBI,
+	CTR_TBF_ALLOC_FAIL_NO_SLOT_AVAIL,
 	CTR_RLC_SENT,
 	CTR_RLC_RESENT,
 	CTR_RLC_RESTARTED,
@@ -114,13 +127,22 @@ enum {
 	CTR_LLC_FRAME_SCHED,
 	CTR_LLC_DL_BYTES,
 	CTR_LLC_UL_BYTES,
+	CTR_PCH_REQUESTS,
+	CTR_PCH_REQUESTS_ALREADY,
+	CTR_PCH_REQUESTS_TIMEDOUT,
 	CTR_RACH_REQUESTS,
-	CTR_11BIT_RACH_REQUESTS,
+	CTR_RACH_REQUESTS_11BIT,
+	CTR_RACH_REQUESTS_ONE_PHASE,
+	CTR_RACH_REQUESTS_TWO_PHASE,
+	CTR_RACH_REQUESTS_UNEXPECTED,
 	CTR_SPB_UL_FIRST_SEGMENT,
 	CTR_SPB_UL_SECOND_SEGMENT,
 	CTR_SPB_DL_FIRST_SEGMENT,
 	CTR_SPB_DL_SECOND_SEGMENT,
 	CTR_IMMEDIATE_ASSIGN_UL_TBF,
+	CTR_IMMEDIATE_ASSIGN_UL_TBF_ONE_PHASE,
+	CTR_IMMEDIATE_ASSIGN_UL_TBF_TWO_PHASE,
+	CTR_IMMEDIATE_ASSIGN_UL_TBF_CONTENTION_RESOLUTION_SUCCESS,
 	CTR_IMMEDIATE_ASSIGN_REJ,
 	CTR_IMMEDIATE_ASSIGN_DL_TBF,
 	CTR_CHANNEL_REQUEST_DESCRIPTION,
@@ -169,6 +191,10 @@ enum {
 
 enum {
 	STAT_MS_PRESENT,
+	STAT_PDCH_AVAILABLE,
+	STAT_PDCH_OCCUPIED,
+	STAT_PDCH_OCCUPIED_GPRS,
+	STAT_PDCH_OCCUPIED_EGPRS,
 };
 
 /* RACH.ind parameters (to be parsed) */
@@ -189,8 +215,6 @@ struct chan_req_params {
 	bool single_block;
 };
 
-struct PollController;
-struct SBAController;
 struct GprsMsStorage;
 struct pcu_l1_meas;
 
@@ -222,10 +246,13 @@ struct gprs_rlcmac_bts {
 
 	uint8_t si1[GSM_MACBLOCK_LEN];
 	bool si1_is_set;
+	uint8_t si2[GSM_MACBLOCK_LEN];
+	bool si2_is_set;
+	struct gsm_sysinfo_freq si2_bcch_cell_list[1024];
 	uint8_t si3[GSM_MACBLOCK_LEN];
 	bool si3_is_set;
 	uint8_t si13[GSM_MACBLOCK_LEN];
-	struct osmo_gsm48_si13_info si31_ro_decoded;
+	struct osmo_gsm48_si13_info si13_ro_decoded;
 	bool si13_is_set;
 
 	/* State for dynamic algorithm selection */
@@ -242,43 +269,47 @@ struct gprs_rlcmac_bts {
 	/* back pointer to PCU object */
 	struct gprs_pcu *pcu;
 
-	int cur_fn;
+	uint32_t cur_fn;
 	int cur_blk_fn;
 	uint8_t max_cs_dl, max_cs_ul;
 	uint8_t max_mcs_dl, max_mcs_ul;
-	struct PollController *pollController;
-	struct SBAController *sba;
 	struct rate_ctr_group *ratectrs;
 	struct osmo_stat_item_group *statg;
 
 	struct GprsMsStorage *ms_store;
 
-	/* list of uplink TBFs */
-	struct llist_head ul_tbfs; /* list of gprs_rlcmac_tbf */
-	/* list of downlink TBFs */
-	struct llist_head dl_tbfs; /* list of gprs_rlcmac_tbf */
+	/* List of struct bts_pch_timer for active PCH pagings */
+	struct llist_head pch_timer;
+
+	struct osmo_time_cc all_allocated_pdch;
 };
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
+struct paging_req_cs {
+	uint8_t chan_needed;
+	uint32_t tlli; /* GSM_RESERVED_TMSI if not present */
+	bool mi_tmsi_present;
+	struct osmo_mobile_identity mi_tmsi;
+	bool mi_imsi_present;
+	struct osmo_mobile_identity mi_imsi;
+};
+
 struct GprsMs *bts_alloc_ms(struct gprs_rlcmac_bts *bts, uint8_t ms_class, uint8_t egprs_ms_class);
-int bts_add_paging(struct gprs_rlcmac_bts *bts, uint8_t chan_needed, const struct osmo_mobile_identity *mi);
+int bts_add_paging(struct gprs_rlcmac_bts *bts, const struct paging_req_cs *req, struct GprsMs *ms);
 
 uint32_t bts_rfn_to_fn(const struct gprs_rlcmac_bts *bts, int32_t rfn);
 
-struct gprs_rlcmac_dl_tbf *bts_dl_tbf_by_poll_fn(struct gprs_rlcmac_bts *bts, uint32_t fn, uint8_t trx, uint8_t ts);
-struct gprs_rlcmac_ul_tbf *bts_ul_tbf_by_poll_fn(struct gprs_rlcmac_bts *bts, uint32_t fn, uint8_t trx, uint8_t ts);
 struct gprs_rlcmac_dl_tbf *bts_dl_tbf_by_tfi(struct gprs_rlcmac_bts *bts, uint8_t tfi, uint8_t trx, uint8_t ts);
 struct gprs_rlcmac_ul_tbf *bts_ul_tbf_by_tfi(struct gprs_rlcmac_bts *bts, uint8_t tfi, uint8_t trx, uint8_t ts);
 
-void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf, bool poll, uint16_t pgroup);
+void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf);
 
-/** TODO: change the number to unsigned */
-void bts_set_current_frame_number(struct gprs_rlcmac_bts *bts, int frame_number);
-void bts_set_current_block_frame_number(struct gprs_rlcmac_bts *bts, int frame_number, unsigned max_delay);
-static inline int bts_current_frame_number(const struct gprs_rlcmac_bts *bts)
+void bts_set_current_frame_number(struct gprs_rlcmac_bts *bts, uint32_t frame_number);
+void bts_set_current_block_frame_number(struct gprs_rlcmac_bts *bts, int frame_number);
+static inline uint32_t bts_current_frame_number(const struct gprs_rlcmac_bts *bts)
 {
 	return bts->cur_fn;
 }
@@ -302,9 +333,7 @@ void bts_send_gsmtap_rach(struct gprs_rlcmac_bts *bts,
 			  enum pcu_gsmtap_category categ, uint8_t channel,
 			  const struct rach_ind_params *rip);
 
-struct SBAController *bts_sba(struct gprs_rlcmac_bts *bts);
-
-struct GprsMsStorage *bts_ms_store(struct gprs_rlcmac_bts *bts);
+struct GprsMsStorage *bts_ms_store(const struct gprs_rlcmac_bts *bts);
 
 struct GprsMs *bts_ms_by_tlli(struct gprs_rlcmac_bts *bts, uint32_t tlli, uint32_t old_tlli);
 
@@ -318,21 +347,28 @@ static inline struct osmo_stat_item_group *bts_stat_items(struct gprs_rlcmac_bts
 	return bts->statg;
 }
 
-static inline void bts_do_rate_ctr_inc(struct gprs_rlcmac_bts *bts, unsigned int ctr_id) {
-	rate_ctr_inc(&bts->ratectrs->ctr[ctr_id]);
+static inline void bts_do_rate_ctr_inc(const struct gprs_rlcmac_bts *bts, unsigned int ctr_id) {
+	rate_ctr_inc(rate_ctr_group_get_ctr(bts->ratectrs, ctr_id));
 }
 
-static inline void bts_do_rate_ctr_add(struct gprs_rlcmac_bts *bts, unsigned int ctr_id, int inc) {
-	rate_ctr_add(&bts->ratectrs->ctr[ctr_id], inc);
+static inline void bts_do_rate_ctr_add(const struct gprs_rlcmac_bts *bts, unsigned int ctr_id, int inc) {
+	rate_ctr_add(rate_ctr_group_get_ctr(bts->ratectrs, ctr_id), inc);
 }
 
 static inline void bts_stat_item_add(struct gprs_rlcmac_bts *bts, unsigned int stat_id, int inc) {
-	int32_t val = osmo_stat_item_get_last(bts->statg->items[stat_id]);
-	osmo_stat_item_set(bts->statg->items[stat_id], val + inc);
+	struct osmo_stat_item *item = osmo_stat_item_group_get_item(bts->statg, stat_id);
+	int32_t val = osmo_stat_item_get_last(item);
+	osmo_stat_item_set(item, val + inc);
 }
 
+#define bts_stat_item_inc(bts, stat_id) bts_stat_item_add(bts, stat_id, 1)
+
+#define bts_stat_item_dec(bts, stat_id) bts_stat_item_add(bts, stat_id, -1)
+
 struct gprs_rlcmac_bts *bts_alloc(struct gprs_pcu *pcu, uint8_t bts_nr);
 
+struct gprs_rlcmac_sba *bts_alloc_sba(struct gprs_rlcmac_bts *bts, uint8_t ta);
+
 void bts_recalc_initial_cs(struct gprs_rlcmac_bts *bts);
 void bts_recalc_initial_mcs(struct gprs_rlcmac_bts *bts);
 void bts_recalc_max_cs(struct gprs_rlcmac_bts *bts);
@@ -349,6 +385,7 @@ void bts_set_max_mcs_ul(struct gprs_rlcmac_bts *bts, uint8_t mcs_ul);
 bool bts_cs_dl_is_supported(const struct gprs_rlcmac_bts *bts, enum CodingScheme cs);
 const struct llist_head* bts_ms_list(struct gprs_rlcmac_bts *bts);
 uint8_t bts_get_ms_pwr_alpha(const struct gprs_rlcmac_bts *bts);
+bool bts_all_pdch_allocated(const struct gprs_rlcmac_bts *bts);
 #ifdef __cplusplus
 }
 #endif
