@@ -361,10 +361,8 @@ static int write_ia_rest_egprs_uplink_sba(const gprs_rlcmac_ul_tbf *tbf, bitvec
  * Immediate assignment reject, sent on the CCCH/AGCH
  * see GSM 44.018, 9.1.20 + 10.5.2.30
  */
-int Encoding::write_immediate_assignment_reject(
-	bitvec *dest, uint16_t ra,
-	uint32_t ref_fn,
-	enum ph_burst_type burst_type)
+int Encoding::write_immediate_assignment_reject(bitvec *dest, uint16_t ra,
+	uint32_t ref_fn, enum ph_burst_type burst_type, uint8_t t3142)
 {
 	unsigned wp = 0;
 	int plen;
@@ -404,8 +402,8 @@ int Encoding::write_immediate_assignment_reject(
 		bitvec_write_field(dest, &wp, ref_fn % 51, 6);          // T3
 		bitvec_write_field(dest, &wp, ref_fn % 26, 5);          // T2
 
-		/* TODO: Make it configurable */
-		bitvec_write_field(dest, &wp, 20, 8); //Wait Indication 1
+		/* 10.5.2.43 Wait Indication */
+		bitvec_write_field(dest, &wp, t3142, 8);
 	}
 
 	plen = wp / 8;
@@ -591,8 +589,7 @@ static void gen_freq_params(Frequency_Parameters_t *freq_params,
 
 /* Generate Packet Uplink Assignment as per 3GPP TS 44.060, section 11.2.29.
  * NOTE: 'block' is expected to be zero-initialized by the caller. */
-void Encoding::write_packet_uplink_assignment(
-	RlcMacDownlink_t * block, uint8_t old_tfi,
+void write_packet_uplink_assignment(RlcMacDownlink_t *block, uint8_t old_tfi,
 	uint8_t old_downlink, uint32_t tlli, uint8_t use_tlli,
 	const struct gprs_rlcmac_ul_tbf *tbf, uint8_t poll, uint8_t rrbp, uint8_t alpha,
 	uint8_t gamma, int8_t ta_idx, bool use_egprs)
@@ -663,7 +660,7 @@ void Encoding::write_packet_uplink_assignment(
 	}
 
 	/* Packet Timing Advance (if known) */
-	if (tbf->ta() <= 63) { /* { 0 | 1  < TIMING_ADVANCE_VALUE : bit (6) > } */
+	if (gsm48_ta_is_valid(tbf->ta())) { /* { 0 | 1  < TIMING_ADVANCE_VALUE : bit (6) > } */
 		pta->Exist_TIMING_ADVANCE_VALUE = 0x01;  // Present
 		pta->TIMING_ADVANCE_VALUE       = tbf->ta();
 	}
@@ -711,11 +708,11 @@ void Encoding::write_packet_uplink_assignment(
 
 /* Generate Packet Downlink Assignment as per 3GPP TS 44.060, section 11.2.7.
  * NOTE: 'block' is expected to be zero-initialized by the caller. */
-void Encoding::write_packet_downlink_assignment(RlcMacDownlink_t * block,
+void write_packet_downlink_assignment(RlcMacDownlink_t * block,
 	bool old_tfi_is_valid, uint8_t old_tfi, uint8_t old_downlink,
 	const struct gprs_rlcmac_dl_tbf *tbf, uint8_t poll, uint8_t rrbp,
 	uint8_t alpha, uint8_t gamma, int8_t ta_idx,
-	uint8_t ta_ts, bool use_egprs)
+	uint8_t ta_ts, bool use_egprs, uint8_t control_ack)
 {
 	PDA_AdditionsR99_t *pda_r99;
 
@@ -742,7 +739,7 @@ void Encoding::write_packet_downlink_assignment(RlcMacDownlink_t * block,
 
 	block->u.Packet_Downlink_Assignment.MAC_MODE            = 0x0;          // Dynamic Allocation
 	block->u.Packet_Downlink_Assignment.RLC_MODE            = RLC_MODE_ACKNOWLEDGED;
-	block->u.Packet_Downlink_Assignment.CONTROL_ACK         = tbf->was_releasing; // NW establishes no new DL TBF for the MS with running timer T3192
+	block->u.Packet_Downlink_Assignment.CONTROL_ACK         = control_ack; // NW establishes no new DL TBF for the MS with running timer T3192
 	block->u.Packet_Downlink_Assignment.TIMESLOT_ALLOCATION = 0;   // timeslot(s)
 	for (tn = 0; tn < 8; tn++) {
 		if (tbf->pdch[tn])
@@ -844,20 +841,12 @@ int Encoding::write_paging_request(bitvec * dest, const struct osmo_mobile_ident
  * The index of the array show_rbb is the bit position inside the rbb
  * (show_rbb[63] relates to BSN ssn-1)
  */
-void Encoding::encode_rbb(const char *show_rbb, uint8_t *rbb)
+void Encoding::encode_rbb(const char *show_rbb, bitvec *bv)
 {
-	uint8_t rbb_byte = 0;
-
 	// RECEIVE_BLOCK_BITMAP
 	for (int i = 0; i < 64; i++) {
 		/* Set bit at the appropriate position (see 3GPP TS 44.060 9.1.8.1) */
-		if (show_rbb[i] == 'R')
-			rbb_byte |= 1<< (7-(i%8));
-
-		if ((i%8) == 7) {
-			rbb[i/8] = rbb_byte;
-			rbb_byte = 0;
-		}
+		bitvec_set_bit(bv, show_rbb[i] == 'R' ? ONE : ZERO);
 	}
 }
 
@@ -892,10 +881,22 @@ static void write_packet_uplink_ack_gprs(
 	bitvec_write_field(dest, &wp, mcs_chan_code(tbf->current_cs()), 2); // CHANNEL_CODING_COMMAND
 	write_packet_ack_nack_desc_gprs(dest, wp, window, is_final);
 
-	bitvec_write_field(dest, &wp, 1, 1); // 1: have CONTENTION_RESOLUTION_TLLI
-	bitvec_write_field(dest, &wp, tbf->tlli(), 32); // CONTENTION_RESOLUTION_TLLI
+	if (tbf->is_tlli_valid()) {
+		bitvec_write_field(dest, &wp, 1, 1); // 1: have CONTENTION_RESOLUTION_TLLI
+		bitvec_write_field(dest, &wp, tbf->tlli(), 32); // CONTENTION_RESOLUTION_TLLI
+	} else {
+		bitvec_write_field(dest, &wp, 0, 1); // 0: don't have CONTENTION_RESOLUTION_TLLI
+	}
+
+	if (gsm48_ta_is_valid(tbf->ta())) {
+		bitvec_write_field(dest, &wp, 1, 1); // 1: have Packet Timing Advance IE (TS 44.060 12.12)
+		bitvec_write_field(dest, &wp, 1, 1); // 1: have TIMING_ADVANCE_VALUE
+		bitvec_write_field(dest, &wp, tbf->ta(), 6); // TIMING_ADVANCE_VALUE
+		bitvec_write_field(dest, &wp, 0, 1); // 0: don't have TIMING_ADVANCE_INDEX
+	} else {
+		bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Packet Timing Advance
+	}
 
-	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Packet Timing Advance
 	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Power Control Parameters
 	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Extension Bits
 	bitvec_write_field(dest, &wp, 0, 1); // fixed 0
@@ -1074,10 +1075,25 @@ static void write_packet_uplink_ack_egprs(
 	bitvec_write_field(dest, &wp, 1, 1); // PRE_EMPTIVE_TRANSMISSION, TODO: This resembles GPRS, change it?
 	bitvec_write_field(dest, &wp, 0, 1); // 0: no PRR_RETRANSMISSION_REQUEST, TODO: clarify
 	bitvec_write_field(dest, &wp, 0, 1); // 0: no ARAC_RETRANSMISSION_REQUEST, TODO: clarify
-	bitvec_write_field(dest, &wp, 1, 1); // 1: have CONTENTION_RESOLUTION_TLLI
-	bitvec_write_field(dest, &wp, tbf->tlli(), 32); // CONTENTION_RESOLUTION_TLLI
+
+	if (tbf->is_tlli_valid()) {
+		bitvec_write_field(dest, &wp, 1, 1); // 1: have CONTENTION_RESOLUTION_TLLI
+		bitvec_write_field(dest, &wp, tbf->tlli(), 32); // CONTENTION_RESOLUTION_TLLI
+	} else {
+		bitvec_write_field(dest, &wp, 0, 1); // 0: don't have CONTENTION_RESOLUTION_TLLI
+	}
+
 	bitvec_write_field(dest, &wp, 1, 1); // TBF_EST (enabled)
-	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Packet Timing Advance
+
+	if (gsm48_ta_is_valid(tbf->ta())) {
+		bitvec_write_field(dest, &wp, 1, 1); // 1: have Packet Timing Advance IE (TS 44.060 12.12)
+		bitvec_write_field(dest, &wp, 1, 1); // 1: have TIMING_ADVANCE_VALUE
+		bitvec_write_field(dest, &wp, tbf->ta(), 6); // TIMING_ADVANCE_VALUE
+		bitvec_write_field(dest, &wp, 0, 1); // 0: don't have TIMING_ADVANCE_INDEX
+	} else {
+		bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Packet Timing Advance
+	}
+
 	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Packet Extended Timing Advance
 	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Power Control Parameters
 	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have Extension Bits
@@ -1090,9 +1106,8 @@ static void write_packet_uplink_ack_egprs(
 	bitvec_write_field(dest, &wp, 0, 1); // 0: don't have REL 5
 };
 
-void Encoding::write_packet_uplink_ack(
-	bitvec * dest, struct gprs_rlcmac_ul_tbf *tbf, bool is_final,
-	uint8_t rrbp)
+void write_packet_uplink_ack(struct bitvec *dest, struct gprs_rlcmac_ul_tbf *tbf,
+			     bool is_final, uint8_t rrbp)
 {
 	unsigned wp = 0;
 
@@ -1712,8 +1727,7 @@ void Encoding::rlc_data_to_dl_append_egprs_li_padding(
  * section 7.1.3.2.1 On receipt of a PACKET RESOURCE REQUEST message
  * 8.1.2.5 Establishment of uplink TBF
  */
-void Encoding::write_packet_access_reject(
-	bitvec * dest, uint32_t tlli)
+void write_packet_access_reject(struct bitvec *dest, uint32_t tlli, unsigned long t3172_ms)
 {
 	unsigned wp = 0;
 
@@ -1727,9 +1741,15 @@ void Encoding::write_packet_access_reject(
 	bitvec_write_field(dest, &wp, 0x0, 1);  //  TLLI / G-RNTI : bit (32)
 	bitvec_write_field(dest, &wp, tlli, 32); // CONTENTION_RESOLUTION_TLLI
 	bitvec_write_field(dest, &wp, 1, 1);  //  WAIT_INDICATION size in seconds
-	/* TODO: make it configurable */
-	bitvec_write_field(dest, &wp, 5, 8);  //  WAIT_INDICATION value
-	bitvec_write_field(dest, &wp, 0, 1);  //  WAIT_INDICATION size in seconds
+
+	/* WAIT_INDICATION, WAIT_INDICATION_SIZE */
+	if (t3172_ms / 20 <= 255) { /* In units of 20 milliseconds */
+		bitvec_write_field(dest, &wp, t3172_ms/20, 8);
+		bitvec_write_field(dest, &wp, 1, 1);
+	} else { /* value too big to fit in ms, do it in seconds */
+		bitvec_write_field(dest, &wp, t3172_ms/1000, 8);
+		bitvec_write_field(dest, &wp, 0, 1);
+	}
 }
 
 void write_packet_neighbour_cell_data(RlcMacDownlink_t *block,
