@@ -33,11 +33,13 @@
 #include <osmocom/gsm/protocol/gsm_23_003.h>
 #include <osmocom/gprs/protocol/gsm_08_16.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/stats.h>
 #include <osmocom/gsm/gsm48.h>
 #include "coding_scheme.h"
 #include "tbf_dl.h"
 #include "llc.h"
 #include "gprs_rlcmac.h"
+#include "bts_pch_timer.h"
 
 /* Tuning parameters for BSSGP flow control */
 #define FC_DEFAULT_LIFE_TIME_SECS 10		/* experimental value, 10s */
@@ -53,6 +55,19 @@ extern void *tall_pcu_ctx;
 extern uint16_t spoof_mcc, spoof_mnc;
 extern bool spoof_mnc_3_digits;
 
+static const struct rate_ctr_desc sgsn_ctr_description[] = {
+	[SGSN_CTR_RX_PAGING_CS] = { "rx_paging_cs", "Amount of paging CS requests received" },
+	[SGSN_CTR_RX_PAGING_PS] = { "rx_paging_ps", "Amount of paging PS requests received" },
+};
+
+static const struct rate_ctr_group_desc sgsn_ctrg_desc = {
+	.group_name_prefix = "pcu:sgsn",
+	.group_description = "SGSN Statistics",
+	.class_id = OSMO_STATS_CLASS_SUBSCRIBER,
+	.num_ctr = ARRAY_SIZE(sgsn_ctr_description),
+	.ctr_desc = sgsn_ctr_description,
+};
+
 static void bvc_timeout(void *_priv);
 
 static int parse_ra_cap(struct tlv_parsed *tp, MS_Radio_Access_capability_t *rac)
@@ -93,12 +108,8 @@ static int gprs_bssgp_pcu_rx_dl_ud(struct msgb *msg, struct tlv_parsed *tp)
 	uint8_t egprs_ms_class = 0;
 	int rc;
 	MS_Radio_Access_capability_t rac;
-	/* TODO: is it really necessary to initialize this as a "000" IMSI? It seems, the function should just return an
-	 * error if no IMSI IE was found. */
-	struct osmo_mobile_identity mi_imsi = {
-		.type = GSM_MI_TYPE_TMSI,
-	};
-	OSMO_STRLCPY_ARRAY(mi_imsi.imsi, "000");
+	const char *imsi = NULL;
+	struct osmo_mobile_identity mi_imsi;
 
 	budh = (struct bssgp_ud_hdr *)msgb_bssgph(msg);
 	tlli = ntohl(budh->tlli);
@@ -129,6 +140,7 @@ static int gprs_bssgp_pcu_rx_dl_ud(struct msgb *msg, struct tlv_parsed *tp)
 			LOGP(DBSSGP, LOGL_NOTICE, "Failed to parse IMSI IE (rc=%d)\n", rc);
 			return bssgp_tx_status(BSSGP_CAUSE_COND_IE_ERR, NULL, msg);
 		}
+		imsi = &mi_imsi.imsi[0];
 	}
 
 	/* parse ms radio access capability */
@@ -165,14 +177,87 @@ static int gprs_bssgp_pcu_rx_dl_ud(struct msgb *msg, struct tlv_parsed *tp)
 				"TLLI (old) IE\n");
 	}
 
-	LOGP(DBSSGP, LOGL_INFO, "LLC [SGSN -> PCU] = TLLI: 0x%08x IMSI: %s len: %d\n", tlli, mi_imsi.imsi, len);
+	LOGP(DBSSGP, LOGL_INFO, "LLC [SGSN -> PCU] = TLLI: 0x%08x IMSI: %s len: %d\n",
+	     tlli, imsi ? : "none", len);
 
-	return dl_tbf_handle(the_pcu->bssgp.bts, tlli, tlli_old, mi_imsi.imsi,
-			ms_class, egprs_ms_class, delay_csec, data, len);
+	return dl_tbf_handle(the_pcu->bssgp.bts, tlli, tlli_old, imsi, ms_class,
+			     egprs_ms_class, delay_csec, data, len);
+}
+
+/* 3GPP TS 48.018 Table 10.3.2. Returns 0 on success, suggested BSSGP cause otherwise */
+static unsigned int get_paging_cs_mi(struct paging_req_cs *req, const struct tlv_parsed *tp)
+{
+	int rc;
+
+	req->chan_needed = tlvp_val8(tp, BSSGP_IE_CHAN_NEEDED, 0);
+
+	if (!TLVP_PRESENT(tp, BSSGP_IE_IMSI)) {
+		LOGP(DBSSGP, LOGL_ERROR, "IMSI Mobile Identity mandatory IE not found\n");
+		return BSSGP_CAUSE_MISSING_MAND_IE;
+	}
+
+	rc = osmo_mobile_identity_decode(&req->mi_imsi, TLVP_VAL(tp, BSSGP_IE_IMSI),
+					 TLVP_LEN(tp, BSSGP_IE_IMSI), true);
+	if (rc < 0 || req->mi_imsi.type != GSM_MI_TYPE_IMSI) {
+		LOGP(DBSSGP, LOGL_ERROR, "Invalid IMSI Mobile Identity\n");
+		return BSSGP_CAUSE_INV_MAND_INF;
+	}
+	req->mi_imsi_present = true;
+
+	/* TMSI is optional */
+	req->mi_tmsi_present = false;
+	if (TLVP_PRESENT(tp, BSSGP_IE_TMSI)) {
+		/* Be safe against an evil SGSN - check the length */
+		if (TLVP_LEN(tp, BSSGP_IE_TMSI) != GSM23003_TMSI_NUM_BYTES) {
+			LOGP(DBSSGP, LOGL_NOTICE, "TMSI IE has odd length (!= 4)\n");
+			return BSSGP_CAUSE_COND_IE_ERR;
+		}
+
+		/* NOTE: TMSI (unlike IMSI) IE comes without MI type header */
+		req->mi_tmsi = (struct osmo_mobile_identity){
+			.type = GSM_MI_TYPE_TMSI,
+		};
+		req->mi_tmsi.tmsi = osmo_load32be(TLVP_VAL(tp, BSSGP_IE_TMSI));
+		req->mi_tmsi_present = true;
+	}
+
+	if (TLVP_PRESENT(tp, BSSGP_IE_TLLI))
+		req->tlli = osmo_load32be(TLVP_VAL(tp, BSSGP_IE_TLLI));
+	else
+		req->tlli = GSM_RESERVED_TMSI;
+
+	return 0;
+}
+
+static int gprs_bssgp_pcu_rx_paging_cs(struct msgb *msg, const struct tlv_parsed *tp)
+{
+	struct paging_req_cs req;
+	struct gprs_rlcmac_bts *bts;
+	struct GprsMs *ms;
+	int rc;
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(the_pcu->bssgp.ctrs, SGSN_CTR_RX_PAGING_CS));
+
+	if ((rc = get_paging_cs_mi(&req, tp)) > 0)
+		return bssgp_tx_status((enum gprs_bssgp_cause) rc, NULL, msg);
+
+	/* We need to page all BTSs since even if a BTS has a matching MS, it
+	 * may have already moved to a newer BTS. On Each BTS, if the MS is
+	 * known, then bts_add_paging() can optimize and page only on PDCHs the
+	 * target MS is using. */
+	llist_for_each_entry(bts, &the_pcu->bts_list, list) {
+		/* TODO: Match by TMSI before IMSI if present?! */
+		ms = bts_ms_by_tlli(bts, req.tlli, req.tlli);
+		if (!ms && req.mi_imsi_present)
+			ms = bts_ms_by_imsi(bts, req.mi_imsi.imsi);
+		bts_add_paging(bts, &req, ms);
+	}
+
+	return 0;
 }
 
 /* Returns 0 on success, suggested BSSGP cause otherwise */
-static unsigned int get_paging_mi(struct osmo_mobile_identity *mi, const struct tlv_parsed *tp)
+static unsigned int get_paging_ps_mi(struct osmo_mobile_identity *mi, const struct tlv_parsed *tp)
 {
 	/* Use TMSI (if present) or IMSI */
 	if (TLVP_PRESENT(tp, BSSGP_IE_TMSI)) {
@@ -202,22 +287,6 @@ static unsigned int get_paging_mi(struct osmo_mobile_identity *mi, const struct
 	return 0;
 }
 
-static int gprs_bssgp_pcu_rx_paging_cs(struct msgb *msg, const struct tlv_parsed *tp)
-{
-	struct osmo_mobile_identity mi;
-	struct gprs_rlcmac_bts *bts;
-	int rc;
-
-	if ((rc = get_paging_mi(&mi, tp)) > 0)
-		return bssgp_tx_status((enum gprs_bssgp_cause) rc, NULL, msg);
-
-	/* FIXME: look if MS is attached a specific BTS and then only page on that one? */
-	llist_for_each_entry(bts, &the_pcu->bts_list, list) {
-		bts_add_paging(bts, tlvp_val8(tp, BSSGP_IE_CHAN_NEEDED, 0), &mi);
-	}
-	return 0;
-}
-
 static int gprs_bssgp_pcu_rx_paging_ps(struct msgb *msg, const struct tlv_parsed *tp)
 {
 	struct osmo_mobile_identity mi_imsi;
@@ -226,6 +295,8 @@ static int gprs_bssgp_pcu_rx_paging_ps(struct msgb *msg, const struct tlv_parsed
 	uint16_t pgroup;
 	int rc;
 
+	rate_ctr_inc(rate_ctr_group_get_ctr(the_pcu->bssgp.ctrs, SGSN_CTR_RX_PAGING_PS));
+
 	if (!TLVP_PRESENT(tp, BSSGP_IE_IMSI)) {
 		LOGP(DBSSGP, LOGL_ERROR, "No IMSI\n");
 		return bssgp_tx_status(BSSGP_CAUSE_MISSING_MAND_IE, NULL, msg);
@@ -242,12 +313,19 @@ static int gprs_bssgp_pcu_rx_paging_ps(struct msgb *msg, const struct tlv_parsed
 		return bssgp_tx_status(BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
 	}
 
-	if ((rc = get_paging_mi(&paging_mi, tp)) > 0)
+	if ((rc = get_paging_ps_mi(&paging_mi, tp)) > 0)
 		return bssgp_tx_status((enum gprs_bssgp_cause) rc, NULL, msg);
 
 	/* FIXME: look if MS is attached a specific BTS and then only page on that one? */
 	llist_for_each_entry(bts, &the_pcu->bts_list, list) {
-		gprs_rlcmac_paging_request(bts, &paging_mi, pgroup);
+		if (bts_pch_timer_get_by_imsi(bts, mi_imsi.imsi)) {
+			LOGP(DBSSGP, LOGL_INFO, "PS-Paging request already pending for IMSI=%s\n", mi_imsi.imsi);
+			bts_do_rate_ctr_inc(bts, CTR_PCH_REQUESTS_ALREADY);
+			continue;
+		}
+		if (gprs_rlcmac_paging_request(bts, &paging_mi, pgroup) < 0)
+			continue;
+		bts_pch_timer_start(bts, &paging_mi, mi_imsi.imsi);
 	}
 	return 0;
 }
@@ -414,8 +492,8 @@ static int gprs_bssgp_pcu_rcvmsg(struct msgb *msg)
 	struct bssgp_ud_hdr *budh = (struct bssgp_ud_hdr *) msgb_bssgph(msg);
 	struct tlv_parsed tp;
 	enum bssgp_pdu_type pdu_type = (enum bssgp_pdu_type) bgph->pdu_type;
-	enum gprs_bssgp_cause cause = BSSGP_CAUSE_OML_INTERV;
 	uint16_t ns_bvci = msgb_bvci(msg), nsei = msgb_nsei(msg);
+	uint16_t bvci;
 	int data_len;
 	int rc = 0;
 	struct bssgp_bvc_ctx *bctx;
@@ -457,20 +535,18 @@ static int gprs_bssgp_pcu_rcvmsg(struct msgb *msg)
 	}
 
 	if (pdu_type == BSSGP_PDUT_BVC_RESET) {
-		rc = bssgp_rcvmsg(msg);
-		if (ns_bvci != BVCI_SIGNALLING)
-			return rc;
+		if (ns_bvci != BVCI_SIGNALLING || !TLVP_PRESENT(&tp, BSSGP_IE_BVCI)) {
+			LOGP(DBSSGP, LOGL_ERROR, "Rx an invalid BVC-RESET %s\n", msgb_hexdump(msg));
+			return bssgp_tx_status(BSSGP_CAUSE_INV_MAND_INF, NULL, msg);
+		}
 
-		if (TLVP_PRES_LEN(&tp, BSSGP_IE_CAUSE, 1))
-			cause = (enum gprs_bssgp_cause)*TLVP_VAL(&tp, BSSGP_IE_CAUSE);
-		else
-			LOGP(DBSSGP, LOGL_ERROR, "NSEI=%u BVC RESET without cause?!\n", nsei);
+		bvci = tlvp_val16be(&tp, BSSGP_IE_BVCI);
+		if (bvci != BVCI_SIGNALLING && bvci != the_pcu->bssgp.bctx->bvci) {
+			LOGP(DBSSGP, LOGL_ERROR, "Rx BVC-RESET for an unknown BVCI %d\n", bvci);
+			return bssgp_tx_status(BSSGP_CAUSE_UNKNOWN_BVCI, &bvci, msg);
+		}
 
-		rc = bssgp_tx_bvc_ptp_reset(nsei, cause);
-		if (rc < 0)
-			LOGP(DBSSGP, LOGL_ERROR, "NSEI=%u BVC PTP reset procedure failed: %d\n", nsei, rc);
-
-		return rc;
+		return bssgp_rcvmsg(msg);
 	}
 
 	/* look-up or create the BTS context for this BVC */
@@ -486,8 +562,8 @@ static int gprs_bssgp_pcu_rcvmsg(struct msgb *msg)
 	if (bctx)
 	{
 		log_set_context(LOG_CTX_GB_BVC, bctx);
-		rate_ctr_inc(&bctx->ctrg->ctr[BSSGP_CTR_PKTS_IN]);
-		rate_ctr_add(&bctx->ctrg->ctr[BSSGP_CTR_BYTES_IN], msgb_bssgp_len(msg));
+		rate_ctr_inc(rate_ctr_group_get_ctr(bctx->ctrg, BSSGP_CTR_PKTS_IN));
+		rate_ctr_add(rate_ctr_group_get_ctr(bctx->ctrg, BSSGP_CTR_BYTES_IN), msgb_bssgp_len(msg));
 	}
 
 	if (ns_bvci == BVCI_SIGNALLING)
@@ -558,12 +634,43 @@ static void handle_nm_status(struct osmo_bssgp_prim *bp)
 int bssgp_prim_cb(struct osmo_prim_hdr *oph, void *ctx)
 {
 	struct osmo_bssgp_prim *bp;
+	int rc;
+	enum gprs_bssgp_cause cause;
 	bp = container_of(oph, struct osmo_bssgp_prim, oph);
 
 	switch (oph->sap) {
 	case SAP_BSSGP_NM:
-		if (oph->primitive == PRIM_NM_STATUS)
+		switch (oph->primitive) {
+		case PRIM_NM_STATUS:
 			handle_nm_status(bp);
+			break;
+		case PRIM_NM_BVC_RESET:
+			/* received a BVC PTP reset */
+			LOGP(DPCU, LOGL_INFO, "Rx BVC_RESET on bvci %d\n", bp->bvci);
+			/* Rx Reset from SGSN */
+			if (bp->bvci == BVCI_SIGNALLING) {
+				if (TLVP_PRES_LEN(bp->tp, BSSGP_IE_CAUSE, 1))
+					cause = (enum gprs_bssgp_cause)*TLVP_VAL(bp->tp, BSSGP_IE_CAUSE);
+				else {
+					LOGP(DBSSGP, LOGL_ERROR, "NSEI=%u BVC RESET without cause?!\n", bp->nsei);
+					break;
+				}
+
+				rc = bssgp_tx_bvc_ptp_reset(bp->nsei, cause);
+				if (rc < 0) {
+					LOGP(DBSSGP, LOGL_ERROR, "NSEI=%u BVC PTP reset procedure failed: %d\n", bp->nsei, rc);
+					break;
+				}
+				the_pcu->bssgp.bvc_sig_reset = 1;
+				the_pcu->bssgp.bvc_reset = 0;
+				the_pcu->bssgp.bvc_unblocked = 0;
+			} else if (bp->bvci == the_pcu->bssgp.bctx->bvci) {
+				the_pcu->bssgp.bvc_reset = 1;
+				the_pcu->bssgp.bvc_unblocked = 0;
+				bvc_timeout(NULL);
+			}
+			break;
+		}
 		break;
 	case SAP_BSSGP_RIM:
 		return handle_rim(bp);
@@ -946,7 +1053,7 @@ static void bvc_timeout(void *_priv)
 	if (!the_pcu->bssgp.bvc_sig_reset) {
 		LOGP(DBSSGP, LOGL_INFO, "Sending reset on BVCI 0\n");
 		bssgp_tx_bvc_reset(the_pcu->bssgp.bctx, 0, BSSGP_CAUSE_OML_INTERV);
-		secs = osmo_tdef_get(the_pcu->T_defs, 2, OSMO_TDEF_S, -1);
+		secs = osmo_tdef_get(the_pcu->T_defs, -102, OSMO_TDEF_S, -1);
 		osmo_timer_schedule(&the_pcu->bssgp.bvc_timer, secs, 0);
 		return;
 	}
@@ -955,7 +1062,7 @@ static void bvc_timeout(void *_priv)
 		LOGP(DBSSGP, LOGL_INFO, "Sending reset on BVCI %d\n",
 			the_pcu->bssgp.bctx->bvci);
 		bssgp_tx_bvc_reset(the_pcu->bssgp.bctx, the_pcu->bssgp.bctx->bvci, BSSGP_CAUSE_OML_INTERV);
-		secs = osmo_tdef_get(the_pcu->T_defs, 2, OSMO_TDEF_S, -1);
+		secs = osmo_tdef_get(the_pcu->T_defs, -102, OSMO_TDEF_S, -1);
 		osmo_timer_schedule(&the_pcu->bssgp.bvc_timer, secs, 0);
 		return;
 	}
@@ -964,7 +1071,7 @@ static void bvc_timeout(void *_priv)
 		LOGP(DBSSGP, LOGL_INFO, "Sending unblock on BVCI %d\n",
 			the_pcu->bssgp.bctx->bvci);
 		bssgp_tx_bvc_unblock(the_pcu->bssgp.bctx);
-		secs = osmo_tdef_get(the_pcu->T_defs, 1, OSMO_TDEF_S, -1);
+		secs = osmo_tdef_get(the_pcu->T_defs, -101, OSMO_TDEF_S, -1);
 		osmo_timer_schedule(&the_pcu->bssgp.bvc_timer, secs, 0);
 		return;
 	}
@@ -992,12 +1099,13 @@ static int ns_configure_nse(struct gprs_rlcmac_bts *bts,
 			    const uint16_t *nsvci,
 			    uint16_t valid)
 {
-	int i, rc;
+	unsigned int i;
+	int rc;
 	uint16_t binds = 0;
 	bool nsvcs = false;
 	struct gprs_ns2_vc *nsvc;
 	struct gprs_ns2_vc_bind *bind[PCU_IF_NUM_NSVC] = { };
-	char name[5];
+	char name[16];
 	bool sns_configured = false;
 
 	if (!valid)
@@ -1016,8 +1124,9 @@ static int ns_configure_nse(struct gprs_rlcmac_bts *bts,
 		if (!(valid & (1 << i)))
 			continue;
 
-		if (!gprs_ns2_ip_bind_by_sockaddr(the_pcu->nsi, &local[i])) {
-			snprintf(name, sizeof(name), "pcu%d", i);
+		bind[i] = gprs_ns2_ip_bind_by_sockaddr(the_pcu->nsi, &local[i]);
+		if (!bind[i]) {
+			snprintf(name, sizeof(name), "pcu%u", i);
 			rc = gprs_ns2_ip_bind(the_pcu->nsi, name, &local[i], 0, &bind[i]);
 			if (rc < 0) {
 				LOGP(DBSSGP, LOGL_ERROR, "Failed to bind to %s\n", osmo_sockaddr_to_str(&local[i]));
@@ -1031,6 +1140,11 @@ static int ns_configure_nse(struct gprs_rlcmac_bts *bts,
 					continue;
 				}
 			}
+
+			if (the_pcu->vty.ns_ip_dscp != -1)
+				gprs_ns2_ip_bind_set_dscp(bind[i], the_pcu->vty.ns_ip_dscp);
+			if (the_pcu->vty.ns_priority != -1)
+				gprs_ns2_ip_bind_set_priority(bind[i], the_pcu->vty.ns_priority);
 		}
 
 		binds |= 1 << i;
@@ -1184,6 +1298,7 @@ struct gprs_bssgp_pcu *gprs_bssgp_init(
 		the_pcu->bssgp.bts->nse = NULL;
 		return NULL;
 	}
+	the_pcu->bssgp.bctx->is_sgsn = false;
 	the_pcu->bssgp.bctx->ra_id.mcc = spoof_mcc ? : mcc;
 	if (spoof_mnc) {
 		the_pcu->bssgp.bctx->ra_id.mnc = spoof_mnc;
@@ -1198,11 +1313,15 @@ struct gprs_bssgp_pcu *gprs_bssgp_init(
 
 	osmo_timer_setup(&the_pcu->bssgp.bvc_timer, bvc_timeout, bts);
 
+	the_pcu->bssgp.ctrs = rate_ctr_group_alloc(the_pcu, &sgsn_ctrg_desc, 0);
+	OSMO_ASSERT(the_pcu->bssgp.ctrs)
+
 	return &the_pcu->bssgp;
 }
 
 void gprs_bssgp_destroy(struct gprs_rlcmac_bts *bts)
 {
+	rate_ctr_group_free(the_pcu->bssgp.ctrs);
 	osmo_timer_del(&the_pcu->bssgp.bvc_timer);
 
 	/* FIXME: blocking... */
