@@ -79,11 +79,10 @@ static struct gpr_ms_callback gprs_default_cb = {
 	.ms_active = gprs_default_cb_ms_active,
 };
 
-void ms_timeout(void *data)
+static void ms_release_timer_cb(void *data)
 {
 	struct GprsMs *ms = (struct GprsMs *) data;
-	LOGP(DRLCMAC, LOGL_INFO, "Timeout for MS object, TLLI = 0x%08x\n",
-		ms_tlli(ms));
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Release timer expired\n");
 
 	if (ms->timer.data) {
 		ms->timer.data = NULL;
@@ -116,7 +115,7 @@ struct GprsMs *ms_alloc(struct gprs_rlcmac_bts *bts, uint32_t tlli)
 
 	ms->imsi[0] = '\0';
 	memset(&ms->timer, 0, sizeof(ms->timer));
-	ms->timer.cb = ms_timeout;
+	ms->timer.cb = ms_release_timer_cb;
 	llc_queue_init(&ms->llc_queue);
 
 	ms_set_mode(ms, GPRS);
@@ -148,7 +147,7 @@ static int ms_talloc_destructor(struct GprsMs *ms)
 {
 	struct llist_item *pos, *tmp;
 
-	LOGP(DRLCMAC, LOGL_INFO, "Destroying MS object, TLLI = 0x%08x\n", ms_tlli(ms));
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Destroying MS object\n");
 
 	ms_set_reserved_slots(ms, NULL, 0, 0);
 
@@ -218,22 +217,26 @@ void ms_unref(struct GprsMs *ms)
 		ms_update_status(ms);
 }
 
-void ms_start_timer(struct GprsMs *ms)
+static void ms_release_timer_start(struct GprsMs *ms)
 {
 	if (ms->delay == 0)
 		return;
 
+	LOGPMS(ms, DRLCMAC, LOGL_DEBUG, "Schedule MS release in %u secs\n", ms->delay);
+
 	if (!ms->timer.data)
 		ms->timer.data = ms_ref(ms);
 
 	osmo_timer_schedule(&ms->timer, ms->delay, 0);
 }
 
-void ms_stop_timer(struct GprsMs *ms)
+static void ms_release_timer_stop(struct GprsMs *ms)
 {
 	if (!ms->timer.data)
 		return;
 
+	LOGPMS(ms, DRLCMAC, LOGL_DEBUG, "Cancel scheduled MS release\n");
+
 	osmo_timer_del(&ms->timer);
 	ms->timer.data = NULL;
 	ms_unref(ms);
@@ -295,8 +298,7 @@ static void ms_attach_ul_tbf(struct GprsMs *ms, struct gprs_rlcmac_ul_tbf *tbf)
 	if (ms->ul_tbf == tbf)
 		return;
 
-	LOGP(DRLCMAC, LOGL_INFO, "Attaching TBF to MS object, TLLI = 0x%08x, TBF = %s\n",
-		ms_tlli(ms), tbf_name((struct gprs_rlcmac_tbf *)tbf));
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Attaching UL TBF: %s\n", tbf_name((struct gprs_rlcmac_tbf *)tbf));
 
 	ms_ref(ms);
 
@@ -306,7 +308,7 @@ static void ms_attach_ul_tbf(struct GprsMs *ms, struct gprs_rlcmac_ul_tbf *tbf)
 	ms->ul_tbf = tbf;
 
 	if (tbf)
-		ms_stop_timer(ms);
+		ms_release_timer_stop(ms);
 
 	ms_unref(ms);
 }
@@ -316,8 +318,7 @@ static void ms_attach_dl_tbf(struct GprsMs *ms, struct gprs_rlcmac_dl_tbf *tbf)
 	if (ms->dl_tbf == tbf)
 		return;
 
-	LOGP(DRLCMAC, LOGL_INFO, "Attaching TBF to MS object, TLLI = 0x%08x, TBF = %s\n",
-		ms_tlli(ms), tbf_name((struct gprs_rlcmac_tbf *)tbf));
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Attaching DL TBF: %s\n", tbf_name((struct gprs_rlcmac_tbf *)tbf));
 
 	ms_ref(ms);
 
@@ -327,7 +328,7 @@ static void ms_attach_dl_tbf(struct GprsMs *ms, struct gprs_rlcmac_dl_tbf *tbf)
 	ms->dl_tbf = tbf;
 
 	if (tbf)
-		ms_stop_timer(ms);
+		ms_release_timer_stop(ms);
 
 	ms_unref(ms);
 }
@@ -364,8 +365,8 @@ void ms_detach_tbf(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf)
 			return;
 	}
 
-	LOGP(DRLCMAC, LOGL_INFO, "Detaching TBF from MS object, TLLI = 0x%08x, TBF = %s\n",
-		ms_tlli(ms), tbf_name(tbf));
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Detaching TBF: %s\n",
+	       tbf_name(tbf));
 
 	if (tbf_ms(tbf) == ms)
 		tbf_set_ms(tbf, NULL);
@@ -374,7 +375,7 @@ void ms_detach_tbf(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf)
 		ms_set_reserved_slots(ms, NULL, 0, 0);
 
 		if (ms_tlli(ms) != 0)
-			ms_start_timer(ms);
+			ms_release_timer_start(ms);
 	}
 
 	ms_update_status(ms);
@@ -382,11 +383,9 @@ void ms_detach_tbf(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf)
 
 void ms_reset(struct GprsMs *ms)
 {
-	LOGP(DRLCMAC, LOGL_INFO,
-		"Clearing MS object, TLLI: 0x%08x, IMSI: '%s'\n",
-		ms_tlli(ms), ms_imsi(ms));
+	LOGPMS(ms, DRLCMAC, LOGL_INFO, "Clearing MS object\n");
 
-	ms_stop_timer(ms);
+	ms_release_timer_stop(ms);
 
 	ms->tlli = GSM_RESERVED_TMSI;
 	ms->new_dl_tlli = ms->tlli;
@@ -523,6 +522,18 @@ void ms_set_imsi(struct GprsMs *ms, const char *imsi)
 	osmo_strlcpy(ms->imsi, imsi, sizeof(ms->imsi));
 }
 
+uint16_t ms_paging_group(struct GprsMs *ms)
+{
+	uint16_t pgroup;
+	if (!ms_imsi_is_valid(ms))
+		return 0; /* 000 is the special "all paging" group */
+	if ((pgroup = imsi2paging_group(ms_imsi(ms))) > 999) {
+		LOGPMS(ms, DRLCMAC, LOGL_ERROR, "IMSI to paging group failed!\n");
+		return 0;
+	}
+	return pgroup;
+}
+
 void ms_set_ta(struct GprsMs *ms, uint8_t ta_)
 {
 	if (ta_ == ms->ta)
