@@ -35,42 +35,30 @@ extern "C" {
 }
 
 struct tbf_sched_candidates {
-	struct gprs_rlcmac_tbf *poll;
 	struct gprs_rlcmac_tbf *ul_ass;
 	struct gprs_rlcmac_tbf *dl_ass;
 	struct gprs_rlcmac_tbf *nacc;
 	struct gprs_rlcmac_ul_tbf *ul_ack;
 };
 
-static uint32_t sched_poll(struct gprs_rlcmac_bts *bts,
-		    uint8_t trx, uint8_t ts, uint32_t fn, uint8_t block_nr,
-		    struct tbf_sched_candidates *tbf_cand)
+static void get_ctrl_msg_tbf_candidates(const struct gprs_rlcmac_pdch *pdch,
+					struct tbf_sched_candidates *tbf_cand)
 {
 	struct gprs_rlcmac_ul_tbf *ul_tbf;
 	struct gprs_rlcmac_dl_tbf *dl_tbf;
 	struct llist_item *pos;
-	uint32_t poll_fn;
 
-	/* check special TBF for events */
-	poll_fn = fn + 4;
-	if ((block_nr % 3) == 2)
-		poll_fn ++;
-	poll_fn = poll_fn % GSM_MAX_FN;
-	llist_for_each_entry(pos, &bts->ul_tbfs, list) {
+	llist_for_each_entry(pos, &pdch->trx->ul_tbfs, list) {
 		ul_tbf = as_ul_tbf((struct gprs_rlcmac_tbf *)pos->entry);
 		OSMO_ASSERT(ul_tbf);
 		/* this trx, this ts */
-		if (ul_tbf->trx->trx_no != trx || !ul_tbf->is_control_ts(ts))
+		if (!ul_tbf->is_control_ts(pdch->ts_no))
 			continue;
-		/* polling for next uplink block */
-		if (ul_tbf->poll_scheduled() && ul_tbf->poll_fn == poll_fn)
-			tbf_cand->poll = ul_tbf;
-		if (ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_SEND_ACK))
+		if (tbf_ul_ack_rts(ul_tbf))
 			tbf_cand->ul_ack = ul_tbf;
-		if (ul_tbf->dl_ass_state_is(GPRS_RLCMAC_DL_ASS_SEND_ASS))
+		if (tbf_dl_ass_rts(ul_tbf))
 			tbf_cand->dl_ass = ul_tbf;
-		if (ul_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS)
-		    || ul_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
+		if (tbf_ul_ass_rts(ul_tbf))
 			tbf_cand->ul_ass = ul_tbf;
 		/* NACC ready to send. TFI assigned is needed to send messages */
 		if (ul_tbf->is_tfi_assigned() && ms_nacc_rts(ul_tbf->ms()))
@@ -78,32 +66,25 @@ static uint32_t sched_poll(struct gprs_rlcmac_bts *bts,
 /* FIXME: Is this supposed to be fair? The last TBF for each wins? Maybe use llist_add_tail and skip once we have all
 states? */
 	}
-	llist_for_each_entry(pos, &bts->dl_tbfs, list) {
+	llist_for_each_entry(pos, &pdch->trx->dl_tbfs, list) {
 		dl_tbf = as_dl_tbf((struct gprs_rlcmac_tbf *)pos->entry);
 		OSMO_ASSERT(dl_tbf);
 		/* this trx, this ts */
-		if (dl_tbf->trx->trx_no != trx || !dl_tbf->is_control_ts(ts))
+		if (!dl_tbf->is_control_ts(pdch->ts_no))
 			continue;
-		/* polling for next uplink block */
-		if (dl_tbf->poll_scheduled() && dl_tbf->poll_fn == poll_fn)
-			tbf_cand->poll = dl_tbf;
-		if (dl_tbf->dl_ass_state_is(GPRS_RLCMAC_DL_ASS_SEND_ASS))
+		if (tbf_dl_ass_rts(dl_tbf))
 			tbf_cand->dl_ass = dl_tbf;
-		if (dl_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS)
-		    || dl_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
+		if (tbf_ul_ass_rts(dl_tbf))
 			tbf_cand->ul_ass = dl_tbf;
 		/* NACC ready to send. TFI assigned is needed to send messages */
 		if (dl_tbf->is_tfi_assigned() && ms_nacc_rts(dl_tbf->ms()))
 			tbf_cand->nacc = dl_tbf;
 	}
-
-	return poll_fn;
 }
 
-static struct gprs_rlcmac_ul_tbf *sched_select_uplink(uint8_t trx, uint8_t ts, uint32_t fn,
-	uint8_t block_nr, struct gprs_rlcmac_pdch *pdch, bool require_gprs_only)
+static struct gprs_rlcmac_ul_tbf *sched_select_uplink(struct gprs_rlcmac_pdch *pdch, bool require_gprs_only)
 {
-	struct gprs_rlcmac_ul_tbf *tbf = NULL;
+	struct gprs_rlcmac_ul_tbf *tbf;
 	uint8_t i, tfi;
 
 	/* select uplink resource */
@@ -117,23 +98,19 @@ static struct gprs_rlcmac_ul_tbf *sched_select_uplink(uint8_t trx, uint8_t ts, u
 		/* we don't need to give resources in FINISHED state,
 		 * because we have received all blocks and only poll
 		 * for packet control ack. */
-		if (tbf->state_is_not(GPRS_RLCMAC_FLOW))
+		if (tbf->state_is_not(TBF_ST_FLOW))
 			continue;
 
 		if (require_gprs_only && tbf->is_egprs_enabled())
 			continue;
 
-		/* use this USF */
-		LOGP(DRLCMACSCHED, LOGL_DEBUG, "Received RTS for PDCH: TRX=%d "
-			"TS=%d FN=%d block_nr=%d scheduling USF=%d for "
-			"required uplink resource of UL TFI=%d\n", trx, ts, fn,
-			block_nr, tbf->m_usf[ts], tfi);
+		/* use this USF (tbf) */
 		/* next TBF to handle resource is the next one */
 		pdch->next_ul_tfi = (tfi + 1) & 31;
-		break;
+		return tbf;
 	}
 
-	return tbf;
+	return NULL;
 }
 
 struct msgb *sched_app_info(struct gprs_rlcmac_tbf *tbf) {
@@ -164,10 +141,8 @@ struct msgb *sched_app_info(struct gprs_rlcmac_tbf *tbf) {
 	return msg;
 }
 
-static struct msgb *sched_select_ctrl_msg(
-		    uint8_t trx, uint8_t ts, uint32_t fn,
-		    uint8_t block_nr, struct gprs_rlcmac_pdch *pdch,
-		    struct tbf_sched_candidates *tbfs)
+static struct msgb *sched_select_ctrl_msg(struct gprs_rlcmac_pdch *pdch, uint32_t fn,
+					  uint8_t block_nr, struct tbf_sched_candidates *tbfs)
 {
 	struct msgb *msg = NULL;
 	struct gprs_rlcmac_tbf *tbf = NULL;
@@ -175,6 +150,7 @@ static struct msgb *sched_select_ctrl_msg(
 						tbfs->dl_ass,
 						tbfs->ul_ack,
 						tbfs->nacc };
+	uint8_t ts = pdch->ts_no;
 
 	/* Send Packet Application Information first (ETWS primary notifications) */
 	msg = sched_app_info(tbfs->dl_ass);
@@ -190,18 +166,14 @@ static struct msgb *sched_select_ctrl_msg(
 			 * because they may kill the TBF when the CONTROL ACK is
 			 * received, thus preventing the others from being processed.
 			 */
-			if (tbf == tbfs->ul_ass && tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
-				msg = tbfs->ul_ass->create_packet_access_reject();
-			else if (tbf == tbfs->ul_ass && tbf->direction ==
-					GPRS_RLCMAC_DL_TBF)
-				if (tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
-					msg = tbfs->ul_ass->create_packet_access_reject();
-				else
-					msg = tbfs->ul_ass->create_ul_ass(fn, ts);
+			if (tbf == tbfs->ul_ass && tbf->ul_ass_state_is(TBF_UL_ASS_SEND_ASS_REJ))
+				msg = tbf_ul_ass_create_rlcmac_msg(tbfs->ul_ass, fn, ts);
+			else if (tbf == tbfs->ul_ass && tbf->direction == GPRS_RLCMAC_DL_TBF)
+				msg = tbf_ul_ass_create_rlcmac_msg(tbfs->ul_ass, fn, ts);
 			else if (tbf == tbfs->dl_ass && tbf->direction == GPRS_RLCMAC_UL_TBF)
-				msg = tbfs->dl_ass->create_dl_ass(fn, ts);
+				msg = tbf_dl_ass_create_rlcmac_msg(tbfs->dl_ass, fn, ts);
 			else if (tbf == tbfs->ul_ack)
-				msg = tbfs->ul_ack->create_ul_ack(fn, ts);
+				msg = tbf_ul_ack_create_rlcmac_msg(tbfs->ul_ack, fn, ts);
 			else if (tbf == tbfs->nacc) {
 				msg = ms_nacc_create_rlcmac_msg(tbf->ms(), tbf, fn, ts);
 			}
@@ -224,34 +196,34 @@ static struct msgb *sched_select_ctrl_msg(
 		 */
 		if (tbfs->dl_ass) {
 			tbf = tbfs->dl_ass;
-			msg = tbfs->dl_ass->create_dl_ass(fn, ts);
+			msg = tbf_dl_ass_create_rlcmac_msg(tbfs->dl_ass, fn, ts);
 		} else if (tbfs->ul_ass) {
 			tbf = tbfs->ul_ass;
-			msg = tbfs->ul_ass->create_ul_ass(fn, ts);
+			msg = tbf_ul_ass_create_rlcmac_msg(tbfs->ul_ass, fn, ts);
 		}
 	}
 
 	/* any message */
 	if (msg) {
 		if (!tbf) {
-			LOGP(DRLCMACSCHED, LOGL_ERROR,
-			     "Control message to be scheduled, but no TBF (TRX=%d, TS=%d)\n", trx, ts);
+			LOGPDCH(pdch, DRLCMACSCHED, LOGL_ERROR, "FN=%" PRIu32
+				" Control message to be scheduled, but no TBF\n", fn);
 			msgb_free(msg);
 			return NULL;
 		}
 		tbf->rotate_in_list();
-		LOGP(DRLCMACSCHED, LOGL_DEBUG, "Scheduling control "
-			"message at RTS for %s (TRX=%d, TS=%d)\n",
-			tbf_name(tbf), trx, ts);
-		rate_ctr_inc(&tbf->ms()->ctrs->ctr[MS_CTR_DL_CTRL_MSG_SCHED]);
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "FN=%" PRIu32
+			" Scheduling control message at RTS for %s\n",
+			fn, tbf_name(tbf));
+		rate_ctr_inc(rate_ctr_group_get_ctr(tbf->ms()->ctrs, MS_CTR_DL_CTRL_MSG_SCHED));
 		return msg;
 	}
 
 	/* schedule PACKET PAGING REQUEST, if any are pending */
 	msg = pdch->packet_paging_request();
 	if (msg) {
-		LOGP(DRLCMACSCHED, LOGL_DEBUG, "Scheduling paging request "
-			"message at RTS for (TRX=%d, TS=%d)\n", trx, ts);
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "FN=%" PRIu32
+			" Scheduling paging request message at RTS\n", fn);
 		return msg;
 	}
 
@@ -267,13 +239,13 @@ static inline enum tbf_dl_prio tbf_compute_priority(const struct gprs_rlcmac_bts
 	int age_thresh1 = msecs_to_frames(200);
 	int age_thresh2 = msecs_to_frames(OSMO_MIN(msecs_t3190/2, dl_tbf_idle_msec));
 
-	if (tbf->is_control_ts(ts) && tbf->need_control_ts())
+	if (tbf->is_control_ts(ts) && tbf->need_poll_for_dl_ack_nack())
 		return DL_PRIO_CONTROL;
 
 	if (tbf->is_control_ts(ts) && age > age_thresh2 && age_thresh2 > 0)
 		return DL_PRIO_HIGH_AGE;
 
-	if ((tbf->state_is(GPRS_RLCMAC_FLOW) && tbf->have_data()) || w->resend_needed() >= 0)
+	if ((tbf->state_is(TBF_ST_FLOW) && tbf->have_data()) || w->resend_needed() >= 0)
 		return DL_PRIO_NEW_DATA;
 
 	if (tbf->is_control_ts(ts) && age > age_thresh1 && tbf->keep_open(fn))
@@ -310,13 +282,14 @@ static bool can_produce_gmsk_data_block_next(struct gprs_rlcmac_dl_tbf *tbf, enu
 	}
 }
 
-static struct msgb *sched_select_downlink(struct gprs_rlcmac_bts *bts,
-		    uint8_t trx, uint8_t ts, uint32_t fn,
-		    uint8_t block_nr, struct gprs_rlcmac_pdch *pdch, enum mcs_kind req_mcs_kind, bool *is_egprs)
+static struct msgb *sched_select_dl_data_msg(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_pdch *pdch,
+					     uint32_t fn, uint8_t block_nr, enum mcs_kind req_mcs_kind,
+					     bool *is_egprs)
 {
 	struct msgb *msg = NULL;
 	struct gprs_rlcmac_dl_tbf *tbf, *prio_tbf = NULL;
 	enum tbf_dl_prio prio, max_prio = DL_PRIO_NONE;
+	uint8_t ts = pdch->ts_no;
 
 	uint8_t i, tfi, prio_tfi;
 	int age;
@@ -332,12 +305,8 @@ static struct msgb *sched_select_downlink(struct gprs_rlcmac_bts *bts,
 		if (tbf->direction != GPRS_RLCMAC_DL_TBF)
 			continue;
 		/* no DL resources needed, go next */
-		if (tbf->state_is_not(GPRS_RLCMAC_FLOW)
-		 && tbf->state_is_not(GPRS_RLCMAC_FINISHED))
-			continue;
-
-		/* waiting for CCCH IMM.ASS confirm */
-		if (tbf->m_wait_confirm)
+		if (tbf->state_is_not(TBF_ST_FLOW)
+		 && tbf->state_is_not(TBF_ST_FINISHED))
 			continue;
 
 		/* If a GPRS (CS1-4) Dl block is required, skip EGPRS(_GSMK) tbfs: */
@@ -355,8 +324,9 @@ static struct msgb *sched_select_downlink(struct gprs_rlcmac_bts *bts,
 		 * below instead of skipping. However, downgrade can only be
 		 * done on new data BSNs (not yet sent) and control blocks. */
 		if (req_mcs_kind == EGPRS_GMSK && !can_produce_gmsk_data_block_next(tbf, prio)) {
-			LOGP(DRLCMACSCHED, LOGL_DEBUG, "%s Cannot downgrade EGPRS TBF with prio %d\n",
-			     tbf_name(tbf), prio);
+			LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "FN=%" PRIu32
+				" Cannot downgrade EGPRS TBF with prio %d for %s\n",
+				fn, prio, tbf_name(tbf));
 			continue;
 		}
 
@@ -369,9 +339,9 @@ static struct msgb *sched_select_downlink(struct gprs_rlcmac_bts *bts,
 	}
 
 	if (prio_tbf) {
-		LOGP(DRLCMACSCHED, LOGL_DEBUG, "Scheduling data message at "
-			"RTS for DL TFI=%d (TRX=%d, TS=%d) prio=%d mcs_mode_restrict=%s\n",
-			prio_tfi, trx, ts, max_prio, mode_name(req_mcs_kind));
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "FN=%" PRIu32
+			" Scheduling data message at RTS for DL TFI=%d prio=%d mcs_mode_restrict=%s\n",
+			fn, prio_tfi, max_prio, mode_name(req_mcs_kind));
 		/* next TBF to handle resource is the next one */
 		pdch->next_dl_tfi = (prio_tfi + 1) & 31;
 		/* generate DL data block */
@@ -436,11 +406,13 @@ int gprs_rlcmac_rcv_rts_block(struct gprs_rlcmac_bts *bts,
 {
 	struct gprs_rlcmac_pdch *pdch;
 	struct tbf_sched_candidates tbf_cand = {0};
-	struct gprs_rlcmac_ul_tbf *usf_tbf;
+	struct gprs_rlcmac_tbf *poll_tbf;
+	struct gprs_rlcmac_ul_tbf *usf_tbf = NULL;
+	struct gprs_rlcmac_sba *sba;
 	uint8_t usf;
 	struct msgb *msg = NULL;
-	uint32_t poll_fn, sba_fn;
-	enum pcu_gsmtap_category gsmtap_cat;
+	uint32_t poll_fn;
+	enum pcu_gsmtap_category gsmtap_cat = PCU_GSMTAP_C_DL_DUMMY; /* init: make gcc happy */
 	bool tx_is_egprs = false;
 	bool require_gprs_only;
 	enum mcs_kind req_mcs_kind; /* Restrict CS/MCS if DL Data block is to be sent */
@@ -450,8 +422,7 @@ int gprs_rlcmac_rcv_rts_block(struct gprs_rlcmac_bts *bts,
 	pdch = &bts->trx[trx].pdch[ts];
 
 	if (!pdch->is_enabled()) {
-		LOGP(DRLCMACSCHED, LOGL_ERROR, "Received RTS on disabled PDCH: "
-			"TRX=%d TS=%d\n", trx, ts);
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_INFO, "Received RTS on disabled TS\n");
 		return -EIO;
 	}
 
@@ -473,52 +444,73 @@ int gprs_rlcmac_rcv_rts_block(struct gprs_rlcmac_bts *bts,
 		req_mcs_kind = EGPRS; /* all kinds are fine */
 	}
 
-	poll_fn = sched_poll(bts, trx, ts, fn, block_nr, &tbf_cand);
-	/* check uplink resource for polling */
-	if (tbf_cand.poll) {
-		LOGP(DRLCMACSCHED, LOGL_DEBUG, "Received RTS for PDCH: TRX=%d "
-			"TS=%d FN=%d block_nr=%d scheduling free USF for "
-			"polling at FN=%d of %s\n", trx, ts, fn,
-			block_nr, poll_fn, tbf_name(tbf_cand.poll));
-		usf = USF_UNUSED;
-	/* else. check for sba */
-	} else if ((sba_fn = bts_sba(bts)->sched(trx, ts, fn, block_nr)) != 0xffffffff) {
-		LOGP(DRLCMACSCHED, LOGL_DEBUG, "Received RTS for PDCH: TRX=%d "
-			"TS=%d FN=%d block_nr=%d scheduling free USF for "
-			"single block allocation at FN=%d\n", trx, ts, fn,
-			block_nr, sba_fn);
-		usf = USF_UNUSED;
-	/* else, we search for uplink resource */
-	} else {
-		usf_tbf = sched_select_uplink(trx, ts, fn, block_nr, pdch, require_gprs_only);
-		if (usf_tbf) {
-			usf = usf_tbf->m_usf[ts];
-			/* If MS selected for USF is GPRS-only, then it will
-			 * only be able to read USF if dl block uses GMSK
-			 * (CS1-4, MCS1-4)
-			 */
-			if (req_mcs_kind == EGPRS && ms_mode(usf_tbf->ms()) != EGPRS)
-				req_mcs_kind = EGPRS_GMSK;
-		} else {
-			usf = USF_UNUSED;
-		}
+	/* polling for next uplink block */
+	poll_fn = rts_next_fn(fn, block_nr);
+	/* check for sba */
+	if ((sba = pdch_ulc_get_sba(pdch->ulc, poll_fn))) {
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "Received RTS for PDCH: "
+			"FN=%d block_nr=%d scheduling free USF for "
+			"single block allocation at FN=%d\n", fn, block_nr, sba->fn);
+	/* else, check uplink resource for polling */
+	} else if ((poll_tbf = pdch_ulc_get_tbf_poll(pdch->ulc, poll_fn))) {
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "Received RTS for PDCH: FN=%d "
+			"block_nr=%d scheduling free USF for polling at FN=%d of %s\n",
+			fn, block_nr, poll_fn, tbf_name(poll_tbf));
+		/* If POLL TBF is UL and already has a USF assigned on this TS,
+		 * let's set its USF in the DL msg. This is not really needed,
+		 * but it helps understand better the flow when looking at
+		 * pcaps. */
+		if (poll_tbf->direction == GPRS_RLCMAC_UL_TBF && as_ul_tbf(poll_tbf)->m_usf[ts] != USF_INVALID)
+			usf_tbf = as_ul_tbf(poll_tbf);
+	/* else, search for uplink tbf */
+	} else if ((usf_tbf = sched_select_uplink(pdch, require_gprs_only))) {
+		LOGPDCH(pdch, DRLCMACSCHED, LOGL_DEBUG, "Received RTS for PDCH: FN=%d "
+			"block_nr=%d scheduling USF=%d for %s, expect answer on UL FN=%d\n",
+			fn, block_nr, usf_tbf->m_usf[pdch->ts_no], tbf_name(usf_tbf), poll_fn);
+		pdch_ulc_reserve_tbf_usf(pdch->ulc, poll_fn, usf_tbf);
 	}
+	/* If MS selected for USF is GPRS-only, then it will only be
+	 * able to read USF if dl block uses GMSK (CS1-4, MCS1-4) */
+	if (usf_tbf && req_mcs_kind == EGPRS && ms_mode(usf_tbf->ms()) != EGPRS)
+		req_mcs_kind = EGPRS_GMSK;
+
+	get_ctrl_msg_tbf_candidates(pdch, &tbf_cand);
 
 	/* Prio 1: select control message */
-	if ((msg = sched_select_ctrl_msg(trx, ts, fn, block_nr, pdch, &tbf_cand))) {
+	if ((msg = sched_select_ctrl_msg(pdch, fn, block_nr, &tbf_cand))) {
 			gsmtap_cat = PCU_GSMTAP_C_DL_CTRL;
 	}
 	/* Prio 2: select data message for downlink */
-	else if((msg = sched_select_downlink(bts, trx, ts, fn, block_nr, pdch, req_mcs_kind, &tx_is_egprs))) {
+	else if((msg = sched_select_dl_data_msg(bts, pdch, fn, block_nr, req_mcs_kind, &tx_is_egprs))) {
 		gsmtap_cat = tx_is_egprs ? PCU_GSMTAP_C_DL_DATA_EGPRS :
 					   PCU_GSMTAP_C_DL_DATA_GPRS;
 	}
-	/* Prio 3: send dummy contol message */
-	else if ((msg = sched_dummy())) {
-		/* increase counter */
-		gsmtap_cat = PCU_GSMTAP_C_DL_DUMMY;
-	} else {
-		return -ENOMEM;
+	/* Prio 3: send dummy control message if need to poll or USF */
+	else {
+		/* If there's no TBF attached to this PDCH, we can submit an empty
+		 * data_req since there's nothing to transmit nor to poll/USF. This
+		 * way we help BTS energy saving (on TRX!=C0) by sending nothing
+		 * instead of a dummy block. The early return is done here and
+		 * not at the start of the function because the condition below
+		 * (num_tbfs==0) may not be enough, because temporary dummy TBFs
+		 * created to send Imm Ass Rej (see handle_tbf_reject()) don't
+		 * have a TFI assigned and hence are not attached to the PDCH
+		 * TS, so they don't show up in the count below.
+		 */
+		const unsigned num_tbfs = pdch->num_tbfs(GPRS_RLCMAC_DL_TBF)
+					+ pdch->num_tbfs(GPRS_RLCMAC_UL_TBF);
+		bool skip_idle = (num_tbfs == 0);
+#ifdef ENABLE_DIRECT_PHY
+		/* In DIRECT_PHY mode we want to always submit something to L1 in
+		 * TRX0, since BTS is not preparing dummy bursts on idle TS for us */
+		skip_idle = skip_idle && trx != 0;
+#endif
+		if (!skip_idle && (msg = sched_dummy())) {
+			/* increase counter */
+			gsmtap_cat = PCU_GSMTAP_C_DL_DUMMY;
+		} else {
+			msg = NULL; /* submit empty frame */
+		}
 	}
 
 	if (tx_is_egprs && pdch->has_gprs_only_tbf_attached()) {
@@ -527,20 +519,23 @@ int gprs_rlcmac_rcv_rts_block(struct gprs_rlcmac_bts *bts,
 		pdch->fn_without_cs14 = 0;
 	}
 
-	/* msg is now available */
-	bts_do_rate_ctr_add(bts, CTR_RLC_DL_BYTES, msg->data_len);
-
-	/* set USF */
-	OSMO_ASSERT(msgb_length(msg) > 0);
-	msg->data[0] = (msg->data[0] & 0xf8) | usf;
-
 	/* Used to measure the leak rate, count all blocks */
 	gprs_bssgp_update_frames_sent();
 
-	/* Send to GSMTAP */
-	tap_n_acc(msg, bts, trx, ts, fn, gsmtap_cat);
+	if (msg) {
+		/* msg is now available */
+		bts_do_rate_ctr_add(bts, CTR_RLC_DL_BYTES, msgb_length(msg));
 
-	/* send PDTCH/PACCH to L1 */
+		/* set USF */
+		OSMO_ASSERT(msgb_length(msg) > 0);
+		usf = usf_tbf ? usf_tbf->m_usf[ts] : USF_UNUSED;
+		msg->data[0] = (msg->data[0] & 0xf8) | usf;
+
+		/* Send to GSMTAP */
+		tap_n_acc(msg, bts, trx, ts, fn, gsmtap_cat);
+	}
+
+	/* send PDTCH/PACCH to L1. msg=NULL accepted too (idle block) */
 	pcu_l1if_tx_pdtch(msg, bts, trx, ts, bts->trx[trx].arfcn, fn, block_nr);
 
 	return 0;
