@@ -298,20 +298,21 @@ static bool idle_pdch_avail(const struct gprs_rlcmac_bts *bts)
 /*! Return free TFI
  *
  *  \param[in] bts Pointer to BTS struct
- *  \param[in] trx Optional pointer to TRX struct
  *  \param[in] ms Pointer to MS object
  *  \param[in] dir DL or UL direction
  *  \param[in] use_trx which TRX to use or -1 if it should be selected based on what MS uses
  *  \param[out] trx_no_ TRX number on which TFI was found
  *  \returns negative error code or 0 on success
  */
-static int tfi_find_free(const struct gprs_rlcmac_bts *bts, const gprs_rlcmac_trx *trx, const GprsMs *ms,
+static int tfi_find_free(const struct gprs_rlcmac_bts *bts, const GprsMs *ms,
 			 enum gprs_rlcmac_tbf_direction dir, int8_t use_trx, uint8_t *trx_no_)
 {
+	const struct gprs_rlcmac_trx *trx;
 	int tfi;
 	uint8_t trx_no;
 
-	if (trx) {
+	/* If MS is already doing stuff on a TRX, set use_trx to it: */
+	if ((trx = ms_current_trx(ms))) {
 		if (use_trx >= 0 && use_trx != trx->trx_no) {
 			LOGP(DRLCMAC, LOGL_ERROR, "- Requested incompatible TRX %d (current is %d)\n",
 			     use_trx, trx->trx_no);
@@ -320,9 +321,6 @@ static int tfi_find_free(const struct gprs_rlcmac_bts *bts, const gprs_rlcmac_tr
 		use_trx = trx->trx_no;
 	}
 
-	if (use_trx == -1 && ms_current_trx(ms))
-		use_trx = ms_current_trx(ms)->trx_no;
-
 	tfi = bts_tfi_find_free(bts, dir, &trx_no, use_trx);
 	if (tfi < 0)
 		return -EBUSY;
@@ -338,13 +336,12 @@ static int tfi_find_free(const struct gprs_rlcmac_bts *bts, const gprs_rlcmac_tr
  * Assign single slot for uplink and downlink
  *
  *  \param[in,out] bts Pointer to BTS struct
- *  \param[in,out] ms_ Pointer to MS object
- *  \param[in,out] tbf_ Pointer to TBF struct
+ *  \param[in,out] tbf Pointer to TBF struct
  *  \param[in] single flag indicating if we should force single-slot allocation
  *  \param[in] use_trx which TRX to use or -1 if it should be selected during allocation
  *  \returns negative error code or 0 on success
  */
-int alloc_algorithm_a(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcmac_tbf *tbf_, bool single,
+int alloc_algorithm_a(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf, bool single,
 		      int8_t use_trx)
 {
 	struct gprs_rlcmac_pdch *pdch;
@@ -355,8 +352,7 @@ int alloc_algorithm_a(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcm
 	int usf = -1;
 	uint8_t mask = 0xff;
 	const char *mask_reason = NULL;
-	const GprsMs *ms = ms_;
-	const gprs_rlcmac_tbf *tbf = tbf_;
+	struct GprsMs *ms = tbf->ms();
 	gprs_rlcmac_trx *trx = ms_current_trx(ms);
 
 	LOGPAL(tbf, "A", single, use_trx, LOGL_DEBUG, "Alloc start\n");
@@ -405,24 +401,24 @@ int alloc_algorithm_a(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcm
 
 	pdch = &trx->pdch[ts];
 
-	/* The allocation will be successful, so the system state and tbf_/ms_
+	/* The allocation will be successful, so the system state and tbf/ms
 	 * may be modified from now on. */
 	if (tbf->direction == GPRS_RLCMAC_UL_TBF) {
-		struct gprs_rlcmac_ul_tbf *ul_tbf = as_ul_tbf(tbf_);
+		struct gprs_rlcmac_ul_tbf *ul_tbf = as_ul_tbf(tbf);
 		LOGPSL(tbf, LOGL_DEBUG, "Assign uplink TS=%d TFI=%d USF=%d\n", ts, tfi, usf);
 		assign_uplink_tbf_usf(pdch, ul_tbf, tfi, usf);
 	} else {
-		struct gprs_rlcmac_dl_tbf *dl_tbf = as_dl_tbf(tbf_);
+		struct gprs_rlcmac_dl_tbf *dl_tbf = as_dl_tbf(tbf);
 		LOGPSL(tbf, LOGL_DEBUG, "Assign downlink TS=%d TFI=%d\n", ts, tfi);
 		assign_dlink_tbf(pdch, dl_tbf, tfi);
 	}
 
-	tbf_->trx = trx;
+	tbf->trx = trx;
 	/* the only one TS is the common TS */
-	tbf_->first_ts = tbf_->first_common_ts = ts;
-	ms_set_reserved_slots(ms_, trx, 1 << ts, 1 << ts);
+	tbf->first_ts = tbf->first_common_ts = ts;
+	ms_set_reserved_slots(ms, trx, 1 << ts, 1 << ts);
 
-	tbf_->upgrade_to_multislot = 0;
+	tbf->upgrade_to_multislot = false;
 	bts_do_rate_ctr_inc(bts, CTR_TBF_ALLOC_ALGO_A);
 	return 0;
 }
@@ -648,6 +644,7 @@ int find_multi_slots(struct gprs_rlcmac_trx *trx, uint8_t mslot_class, uint8_t *
 	if (!max_ul_slots || !max_dl_slots) {
 		LOGP(DRLCMAC, LOGL_NOTICE,
 			"No valid UL/DL slot combination found\n");
+		bts_do_rate_ctr_inc(trx->bts, CTR_TBF_ALLOC_FAIL_NO_SLOT_COMBI);
 		return -EINVAL;
 	}
 
@@ -662,12 +659,12 @@ int find_multi_slots(struct gprs_rlcmac_trx *trx, uint8_t mslot_class, uint8_t *
  *  \param[in] slots Timeslots in use
  *  \param[in] reserved_slots Reserved timeslots
  *  \param[out] slotcount Number of TS in use
- *  \param[out] avail_count Number of reserved TS
+ *  \param[out] reserve_count Number of reserved TS
  */
-static void update_slot_counters(uint8_t slots, uint8_t reserved_slots, uint8_t *slotcount, uint8_t *avail_count)
+static void count_slots(uint8_t slots, uint8_t reserved_slots, uint8_t *slotcount, uint8_t *reserve_count)
 {
 	(*slotcount) = pcu_bitcount(slots);
-	(*avail_count) = pcu_bitcount(reserved_slots);
+	(*reserve_count) = pcu_bitcount(reserved_slots);
 }
 
 /*! Return slot mask with single TS from a given UL/DL set according to TBF's direction, ts pointer is set to that TS
@@ -711,7 +708,8 @@ static int tbf_select_slot_set(const gprs_rlcmac_tbf *tbf, const gprs_rlcmac_trx
 			       uint8_t reserved_ul_slots, uint8_t reserved_dl_slots,
 			       int8_t first_common_ts)
 {
-	uint8_t sl = tbf->direction != GPRS_RLCMAC_DL_TBF ? ul_slots : dl_slots;
+	bool is_ul = tbf->direction == GPRS_RLCMAC_UL_TBF;
+	uint8_t sl = is_ul ? ul_slots : dl_slots;
 	char slot_info[9] = { 0 };
 
 	if (single)
@@ -719,59 +717,63 @@ static int tbf_select_slot_set(const gprs_rlcmac_tbf *tbf, const gprs_rlcmac_trx
 
 	if (!sl) {
 		LOGP(DRLCMAC, LOGL_NOTICE, "No %s slots available\n",
-		     tbf->direction != GPRS_RLCMAC_DL_TBF ? "uplink" : "downlink");
+		     is_ul ? "uplink" : "downlink");
+		bts_do_rate_ctr_inc(trx->bts, CTR_TBF_ALLOC_FAIL_NO_SLOT_AVAIL);
 		return -EINVAL;
 	}
 
-	if (tbf->direction != GPRS_RLCMAC_DL_TBF) {
+	if (is_ul) {
 		snprintf(slot_info, 9, OSMO_BIT_SPEC, OSMO_BIT_PRINT_EX(reserved_ul_slots, 'u'));
 		masked_override_with(slot_info, sl, 'U');
-		LOGP(DRLCMAC, LOGL_DEBUG, "- Selected UL");
 	} else {
 		snprintf(slot_info, 9, OSMO_BIT_SPEC, OSMO_BIT_PRINT_EX(reserved_dl_slots, 'd'));
 		masked_override_with(slot_info, sl, 'D');
-		LOGP(DRLCMAC, LOGL_DEBUG, "- Selected DL");
 	}
 
-	LOGPC(DRLCMAC, LOGL_DEBUG, " slots: (TS=0)\"%s\"(TS=7), %s\n", slot_info, single ? "single" : "multi");
+	LOGPC(DRLCMAC, LOGL_DEBUG, "Selected %s slots: (TS=0)\"%s\"(TS=7), %s\n",
+	      is_ul ? "UL" : "DL",
+	      slot_info, single ? "single" : "multi");
 
 	return sl;
 }
 
 /*! Allocate USF according to a given UL TS mapping
  *
- * N. B: this is legacy implementation which ignores given selected_ul_slots
  *  \param[in] trx Pointer to TRX object
- *  \param[in] tbf Pointer to TBF object
- *  \param[in] first_common_ts First TS which is common to both UL and DL
  *  \param[in] selected_ul_slots set of UL timeslots selected for allocation
  *  \param[in] dl_slots set of DL timeslots
  *  \param[out] usf array for allocated USF
- *  \returns updated UL TS or negative on error
+ *  \returns updated UL TS mask or negative on error
  */
-static int allocate_usf(const gprs_rlcmac_trx *trx, int8_t first_common_ts, uint8_t selected_ul_slots, uint8_t dl_slots,
-			int *usf)
+static int allocate_usf(const gprs_rlcmac_trx *trx, uint8_t selected_ul_slots, uint8_t dl_slots,
+			int *usf_list)
 {
-	int free_usf = -1, ts;
-	uint8_t ul_slots = selected_ul_slots;
+	uint8_t ul_slots = selected_ul_slots & dl_slots;
+	unsigned int ts;
 
-	if (first_common_ts >= 0)
-		ul_slots = 1 << first_common_ts;
-	else
-		ul_slots = ul_slots & dl_slots;
+	for (ts = 0; ts < ARRAY_SIZE(trx->pdch); ts++) {
+		const struct gprs_rlcmac_pdch *pdch = &trx->pdch[ts];
+		int8_t free_usf;
 
-	ts = find_least_busy_pdch(trx, GPRS_RLCMAC_UL_TBF, ul_slots, compute_usage_by_num_tbfs, NULL, &free_usf);
+		if (((1 << ts) & ul_slots) == 0)
+			continue;
 
-	if (free_usf < 0 || ts < 0) {
-		LOGP(DRLCMAC, LOGL_NOTICE, "No USF available\n");
-		return -EBUSY;
+		free_usf = find_free_usf(pdch->assigned_usf());
+		if (free_usf < 0) {
+			LOGP(DRLCMAC, LOGL_DEBUG,
+				"- Skipping TS %d, because "
+				"no USF available\n", ts);
+			ul_slots &= (~(1 << ts)) & 0xff;
+			continue;
+		}
+		usf_list[ts] = free_usf;
 	}
 
-	OSMO_ASSERT(ts >= 0 && ts <= 8);
-
-	/* We will stick to that single UL slot, unreserve the others */
-	ul_slots = 1 << ts;
-	usf[ts] = free_usf;
+	if (!ul_slots) {
+		LOGP(DRLCMAC, LOGL_NOTICE, "No USF available\n");
+		bts_do_rate_ctr_inc(trx->bts, CTR_TBF_ALLOC_FAIL_NO_USF);
+		return -EBUSY;
+	}
 
 	return ul_slots;
 }
@@ -851,13 +853,12 @@ static void assign_dl_tbf_slots(struct gprs_rlcmac_dl_tbf *dl_tbf, gprs_rlcmac_t
  * Assign one uplink slot. (With free USF)
  *
  *  \param[in,out] bts Pointer to BTS struct
- *  \param[in,out] ms_ Pointer to MS object
- *  \param[in,out] tbf_ Pointer to TBF struct
+ *  \param[in,out] tbf Pointer to TBF struct
  *  \param[in] single flag indicating if we should force single-slot allocation
  *  \param[in] use_trx which TRX to use or -1 if it should be selected during allocation
  *  \returns negative error code or 0 on success
  */
-int alloc_algorithm_b(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcmac_tbf *tbf_, bool single,
+int alloc_algorithm_b(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf, bool single,
 		      int8_t use_trx)
 {
 	uint8_t dl_slots;
@@ -866,48 +867,39 @@ int alloc_algorithm_b(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcm
 	uint8_t reserved_ul_slots;
 	int8_t first_common_ts;
 	uint8_t slotcount = 0;
-	uint8_t avail_count = 0, trx_no;
+	uint8_t reserve_count = 0, trx_no;
 	int first_ts = -1;
 	int usf[8] = {-1, -1, -1, -1, -1, -1, -1, -1};
 	int rc;
 	int tfi;
-	const GprsMs *ms = ms_;
-	const gprs_rlcmac_tbf *tbf = tbf_;
+	struct GprsMs *ms = tbf->ms();
 	gprs_rlcmac_trx *trx;
 
 	LOGPAL(tbf, "B", single, use_trx, LOGL_DEBUG, "Alloc start\n");
 
 	/* Step 1: Get current state from the MS object */
 
-	if (!ms) {
-		LOGP(DRLCMAC, LOGL_ERROR, "MS not set\n");
-		return -EINVAL;
-	}
-
-	dl_slots = ms_reserved_dl_slots(ms);
-	ul_slots = ms_reserved_ul_slots(ms);
+	reserved_dl_slots = ms_reserved_dl_slots(ms);
+	reserved_ul_slots = ms_reserved_ul_slots(ms);
 	first_common_ts = ms_first_common_ts(ms);
-	trx = ms_current_trx(ms);
 
 	/* Step 2a: Find usable TRX and TFI */
-	tfi = tfi_find_free(bts, trx, ms, tbf->direction, use_trx, &trx_no);
+	tfi = tfi_find_free(bts, ms, tbf->direction, use_trx, &trx_no);
 	if (tfi < 0) {
 		LOGPAL(tbf, "B", single, use_trx, LOGL_NOTICE, "failed to allocate a TFI\n");
 		return tfi;
 	}
 
 	/* Step 2b: Reserve slots on the TRX for the MS */
-	if (!trx)
-		trx = &bts->trx[trx_no];
+	trx = &bts->trx[trx_no];
 
-	if (!dl_slots || !ul_slots) {
-		rc = find_multi_slots(trx, ms_ms_class(ms), &ul_slots, &dl_slots);
+	if (!reserved_dl_slots || !reserved_ul_slots) {
+		rc = find_multi_slots(trx, ms_ms_class(ms), &reserved_ul_slots, &reserved_dl_slots);
 		if (rc < 0)
 			return rc;
 	}
-
-	reserved_dl_slots = dl_slots;
-	reserved_ul_slots = ul_slots;
+	dl_slots = reserved_dl_slots;
+	ul_slots = reserved_ul_slots;
 
 	/* Step 3a: Derive the slot set for the current TBF */
 	rc = tbf_select_slot_set(tbf, trx, single, ul_slots, dl_slots, reserved_ul_slots, reserved_dl_slots,
@@ -918,17 +910,16 @@ int alloc_algorithm_b(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcm
 	/* Step 3b: Derive the slot set for a given direction */
 	if (tbf->direction == GPRS_RLCMAC_DL_TBF) {
 		dl_slots = rc;
-		update_slot_counters(dl_slots, reserved_dl_slots, &slotcount, &avail_count);
+		count_slots(dl_slots, reserved_dl_slots, &slotcount, &reserve_count);
 	} else {
-		rc = allocate_usf(trx, first_common_ts, rc, dl_slots, usf);
+		rc = allocate_usf(trx, rc, dl_slots, usf);
 		if (rc < 0)
 			return rc;
 
-		/* We will stick to that single UL slot, unreserve the others */
 		ul_slots = rc;
 		reserved_ul_slots = ul_slots;
 
-		update_slot_counters(ul_slots, reserved_ul_slots, &slotcount, &avail_count);
+		count_slots(ul_slots, reserved_ul_slots, &slotcount, &reserve_count);
 	}
 
 	first_ts = ffs(rc) - 1;
@@ -945,28 +936,28 @@ int alloc_algorithm_b(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcm
 	}
 
 	if (single && slotcount) {
-		tbf_->upgrade_to_multislot = (avail_count > slotcount);
+		tbf->upgrade_to_multislot = (reserve_count > slotcount);
 		LOGPAL(tbf, "B", single, use_trx, LOGL_INFO, "using single slot at TS %d\n", first_ts);
 	} else {
-		tbf_->upgrade_to_multislot = 0;
+		tbf->upgrade_to_multislot = false;
 		LOGPAL(tbf, "B", single, use_trx, LOGL_INFO, "using %d slots\n", slotcount);
 	}
 
-	/* The allocation will be successful, so the system state and tbf_/ms_
+	/* The allocation will be successful, so the system state and tbf/ms
 	 * may be modified from now on. */
 
 	/* Step 4: Update MS and TBF and really allocate the resources */
 
-	update_ms_reserved_slots(trx, ms_, reserved_ul_slots, reserved_dl_slots, ul_slots, dl_slots);
+	update_ms_reserved_slots(trx, ms, reserved_ul_slots, reserved_dl_slots, ul_slots, dl_slots);
 
-	tbf_->trx = trx;
-	tbf_->first_common_ts = first_common_ts;
-	tbf_->first_ts = first_ts;
+	tbf->trx = trx;
+	tbf->first_common_ts = first_common_ts;
+	tbf->first_ts = first_ts;
 
 	if (tbf->direction == GPRS_RLCMAC_DL_TBF)
-		assign_dl_tbf_slots(as_dl_tbf(tbf_), trx, dl_slots, tfi);
+		assign_dl_tbf_slots(as_dl_tbf(tbf), trx, dl_slots, tfi);
 	else
-		assign_ul_tbf_slots(as_ul_tbf(tbf_), trx, ul_slots, tfi, usf);
+		assign_ul_tbf_slots(as_ul_tbf(tbf), trx, ul_slots, tfi, usf);
 
 	bts_do_rate_ctr_inc(bts, CTR_TBF_ALLOC_ALGO_B);
 
@@ -982,13 +973,12 @@ int alloc_algorithm_b(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcm
  * goal is to provide the highest possible bandwidth per MS.
  *
  *  \param[in,out] bts Pointer to BTS struct
- *  \param[in,out] ms_ Pointer to MS object
- *  \param[in,out] tbf_ Pointer to TBF struct
+ *  \param[in,out] tbf Pointer to TBF struct
  *  \param[in] single flag indicating if we should force single-slot allocation
  *  \param[in] use_trx which TRX to use or -1 if it should be selected during allocation
  *  \returns negative error code or 0 on success
  */
-int alloc_algorithm_dynamic(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gprs_rlcmac_tbf *tbf_, bool single,
+int alloc_algorithm_dynamic(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_tbf *tbf, bool single,
 			    int8_t use_trx)
 {
 	int rc;
@@ -1001,7 +991,7 @@ int alloc_algorithm_dynamic(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gpr
 	}
 
 	if (!bts->multislot_disabled) {
-		rc = alloc_algorithm_b(bts, ms_, tbf_, single, use_trx);
+		rc = alloc_algorithm_b(bts, tbf, single, use_trx);
 		if (rc >= 0)
 			return rc;
 
@@ -1010,7 +1000,7 @@ int alloc_algorithm_dynamic(struct gprs_rlcmac_bts *bts, GprsMs *ms_, struct gpr
 		bts->multislot_disabled = 1;
 	}
 
-	return alloc_algorithm_a(bts, ms_, tbf_, single, use_trx);
+	return alloc_algorithm_a(bts, tbf, single, use_trx);
 }
 
 int gprs_alloc_max_dl_slots_per_ms(const struct gprs_rlcmac_bts *bts, uint8_t ms_class)
