@@ -44,6 +44,8 @@ extern "C" {
 #include <osmocom/gsm/protocol/gsm_04_08.h>
 #include <osmocom/gsm/gsm48_rest_octets.h>
 #include <osmocom/gsm/sysinfo.h>
+
+#include <nacc_fsm.h>
 }
 
 #include <gprs_rlcmac.h>
@@ -95,6 +97,7 @@ uint16_t imsi2paging_group(const char* imsi)
  * PCU messages
  */
 
+/* Can be used to allocate message with non-variable size */
 struct msgb *pcu_msgb_alloc(uint8_t msg_type, uint8_t bts_nr)
 {
 	struct msgb *msg;
@@ -111,6 +114,21 @@ struct msgb *pcu_msgb_alloc(uint8_t msg_type, uint8_t bts_nr)
 	return msg;
 }
 
+/* Allocate message with extra size, only reserve pcuif msg hdr */
+static struct msgb *pcu_msgb_alloc_ext_size(uint8_t msg_type, uint8_t bts_nr, size_t extra_size)
+{
+	struct msgb *msg;
+	struct gsm_pcu_if *pcu_prim;
+	msg = msgb_alloc(sizeof(struct gsm_pcu_if) + extra_size, "pcu_sock_tx");
+	/* Only header is filled, caller is responible for reserving + filling
+	 * message type specific contents: */
+	msgb_put(msg, PCUIF_HDR_SIZE);
+	pcu_prim = (struct gsm_pcu_if *) msgb_data(msg);
+	pcu_prim->msg_type = msg_type;
+	pcu_prim->bts_nr = bts_nr;
+	return msg;
+}
+
 const struct value_string gsm_pcu_if_text_type_names[] = {
 	OSMO_VALUE_STRING(PCU_VERSION),
 	OSMO_VALUE_STRING(PCU_OML_ALERT),
@@ -147,14 +165,15 @@ int pcu_tx_txt_ind(enum gsm_pcu_if_text_type t, const char *fmt, ...)
 	return pcu_sock_send(msg);
 }
 
-static int pcu_tx_act_req(struct gprs_rlcmac_bts *bts, uint8_t trx, uint8_t ts, uint8_t activate)
+static int pcu_tx_act_req(struct gprs_rlcmac_bts *bts, const struct gprs_rlcmac_pdch *pdch,
+			  uint8_t activate)
 {
 	struct msgb *msg;
 	struct gsm_pcu_if *pcu_prim;
 	struct gsm_pcu_if_act_req *act_req;
 
-	LOGP(DL1IF, LOGL_INFO, "Sending %s request: trx=%d ts=%d\n",
-		(activate) ? "activate" : "deactivate", trx, ts);
+	LOGPDCH(pdch, DL1IF, LOGL_INFO, "Sending %s request\n",
+		(activate) ? "activate" : "deactivate");
 
 	msg = pcu_msgb_alloc(PCU_IF_MSG_ACT_REQ, bts->nr);
 	if (!msg)
@@ -162,8 +181,8 @@ static int pcu_tx_act_req(struct gprs_rlcmac_bts *bts, uint8_t trx, uint8_t ts,
 	pcu_prim = (struct gsm_pcu_if *) msg->data;
 	act_req = &pcu_prim->u.act_req;
 	act_req->activate = activate;
-	act_req->trx_nr = trx;
-	act_req->ts_nr = ts;
+	act_req->trx_nr = pdch->trx_no();
+	act_req->ts_nr = pdch->ts_no;
 
 	return pcu_sock_send(msg);
 }
@@ -177,9 +196,9 @@ static int pcu_tx_data_req(struct gprs_rlcmac_bts *bts, uint8_t trx, uint8_t ts,
 	struct gsm_pcu_if_data *data_req;
 	int current_fn = bts_current_frame_number(bts);
 
-	LOGP(DL1IF, LOGL_DEBUG, "Sending data request: trx=%d ts=%d sapi=%d "
-		"arfcn=%d fn=%d cur_fn=%d block=%d data=%s\n", trx, ts, sapi, arfcn, fn, current_fn,
-		block_nr, osmo_hexdump(data, len));
+	LOGP(DL1IF, LOGL_DEBUG, "(bts=%u,trx=%u,ts=%u) FN=%u Sending data request: sapi=%d "
+	     "arfcn=%d cur_fn=%d block=%d data=%s\n", bts->nr, trx, ts, fn, sapi,
+	     arfcn, current_fn, block_nr, osmo_hexdump(data, len));
 
 	msg = pcu_msgb_alloc(PCU_IF_MSG_DATA_REQ, bts->nr);
 	if (!msg)
@@ -193,7 +212,8 @@ static int pcu_tx_data_req(struct gprs_rlcmac_bts *bts, uint8_t trx, uint8_t ts,
 	data_req->trx_nr = trx;
 	data_req->ts_nr = ts;
 	data_req->block_nr = block_nr;
-	memcpy(data_req->data, data, len);
+	if (len)
+		memcpy(data_req->data, data, len);
 	data_req->len = len;
 
 	return pcu_sock_send(msg);
@@ -204,12 +224,20 @@ void pcu_l1if_tx_pdtch(msgb *msg, struct gprs_rlcmac_bts *bts, uint8_t trx, uint
 {
 #ifdef ENABLE_DIRECT_PHY
 	if (bts->trx[trx].fl1h) {
+		if (!msg) /* Simply skip sending idle frames to L1 */
+			return;
 		l1if_pdch_req(bts->trx[trx].fl1h, ts, 0, fn, arfcn, block_nr,
 			msg->data, msg->len);
 		msgb_free(msg);
 		return;
 	}
 #endif
+	if (!msg) {
+		pcu_tx_data_req(bts, trx, ts, PCU_IF_SAPI_PDTCH, arfcn, fn, block_nr,
+				NULL, 0);
+		return;
+	}
+
 	pcu_tx_data_req(bts, trx, ts, PCU_IF_SAPI_PDTCH, arfcn, fn, block_nr,
 			msg->data, msg->len);
 	msgb_free(msg);
@@ -224,10 +252,17 @@ void pcu_l1if_tx_ptcch(struct gprs_rlcmac_bts *bts,
 		gsmtap_send(the_pcu->gsmtap, arfcn, ts, GSMTAP_CHANNEL_PTCCH, 0, fn, 0, 0, data, data_len);
 #ifdef ENABLE_DIRECT_PHY
 	if (bts->trx[trx].fl1h) {
+		if (!data_len) /* Simply skip sending idle frames to L1 */
+			return;
 		l1if_pdch_req(bts->trx[trx].fl1h, ts, 1, fn, arfcn, block_nr, data, data_len);
 		return;
 	}
 #endif
+	if (!data_len) {
+		pcu_tx_data_req(bts, trx, ts, PCU_IF_SAPI_PTCCH, arfcn, fn, block_nr, NULL, 0);
+		return;
+	}
+
 	pcu_tx_data_req(bts, trx, ts, PCU_IF_SAPI_PTCCH, arfcn, fn, block_nr, data, data_len);
 }
 
@@ -270,29 +305,75 @@ void pcu_l1if_tx_pch(struct gprs_rlcmac_bts *bts, bitvec * block, int plen, uint
 	pcu_tx_data_req(bts, 0, 0, PCU_IF_SAPI_PCH, 0, 0, 0, data, PAGING_GROUP_LEN + GSM_MACBLOCK_LEN);
 }
 
+int pcu_tx_neigh_addr_res_req(struct gprs_rlcmac_bts *bts, const struct neigh_cache_entry_key *neigh_key)
+{
+	struct msgb *msg;
+	struct gsm_pcu_if *pcu_prim;
+	struct gsm_pcu_if_neigh_addr_req *naddr_req;
+
+	LOGP(DL1IF, LOGL_DEBUG, "(bts=%u) Tx Neighbor Address Resolution Request: " NEIGH_CACHE_ENTRY_KEY_FMT "\n",
+	     bts->nr, NEIGH_CACHE_ENTRY_KEY_ARGS(neigh_key));
+
+	msg = pcu_msgb_alloc_ext_size(PCU_IF_MSG_CONTAINER, bts->nr, sizeof(struct gsm_pcu_if_neigh_addr_req));
+	if (!msg)
+		return -ENOMEM;
+	pcu_prim = (struct gsm_pcu_if *) msgb_data(msg);
+	naddr_req = (struct gsm_pcu_if_neigh_addr_req *)&pcu_prim->u.container.data[0];
+
+	msgb_put(msg, sizeof(pcu_prim->u.container) + sizeof(struct gsm_pcu_if_neigh_addr_req));
+	pcu_prim->u.container.msg_type = PCU_IF_MSG_NEIGH_ADDR_REQ;
+	osmo_store16be(sizeof(struct gsm_pcu_if_neigh_addr_req), &pcu_prim->u.container.length);
+
+	osmo_store16be(neigh_key->local_lac, &naddr_req->local_lac);
+	osmo_store16be(neigh_key->local_ci, &naddr_req->local_ci);
+	osmo_store16be(neigh_key->tgt_arfcn, &naddr_req->tgt_arfcn);
+	naddr_req->tgt_bsic = neigh_key->tgt_bsic;
+
+	return pcu_sock_send(msg);
+}
+
 void pcu_rx_block_time(struct gprs_rlcmac_bts *bts, uint16_t arfcn, uint32_t fn, uint8_t ts_no)
 {
-	bts_set_current_block_frame_number(bts, fn, 0);
+	bts_set_current_block_frame_number(bts, fn);
 }
 
-void pcu_rx_ra_time(struct gprs_rlcmac_bts *bts, uint16_t arfcn, uint32_t fn, uint8_t ts_no)
-{
-	/* access bursts may arrive some bursts earlier */
-	bts_set_current_block_frame_number(bts, fn, 5);
-}
-
-int pcu_rx_data_ind_pdtch(struct gprs_rlcmac_bts *bts, uint8_t trx_no, uint8_t ts_no, uint8_t *data,
+int pcu_rx_data_ind_pdtch(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_pdch *pdch, uint8_t *data,
 	uint8_t len, uint32_t fn, struct pcu_l1_meas *meas)
 {
-	struct gprs_rlcmac_pdch *pdch;
+	int rc;
 
-	pdch = &bts->trx[trx_no].pdch[ts_no];
-	return pdch->rcv_block(data, len, fn, meas);
+	if (!pdch->is_enabled()) {
+		LOGPDCH(pdch, DL1IF, LOGL_INFO, "Received DATA.ind (PDTCH) on disabled TS\n");
+		return -EINVAL;
+	}
+
+	rc = pdch->rcv_block(data, len, fn, meas);
+	pdch_ulc_expire_fn(pdch->ulc, fn);
+	return rc;
+}
+
+static int list_arfcn(const struct gprs_rlcmac_bts *bts, const struct gsm_sysinfo_freq *freq, const char *text)
+{
+	int n = 0, i;
+	for (i = 0; i < 1024; i++) {
+		if (freq[i].mask) {
+			if (!n)
+				LOGP(DL1IF, LOGL_INFO, "BTS%d: %s", bts->nr, text);
+			LOGPC(DL1IF, LOGL_INFO, " %d", i);
+			n++;
+		}
+	}
+	if (n)
+		LOGPC(DL1IF, LOGL_INFO, "\n");
+
+	return n;
 }
 
 static int pcu_rx_data_ind_bcch(struct gprs_rlcmac_bts *bts, uint8_t *data, uint8_t len)
 {
+	struct gsm48_system_information_type_2 *si2;
 	const uint8_t *si_ro;
+
 	switch (len) {
 	case 0:
 		/* Due to historical reasons also accept a completely empty message as
@@ -309,6 +390,9 @@ static int pcu_rx_data_ind_bcch(struct gprs_rlcmac_bts *bts, uint8_t *data, uint
 		case SYSINFO_TYPE_1:
 			bts->si1_is_set = false;
 			break;
+		case SYSINFO_TYPE_2:
+			bts->si2_is_set = false;
+			break;
 		case SYSINFO_TYPE_3:
 			bts->si3_is_set = false;
 			break;
@@ -332,6 +416,14 @@ static int pcu_rx_data_ind_bcch(struct gprs_rlcmac_bts *bts, uint8_t *data, uint
 			memcpy(bts->si1, data, GSM_MACBLOCK_LEN);
 			bts->si1_is_set = true;
 			break;
+		case GSM48_MT_RR_SYSINFO_2:
+			memcpy(bts->si2, data, GSM_MACBLOCK_LEN);
+			bts->si2_is_set = true;
+			si2 = (struct gsm48_system_information_type_2 *)bts->si2;
+			gsm48_decode_freq_list(bts->si2_bcch_cell_list, si2->bcch_frequency_list,
+					       sizeof(si2->bcch_frequency_list), 0xce, 1);
+			list_arfcn(bts, bts->si2_bcch_cell_list, "SI2 Neighbour cells in same band:");
+			break;
 		case GSM48_MT_RR_SYSINFO_3:
 			memcpy(bts->si3, data, GSM_MACBLOCK_LEN);
 			bts->si3_is_set = true;
@@ -340,7 +432,7 @@ static int pcu_rx_data_ind_bcch(struct gprs_rlcmac_bts *bts, uint8_t *data, uint
 			memcpy(bts->si13, data, GSM_MACBLOCK_LEN);
 			bts->si13_is_set = true;
 			si_ro = ((struct gsm48_system_information_type_13*)data)->rest_octets;
-			if (osmo_gsm48_rest_octets_si13_decode(&bts->si31_ro_decoded, si_ro) < 0)
+			if (osmo_gsm48_rest_octets_si13_decode(&bts->si13_ro_decoded, si_ro) < 0)
 				LOGP(DPCU, LOGL_ERROR, "Error decoding SI13\n");
 			break;
 		default:
@@ -366,27 +458,30 @@ static int pcu_rx_data_ind(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_data *
 	int rc;
 	int current_fn = bts_current_frame_number(bts);
 	struct pcu_l1_meas meas = {0};
+	struct gprs_rlcmac_pdch *pdch;
 	uint8_t gsmtap_chantype;
 
-	LOGP(DL1IF, LOGL_DEBUG, "Data indication received: sapi=%d arfcn=%d "
-		"fn=%d cur_fn=%d block=%d data=%s\n", data_ind->sapi,
-		data_ind->arfcn, data_ind->fn, current_fn, data_ind->block_nr,
-		osmo_hexdump(data_ind->data, data_ind->len));
+	LOGP(DL1IF, LOGL_DEBUG, "(bts=%" PRIu8 ",trx=%" PRIu8 ",ts=%" PRIu8 ") FN=%u "
+		"Rx DATA.ind: sapi=%d arfcn=%d cur_fn=%d "
+		"block=%d data=%s\n", bts->nr, data_ind->trx_nr, data_ind->ts_nr,
+		data_ind->fn, data_ind->sapi, data_ind->arfcn, current_fn,
+		data_ind->block_nr, osmo_hexdump(data_ind->data, data_ind->len));
 
 	switch (data_ind->sapi) {
 	case PCU_IF_SAPI_PDTCH:
+		pdch = &bts->trx[data_ind->trx_nr].pdch[data_ind->ts_nr];
 		pcu_l1_meas_set_rssi(&meas, data_ind->rssi);
 		/* convert BER to % value */
 		pcu_l1_meas_set_ber(&meas, data_ind->ber10k / 100);
 		pcu_l1_meas_set_bto(&meas, data_ind->ta_offs_qbits);
 		pcu_l1_meas_set_link_qual(&meas, data_ind->lqual_cb / 10);
 
-		LOGP(DL1IF, LOGL_DEBUG, "Data indication with raw measurements received: BER10k = %d, BTO = %d, Q = %d\n",
-		     data_ind->ber10k, data_ind->ta_offs_qbits, data_ind->lqual_cb);
+		LOGPDCH(pdch, DL1IF, LOGL_DEBUG, "FN=%u Rx DATA.ind PDTCH: "
+			"BER10k = %d, BTO = %d, Q = %d\n", data_ind->fn,
+			data_ind->ber10k, data_ind->ta_offs_qbits, data_ind->lqual_cb);
 
-		rc = pcu_rx_data_ind_pdtch(bts, data_ind->trx_nr, data_ind->ts_nr,
-			data_ind->data, data_ind->len, data_ind->fn,
-			&meas);
+		rc = pcu_rx_data_ind_pdtch(bts, pdch, data_ind->data, data_ind->len,
+					   data_ind->fn, &meas);
 		gsmtap_chantype = GSMTAP_CHANNEL_PDTCH;
 		break;
 	case PCU_IF_SAPI_BCCH:
@@ -394,8 +489,9 @@ static int pcu_rx_data_ind(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_data *
 		gsmtap_chantype = GSMTAP_CHANNEL_BCCH;
 		break;
 	default:
-		LOGP(DL1IF, LOGL_ERROR, "Received PCU data indication with "
-			"unsupported sapi %d\n", data_ind->sapi);
+		LOGP(DL1IF, LOGL_ERROR, "(bts=%" PRIu8 ",trx=%" PRIu8 ",ts=%" PRIu8 ") "
+		     "FN=%u Rx DATA.ind with unsupported sapi %d\n",
+		     bts->nr, data_ind->trx_nr, data_ind->ts_nr, data_ind->fn, data_ind->sapi);
 		rc = -EINVAL;
 		gsmtap_chantype = GSMTAP_CHANNEL_UNKNOWN;
 	}
@@ -451,6 +547,25 @@ int pcu_rx_rts_req_ptcch(struct gprs_rlcmac_bts *bts, uint8_t trx, uint8_t ts,
 	if (!pdch->m_is_enabled)
 		return -EAGAIN;
 
+	/* If there's no TBF attached to this PDCH, we can skip Tx of PTCCH
+	 * since there's nothing worthy of being transmitted. This way BTS can
+	 * identify idle blocks and send nothing or dumy blocks with reduced
+	 * energy for the sake of energy saving.
+	 */
+	const unsigned num_tbfs = pdch->num_tbfs(GPRS_RLCMAC_DL_TBF)
+				+ pdch->num_tbfs(GPRS_RLCMAC_UL_TBF);
+	bool skip_idle = (num_tbfs == 0);
+#ifdef ENABLE_DIRECT_PHY
+		/* In DIRECT_PHY mode we want to always submit something to L1 in
+		 * TRX0, since BTS is not preparing dummy bursts on idle TS for us: */
+		skip_idle = skip_idle && trx != 0;
+#endif
+	if (skip_idle) {
+		pcu_l1if_tx_ptcch(bts, trx, ts, bts->trx[trx].arfcn, fn, block_nr,
+				  NULL, 0);
+		return 0;
+	}
+
 	pcu_l1if_tx_ptcch(bts, trx, ts, bts->trx[trx].arfcn, fn, block_nr,
 			  pdch->ptcch_msg, GSM_MACBLOCK_LEN);
 	return 0;
@@ -460,10 +575,12 @@ static int pcu_rx_rts_req(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_rts_req
 {
 	int rc = 0;
 	int current_fn = bts_current_frame_number(bts);
+	const struct gprs_rlcmac_pdch *pdch;
+	pdch = &bts->trx[rts_req->trx_nr].pdch[rts_req->ts_nr];
 
-	LOGP(DL1IF, LOGL_DEBUG, "RTS request received: trx=%d ts=%d sapi=%d "
-		"arfcn=%d fn=%d cur_fn=%d block=%d\n", rts_req->trx_nr, rts_req->ts_nr,
-		rts_req->sapi, rts_req->arfcn, rts_req->fn, current_fn, rts_req->block_nr);
+	LOGPDCH(pdch, DL1IF, LOGL_DEBUG, "FN=%u RX RTS.req: sapi=%d "
+		"arfcn=%d cur_fn=%d block=%d\n", rts_req->fn,
+		rts_req->sapi, rts_req->arfcn, current_fn, rts_req->block_nr);
 
 	switch (rts_req->sapi) {
 	case PCU_IF_SAPI_PDTCH:
@@ -475,8 +592,9 @@ static int pcu_rx_rts_req(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_rts_req
 			rts_req->fn, rts_req->block_nr);
 		break;
 	default:
-		LOGP(DL1IF, LOGL_ERROR, "Received PCU RTS request with "
-			"unsupported sapi %d\n", rts_req->sapi);
+		LOGP(DL1IF, LOGL_ERROR, "(bts=%u,trx=%u,ts=%u) FN=%u RX RTS.req with "
+		     "unsupported sapi %d\n", bts->nr, rts_req->trx_nr, rts_req->ts_nr,
+		     rts_req->fn, rts_req->sapi);
 		rc = -EINVAL;
 	}
 
@@ -617,7 +735,7 @@ bssgp_failed:
 		for (trx_nr = 0; trx_nr < ARRAY_SIZE(bts->trx); trx_nr++) {
 			bts->trx[trx_nr].arfcn = info_ind->trx[trx_nr].arfcn;
 			for (ts_nr = 0; ts_nr < ARRAY_SIZE(bts->trx[0].pdch); ts_nr++)
-				bts->trx[trx_nr].pdch[ts_nr].free_resources();
+				bts->trx[trx_nr].pdch[ts_nr].disable();
 		}
 		gprs_bssgp_destroy(bts);
 		exit(0);
@@ -758,7 +876,7 @@ bssgp_failed:
 						l1if_connect_pdch(
 							bts->trx[trx_nr].fl1h, ts_nr);
 #endif
-					pcu_tx_act_req(bts, trx_nr, ts_nr, 1);
+					pcu_tx_act_req(bts, pdch, 1);
 					pdch->enable();
 				}
 
@@ -784,8 +902,7 @@ bssgp_failed:
 				     trx_nr, ts_nr, pdch->tsc, pdch->fh.enabled ? "yes" : "no");
 			} else {
 				if (pdch->is_enabled()) {
-					pcu_tx_act_req(bts, trx_nr, ts_nr, 0);
-					pdch->free_resources();
+					pcu_tx_act_req(bts, pdch, 0);
 					pdch->disable();
 				}
 			}
@@ -806,13 +923,17 @@ static int pcu_rx_time_ind(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_time_i
 
 	LOGP(DL1IF, LOGL_DEBUG, "Time indication received: %d\n", time_ind->fn % 52);
 
-	bts_set_current_frame_number(bts, time_ind->fn);
+	/* Ignore TIME.ind completely, we nowadays relay on DATA.ind always
+	 * providing all block FNs. */
 	return 0;
 }
 
 static int pcu_rx_pag_req(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_pag_req *pag_req)
 {
 	struct osmo_mobile_identity mi;
+	struct GprsMs *ms = NULL;
+	struct paging_req_cs req = { .chan_needed = pag_req->chan_needed,
+				     .tlli = GSM_RESERVED_TMSI };
 	int rc;
 
 	LOGP(DL1IF, LOGL_DEBUG, "Paging request received: chan_needed=%d "
@@ -831,7 +952,23 @@ static int pcu_rx_pag_req(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_pag_req
 		return -EINVAL;
 	}
 
-	return bts_add_paging(bts, pag_req->chan_needed, &mi);
+	switch (mi.type) {
+	case GSM_MI_TYPE_TMSI:
+		req.mi_tmsi = mi;
+		req.mi_tmsi_present = true;
+		/* TODO: look up MS by TMSI? Derive TLLI? */
+		break;
+	case GSM_MI_TYPE_IMSI:
+		req.mi_imsi = mi;
+		req.mi_imsi_present = true;
+		ms = bts_ms_by_imsi(bts, req.mi_imsi.imsi);
+		break;
+	default:
+		LOGP(DL1IF, LOGL_ERROR, "Unexpected MI type %u\n", mi.type);
+		return -EINVAL;
+	}
+
+	return bts_add_paging(bts, &req, ms);
 }
 
 static int pcu_rx_susp_req(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_susp_req *susp_req)
@@ -898,9 +1035,81 @@ static int pcu_rx_app_info_req(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_ap
 	return 0;
 }
 
-int pcu_rx(uint8_t msg_type, struct gsm_pcu_if *pcu_prim)
+static int pcu_rx_neigh_addr_cnf(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_neigh_addr_cnf *naddr_cnf)
+{
+	struct llist_head *tmp;
+	struct osmo_cell_global_id_ps cgi_ps;
+	struct osmo_cell_global_id_ps *cgi_ps_ptr = &cgi_ps;
+
+	struct neigh_cache_entry_key neigh_key = {
+		.local_lac = osmo_load16be(&naddr_cnf->orig_req.local_lac),
+		.local_ci = osmo_load16be(&naddr_cnf->orig_req.local_ci),
+		.tgt_arfcn = osmo_load16be(&naddr_cnf->orig_req.tgt_arfcn),
+		.tgt_bsic = naddr_cnf->orig_req.tgt_bsic,
+	};
+
+	if (naddr_cnf->err_code == 0) {
+		cgi_ps.rai.lac.plmn.mcc = osmo_load16be(&naddr_cnf->cgi_ps.mcc);
+		cgi_ps.rai.lac.plmn.mnc = osmo_load16be(&naddr_cnf->cgi_ps.mnc);
+		cgi_ps.rai.lac.plmn.mnc_3_digits = naddr_cnf->cgi_ps.mnc_3_digits;
+		cgi_ps.rai.lac.lac = osmo_load16be(&naddr_cnf->cgi_ps.lac);
+		cgi_ps.rai.rac = naddr_cnf->cgi_ps.rac;
+		cgi_ps.cell_identity = osmo_load16be(&naddr_cnf->cgi_ps.cell_identity);
+
+		LOGP(DL1IF, LOGL_INFO, "Rx Neighbor Address Resolution Confirmation for " NEIGH_CACHE_ENTRY_KEY_FMT ": %s\n",
+		     NEIGH_CACHE_ENTRY_KEY_ARGS(&neigh_key), osmo_cgi_ps_name(&cgi_ps));
+
+		/* Cache the cgi_ps so we can avoid requesting again same resolution for a while */
+		neigh_cache_add(bts->pcu->neigh_cache, &neigh_key, &cgi_ps);
+	} else {
+		cgi_ps_ptr = NULL;
+		LOGP(DL1IF, LOGL_INFO, "Rx Neighbor Address Resolution Confirmation for " NEIGH_CACHE_ENTRY_KEY_FMT ": failed with err_code=%u\n",
+		     NEIGH_CACHE_ENTRY_KEY_ARGS(&neigh_key), naddr_cnf->err_code);
+	}
+
+	llist_for_each(tmp, bts_ms_store(bts)->ms_list()) {
+		GprsMs *ms = llist_entry(tmp, typeof(*ms), list);
+		if (ms->nacc && nacc_fsm_is_waiting_addr_resolution(ms->nacc, &neigh_key))
+			osmo_fsm_inst_dispatch(ms->nacc->fi, NACC_EV_RX_RAC_CI, cgi_ps_ptr);
+	}
+	return 0;
+}
+
+static int pcu_rx_container(struct gprs_rlcmac_bts *bts, struct gsm_pcu_if_container *container)
+{
+	int rc;
+	uint16_t data_length = osmo_load16be(&container->length);
+
+	switch (container->msg_type) {
+	case PCU_IF_MSG_NEIGH_ADDR_CNF:
+		if (data_length < sizeof(struct gsm_pcu_if_neigh_addr_cnf)) {
+			LOGP(DL1IF, LOGL_ERROR, "Rx container(NEIGH_ADDR_CNF) message too short: %u vs exp %zu\n",
+			     data_length, sizeof(struct gsm_pcu_if_neigh_addr_cnf));
+			return -EINVAL;
+		}
+		rc = pcu_rx_neigh_addr_cnf(bts, (struct gsm_pcu_if_neigh_addr_cnf*)&container->data);
+		break;
+	default:
+		LOGP(DL1IF, LOGL_NOTICE, "(bts=%d) Rx unexpected msg type (%u) inside container!\n",
+		     bts->nr, container->msg_type);
+		rc = -1;
+	}
+	return rc;
+}
+
+#define CHECK_IF_MSG_SIZE(prim_len, prim_msg) \
+	do { \
+		size_t _len = PCUIF_HDR_SIZE + sizeof(prim_msg); \
+		if (prim_len < _len) { \
+			LOGP(DL1IF, LOGL_ERROR, "Received %zu bytes on PCU Socket, but primitive %s " \
+			     "size is %zu, discarding\n", prim_len, #prim_msg, _len); \
+			return -EINVAL; \
+		} \
+	} while(0);
+int pcu_rx(struct gsm_pcu_if *pcu_prim, size_t pcu_prim_length)
 {
 	int rc = 0;
+	size_t exp_len;
 	struct gprs_rlcmac_bts *bts = gprs_pcu_get_bts_by_nr(the_pcu, pcu_prim->bts_nr);
 	if (!bts) {
 		LOGP(DL1IF, LOGL_NOTICE, "Received message for new BTS%d\n", pcu_prim->bts_nr);
@@ -911,37 +1120,59 @@ int pcu_rx(uint8_t msg_type, struct gsm_pcu_if *pcu_prim)
 		}
 	}
 
-	switch (msg_type) {
+	switch (pcu_prim->msg_type) {
 	case PCU_IF_MSG_DATA_IND:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.data_ind);
 		rc = pcu_rx_data_ind(bts, &pcu_prim->u.data_ind);
 		break;
 	case PCU_IF_MSG_DATA_CNF:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.data_cnf);
 		rc = pcu_rx_data_cnf(bts, &pcu_prim->u.data_cnf);
 		break;
 	case PCU_IF_MSG_RTS_REQ:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.rts_req);
 		rc = pcu_rx_rts_req(bts, &pcu_prim->u.rts_req);
 		break;
 	case PCU_IF_MSG_RACH_IND:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.rach_ind);
 		rc = pcu_rx_rach_ind(bts, &pcu_prim->u.rach_ind);
 		break;
 	case PCU_IF_MSG_INFO_IND:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.info_ind);
 		rc = pcu_rx_info_ind(bts, &pcu_prim->u.info_ind);
 		break;
 	case PCU_IF_MSG_TIME_IND:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.time_ind);
 		rc = pcu_rx_time_ind(bts, &pcu_prim->u.time_ind);
 		break;
 	case PCU_IF_MSG_PAG_REQ:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.pag_req);
 		rc = pcu_rx_pag_req(bts, &pcu_prim->u.pag_req);
 		break;
 	case PCU_IF_MSG_SUSP_REQ:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.susp_req);
 		rc = pcu_rx_susp_req(bts, &pcu_prim->u.susp_req);
 		break;
 	case PCU_IF_MSG_APP_INFO_REQ:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.app_info_req);
 		rc = pcu_rx_app_info_req(bts, &pcu_prim->u.app_info_req);
 		break;
+	case PCU_IF_MSG_INTERF_IND:
+		/* TODO: handle interference reports */
+		break;
+	case PCU_IF_MSG_CONTAINER:
+		CHECK_IF_MSG_SIZE(pcu_prim_length, pcu_prim->u.container);
+		/* ^ check if we can access container fields, v check with container data length */
+		exp_len = PCUIF_HDR_SIZE + sizeof(pcu_prim->u.container) + osmo_load16be(&pcu_prim->u.container.length);
+		if (pcu_prim_length < exp_len) {
+			LOGP(DL1IF, LOGL_ERROR, "Received %zu bytes on PCU Socket, but primitive container size" \
+			     "is %zu, discarding\n", pcu_prim_length, exp_len);
+		}
+		rc = pcu_rx_container(bts, &pcu_prim->u.container);
+		break;
 	default:
 		LOGP(DL1IF, LOGL_ERROR, "Received unknown PCU msg type %d\n",
-			msg_type);
+			pcu_prim->msg_type);
 		rc = -EINVAL;
 	}
 
