@@ -70,6 +70,49 @@ static const struct value_string pcu_gsmtap_categ_help[] = {
 	{ 0, NULL }
 };
 
+DEFUN(cfg_pcu_gsmtap_remote_host,
+      cfg_pcu_gsmtap_remote_host_cmd,
+      "gsmtap-remote-host [HOSTNAME]",
+      "Enable GSMTAP Um logging (see also 'gsmtap-category')\n"
+      "Remote IP address or hostname ('localhost' if omitted)\n")
+{
+	osmo_talloc_replace_string(the_pcu, &the_pcu->gsmtap_remote_host,
+				   argc > 0 ? argv[0] : "localhost");
+
+	if (vty->type != VTY_FILE)
+		vty_out(vty, "%% This command requires restart%s", VTY_NEWLINE);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_pcu_no_gsmtap_remote_host,
+      cfg_pcu_no_gsmtap_remote_host_cmd,
+      "no gsmtap-remote-host",
+      NO_STR "Disable GSMTAP Um logging\n")
+{
+	if (the_pcu->gsmtap_remote_host)
+		TALLOC_FREE(the_pcu->gsmtap_remote_host);
+
+	if (vty->type != VTY_FILE)
+		vty_out(vty, "%% This command requires restart%s", VTY_NEWLINE);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_pcu_gsmtap_sapi_all, pcucfg_pcu_gsmtap_categ_all_cmd,
+	"gsmtap-category (enable-all|disable-all)",
+	"Enable/disable sending of UL/DL messages over GSMTAP\n"
+	"Enable all kinds of messages (all categories)\n"
+	"Disable all kinds of messages (all categories)\n")
+{
+
+	if (strcmp(argv[0], "enable-all") == 0)
+		the_pcu->gsmtap_categ_mask = UINT32_MAX;
+	else
+		the_pcu->gsmtap_categ_mask = 0x00;
+
+	return CMD_SUCCESS;
+}
 
 DEFUN(cfg_pcu_gsmtap_categ, cfg_pcu_gsmtap_categ_cmd, "HIDDEN", "HIDDEN")
 {
@@ -238,11 +281,14 @@ static int config_write_pcu(struct vty *vty)
 	if (strcmp(the_pcu->pcu_sock_path, PCU_SOCK_DEFAULT))
 		vty_out(vty, " pcu-socket %s%s", the_pcu->pcu_sock_path, VTY_NEWLINE);
 
+	if (the_pcu->gsmtap_remote_host)
+		vty_out(vty, " gsmtap-remote-host %s%s", the_pcu->gsmtap_remote_host, VTY_NEWLINE);
 	for (i = 0; i < 32; i++) {
-		unsigned int cs = (1 << i);
-		if (the_pcu->gsmtap_categ_mask & cs) {
-			vty_out(vty, " gsmtap-category %s%s",
-				get_value_string(pcu_gsmtap_categ_names, i), VTY_NEWLINE);
+		if (the_pcu->gsmtap_categ_mask & ((uint32_t)1 << i)) {
+			const char *category_buf;
+			if (!(category_buf = get_value_string_or_null(pcu_gsmtap_categ_names, i)))
+				continue;
+			vty_out(vty, " gsmtap-category %s%s", category_buf, VTY_NEWLINE);
 		}
 	}
 
@@ -251,6 +297,11 @@ static int config_write_pcu(struct vty *vty)
 	else
 		vty_out(vty, " gb-dialect classic%s", VTY_NEWLINE);
 
+	if (the_pcu->vty.ns_ip_dscp != -1)
+		vty_out(vty, " gb ip-dscp %d%s", the_pcu->vty.ns_ip_dscp, VTY_NEWLINE);
+	if (the_pcu->vty.ns_priority != -1)
+		vty_out(vty, " gb socket-priority %d%s", the_pcu->vty.ns_priority, VTY_NEWLINE);
+
 	if (the_pcu->vty.neigh_ctrl_addr) {
 		vty_out(vty, " neighbor resolution %s %u%s",
 			the_pcu->vty.neigh_ctrl_addr, the_pcu->vty.neigh_ctrl_port, VTY_NEWLINE);
@@ -1031,6 +1082,30 @@ DEFUN_USRATTR(cfg_pcu_gb_dialect,
 	return CMD_SUCCESS;
 }
 
+DEFUN_USRATTR(cfg_pcu_gb_ip_dscp,
+	      cfg_pcu_gb_ip_dscp_cmd,
+	      X(PCU_VTY_ATTR_NS_RESET),
+	      "gb ip-dscp <0-63>",
+	      "Configure Gb interface\n"
+	      "Set IP DSCP value for outbound packets\n"
+	      "IP DSCP value to use\n")
+{
+	the_pcu->vty.ns_ip_dscp = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
+DEFUN_USRATTR(cfg_pcu_gb_priority,
+	      cfg_pcu_gb_priority_cmd,
+	      X(PCU_VTY_ATTR_NS_RESET),
+	      "gb socket-priority <0-255>",
+	      "Configure Gb interface\n"
+	      "Set socket priority value for outbound packets\n"
+	      "Socket priority value to use (>6 requires CAP_NET_ADMIN)")
+{
+	the_pcu->vty.ns_priority = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_neighbor_resolution, cfg_neighbor_resolution_cmd,
        "neighbor resolution " VTY_IPV46_CMD " [<0-65535>]",
        "Manage local and remote-BSS neighbor cells\n"
@@ -1038,6 +1113,10 @@ DEFUN(cfg_neighbor_resolution, cfg_neighbor_resolution_cmd,
        "IPv4 address to connect to\n" "IPv6 address to connect to\n"
        "Port to connect to (default 4248)\n")
 {
+	vty_out(vty, "%% Warning: The CTRL interface for Neighbor Address Resolution is now deprecated."
+		"Upgrade osmo-bsc and drop the 'neighbor resolution " VTY_IPV46_CMD " [<0-65535>]' VTY "
+		"option in order to let osmo-pcu use the new resoluton method using the PCUIF over IPA "
+		"multiplex, which will work out of the box without required configuration.%s", VTY_NEWLINE);
 	osmo_talloc_replace_string(the_pcu, &the_pcu->vty.neigh_ctrl_addr, argv[0]);
 	if (argc > 1)
 		the_pcu->vty.neigh_ctrl_port = atoi(argv[1]);
@@ -1261,10 +1340,15 @@ int pcu_vty_init(void)
 	install_element(PCU_NODE, &cfg_pcu_no_dl_tbf_preemptive_retransmission_cmd);
 	install_element(PCU_NODE, &cfg_pcu_ms_idle_time_cmd);
 	install_element(PCU_NODE, &cfg_pcu_no_ms_idle_time_cmd);
+	install_element(PCU_NODE, &cfg_pcu_gsmtap_remote_host_cmd);
+	install_element(PCU_NODE, &cfg_pcu_no_gsmtap_remote_host_cmd);
+	install_element(PCU_NODE, &pcucfg_pcu_gsmtap_categ_all_cmd);
 	install_element(PCU_NODE, &cfg_pcu_gsmtap_categ_cmd);
 	install_element(PCU_NODE, &cfg_pcu_no_gsmtap_categ_cmd);
 	install_element(PCU_NODE, &cfg_pcu_sock_cmd);
 	install_element(PCU_NODE, &cfg_pcu_gb_dialect_cmd);
+	install_element(PCU_NODE, &cfg_pcu_gb_ip_dscp_cmd);
+	install_element(PCU_NODE, &cfg_pcu_gb_priority_cmd);
 	install_element(PCU_NODE, &cfg_neighbor_resolution_cmd);
 	install_element(PCU_NODE, &cfg_pcu_timer_cmd);
 
