@@ -21,14 +21,15 @@
 
 #pragma once
 
+#include <inttypes.h>
+
 #ifdef __cplusplus
 extern "C" {
 #include <osmocom/core/linuxlist.h>
 #include "gsm_rlcmac.h"
 #include "coding_scheme.h"
+#include "pdch_ul_controller.h"
 }
-
-#include <bts.h>
 #endif
 
 #include <tbf.h>
@@ -54,8 +55,6 @@ struct gprs_rlcmac_pdch {
 
 	bool add_paging(uint8_t chan_needed, const struct osmo_mobile_identity *mi);
 
-	void free_resources();
-
 	bool is_enabled() const;
 
 	void enable();
@@ -79,6 +78,7 @@ struct gprs_rlcmac_pdch {
 	void detach_tbf(gprs_rlcmac_tbf *tbf);
 
 	unsigned num_tbfs(enum gprs_rlcmac_tbf_direction dir) const;
+	void num_tbfs_update(gprs_rlcmac_tbf *tbf, bool is_attach);
 
 	void reserve(enum gprs_rlcmac_tbf_direction dir);
 	void unreserve(enum gprs_rlcmac_tbf_direction dir);
@@ -103,8 +103,6 @@ struct gprs_rlcmac_pdch {
 	/* PTCCH (Packet Timing Advance Control Channel) */
 	uint8_t ptcch_msg[GSM_MACBLOCK_LEN]; /* 'ready to use' PTCCH/D message */
 #ifdef __cplusplus
-	/* Initialize the PTCCH/D message */
-	void init_ptcch_msg(void);
 	/* Obtain an unused TA Index for a TBF */
 	uint8_t reserve_tai(uint8_t ta);
 	/* Mark a given TA Index as free, so it can be used again */
@@ -128,6 +126,7 @@ struct gprs_rlcmac_pdch {
 	/* back pointers */
 	struct gprs_rlcmac_trx *trx;
 	uint8_t ts_no;
+	struct pdch_ulc *ulc;
 
 #ifdef __cplusplus
 private:
@@ -145,9 +144,12 @@ private:
 		enum gprs_rlcmac_tbf_direction dir);
 	gprs_rlcmac_tbf *tbf_by_tfi(uint8_t tfi,
 		enum gprs_rlcmac_tbf_direction dir);
+
+	void free_resources();
 #endif
 
-	uint8_t m_num_tbfs[2];
+	uint8_t m_num_tbfs_gprs[2];
+	uint8_t m_num_tbfs_egprs[2];
 	uint8_t m_num_reserved[2];
 	uint8_t m_assigned_usf; /* bit set */
 	uint32_t m_assigned_tfi[2]; /* bit set */
@@ -158,7 +160,7 @@ private:
 
 inline unsigned gprs_rlcmac_pdch::num_tbfs(enum gprs_rlcmac_tbf_direction dir) const
 {
-	return m_num_tbfs[dir];
+	return m_num_tbfs_gprs[dir] + m_num_tbfs_egprs[dir];
 }
 
 inline unsigned gprs_rlcmac_pdch::num_reserved(
@@ -188,8 +190,16 @@ inline bool gprs_rlcmac_pdch::is_enabled() const
 #ifdef __cplusplus
 extern "C" {
 #endif
+void pdch_init(struct gprs_rlcmac_pdch *pdch, struct gprs_rlcmac_trx *trx, uint8_t ts_nr);
 void pdch_free_all_tbf(struct gprs_rlcmac_pdch *pdch);
 void pdch_disable(struct gprs_rlcmac_pdch *pdch);
+bool pdch_is_enabled(const struct gprs_rlcmac_pdch *pdch);
+bool pdch_is_full(const struct gprs_rlcmac_pdch *pdch);
 #ifdef __cplusplus
 }
 #endif
+
+#define LOGPDCH(pdch, category, level, fmt, args...) \
+	LOGP(category, level, "PDCH(bts=%" PRIu8 ",trx=%" PRIu8 ",ts=%" PRIu8 ") " fmt, \
+	     (pdch)->trx->bts->nr, (pdch)->trx->trx_no, (pdch)->ts_no, \
+	     ## args)
