@@ -98,7 +98,7 @@ static void llc_timer_cb(void *_tbf)
 {
 	struct gprs_rlcmac_dl_tbf *tbf = (struct gprs_rlcmac_dl_tbf *)_tbf;
 
-	if (tbf->state_is_not(GPRS_RLCMAC_FLOW))
+	if (tbf->state_is_not(TBF_ST_FLOW))
 		return;
 
 	LOGPTBFDL(tbf, LOGL_DEBUG, "LLC receive timeout, requesting DL ACK\n");
@@ -129,9 +129,8 @@ struct gprs_rlcmac_dl_tbf *tbf_alloc_dl_tbf(struct gprs_rlcmac_bts *bts, GprsMs
 
 	OSMO_ASSERT(ms != NULL);
 
-	LOGP(DTBF, LOGL_DEBUG, "********** DL-TBF starts here **********\n");
-	LOGP(DTBF, LOGL_INFO, "Allocating DL TBF: MS_CLASS=%d/%d\n",
-	     ms_ms_class(ms), ms_egprs_ms_class(ms));
+	LOGPMS(ms, DTBF, LOGL_DEBUG, "********** DL-TBF starts here **********\n");
+	LOGPMS(ms, DTBF, LOGL_INFO, "Allocating DL TBF\n");
 
 	tbf = talloc(tall_pcu_ctx, struct gprs_rlcmac_dl_tbf);
 
@@ -169,25 +168,32 @@ struct gprs_rlcmac_dl_tbf *tbf_alloc_dl_tbf(struct gprs_rlcmac_bts *bts, GprsMs
 		}
 	}
 
-	llist_add(tbf_bts_list((struct gprs_rlcmac_tbf *)tbf), &bts->dl_tbfs);
+	llist_add(tbf_trx_list((struct gprs_rlcmac_tbf *)tbf), &tbf->trx->dl_tbfs);
 	bts_do_rate_ctr_inc(tbf->bts, CTR_TBF_DL_ALLOCATED);
 
-	tbf->m_last_dl_poll_fn = -1;
-	tbf->m_last_dl_drained_fn = -1;
-
 	osmo_clock_gettime(CLOCK_MONOTONIC, &tbf->m_bw.dl_bw_tv);
 	osmo_clock_gettime(CLOCK_MONOTONIC, &tbf->m_bw.dl_loss_tv);
 
 	return tbf;
 }
 
+gprs_rlcmac_dl_tbf::~gprs_rlcmac_dl_tbf()
+{
+	osmo_timer_del(&m_llc_timer);
+	if (is_egprs_enabled()) {
+		rate_ctr_group_free(m_dl_egprs_ctrs);
+	} else {
+		rate_ctr_group_free(m_dl_gprs_ctrs);
+	}
+	/* ~gprs_rlcmac_tbf() is called automatically upon return */
+}
+
 gprs_rlcmac_dl_tbf::gprs_rlcmac_dl_tbf(struct gprs_rlcmac_bts *bts_, GprsMs *ms) :
 	gprs_rlcmac_tbf(bts_, ms, GPRS_RLCMAC_DL_TBF),
 	m_tx_counter(0),
-	m_wait_confirm(0),
 	m_dl_ack_requested(false),
-	m_last_dl_poll_fn(0),
-	m_last_dl_drained_fn(0),
+	m_last_dl_poll_fn(-1),
+	m_last_dl_drained_fn(-1),
 	m_dl_gprs_ctrs(NULL),
 	m_dl_egprs_ctrs(NULL)
 {
@@ -195,11 +201,6 @@ gprs_rlcmac_dl_tbf::gprs_rlcmac_dl_tbf(struct gprs_rlcmac_bts *bts_, GprsMs *ms)
 	osmo_timer_setup(&m_llc_timer, llc_timer_cb, this);
 }
 
-void gprs_rlcmac_dl_tbf::cleanup()
-{
-	osmo_timer_del(&m_llc_timer);
-}
-
 void gprs_rlcmac_dl_tbf::start_llc_timer()
 {
 	if (the_pcu->vty.llc_idle_ack_csec > 0) {
@@ -225,7 +226,7 @@ int gprs_rlcmac_dl_tbf::append_data(uint16_t pdu_delay_csec,
 	llc_queue()->enqueue(llc_msg, &expire_time);
 	start_llc_timer();
 
-	if (state_is(GPRS_RLCMAC_WAIT_RELEASE)) {
+	if (state_is(TBF_ST_WAIT_RELEASE)) {
 		LOGPTBFDL(this, LOGL_DEBUG, "in WAIT RELEASE state (T3193), so reuse TBF\n");
 		establish_dl_tbf_on_pacch();
 	}
@@ -243,8 +244,12 @@ static int tbf_new_dl_assignment(struct gprs_rlcmac_bts *bts, GprsMs *ms,
 
 	ul_tbf = ms_ul_tbf(ms);
 
+	/* 3GPP TS 44.060 sec 7.1.3.1 Initiation of the Packet resource request procedure:
+	* "Furthermore, the mobile station shall not respond to PACKET DOWNLINK ASSIGNMENT
+	* or MULTIPLE TBF DOWNLINK ASSIGNMENT messages before contention resolution is
+	* completed on the mobile station side." */
 	if (ul_tbf && ul_tbf->m_contention_resolution_done
-	 && !ul_tbf->m_final_ack_sent) {
+	 && !tbf_ul_ack_waiting_cnf_final_ack(ul_tbf)) {
 		use_trx = ul_tbf->trx->trx_no;
 		ss = false;
 		old_ul_tbf = ul_tbf;
@@ -260,7 +265,7 @@ static int tbf_new_dl_assignment(struct gprs_rlcmac_bts *bts, GprsMs *ms,
 	dl_tbf = tbf_alloc_dl_tbf(bts, ms, use_trx, ss);
 
 	if (!dl_tbf) {
-		LOGP(DTBF, LOGL_NOTICE, "No PDCH resource\n");
+		LOGPMS(ms, DTBF, LOGL_NOTICE, "No PDCH resource\n");
 		return -EBUSY;
 	}
 
@@ -291,7 +296,9 @@ int dl_tbf_handle(struct gprs_rlcmac_bts *bts,
 	/* check for existing TBF */
 	ms = bts_ms_store(bts)->get_ms(tlli, tlli_old, imsi);
 
-	if (ms && strlen(ms_imsi(ms)) == 0) {
+	/* If we got MS by TLLI above let's see if we already have another MS
+	 * object identified by IMSI and merge them */
+	if (ms && !ms_imsi_is_valid(ms) && imsi) {
 		ms_old = bts_ms_store(bts)->get_ms(0, 0, imsi);
 		if (ms_old && ms_old != ms) {
 			/* The TLLI has changed (RAU), so there are two MS
@@ -305,7 +312,7 @@ int dl_tbf_handle(struct gprs_rlcmac_bts *bts,
 			if (!ms_dl_tbf(ms) && ms_dl_tbf(ms_old)) {
 				LOGP(DTBF, LOGL_NOTICE,
 				     "IMSI %s, old TBF %s: moving DL TBF to new MS object\n",
-				     imsi, ms_dl_tbf(ms_old)->name());
+				     imsi ? : "unknown", ms_dl_tbf(ms_old)->name());
 				dl_tbf = ms_dl_tbf(ms_old);
 				/* Move the DL TBF to the new MS */
 				dl_tbf->set_ms(ms);
@@ -318,7 +325,8 @@ int dl_tbf_handle(struct gprs_rlcmac_bts *bts,
 
 	if (!ms)
 		ms = bts_alloc_ms(bts, ms_class, egprs_ms_class);
-	ms_set_imsi(ms, imsi);
+	if (imsi)
+		ms_set_imsi(ms, imsi);
 	ms_confirm_tlli(ms, tlli);
 	if (!ms_ms_class(ms) && ms_class) {
 		ms_set_ms_class(ms, ms_class);
@@ -510,7 +518,7 @@ int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
 		/* re-send block with negative aknowlegement */
 		m_window.m_v_b.mark_unacked(bsn);
 		bts_do_rate_ctr_inc(bts, CTR_RLC_RESENT);
-	} else if (state_is(GPRS_RLCMAC_FINISHED)) {
+	} else if (state_is(TBF_ST_FINISHED)) {
 		/* If the TBF is in finished, we already sent all packages at least once.
 		 * If any packages could have been sent (because of unacked) it should have
 		 * been catched up by the upper if(bsn >= 0) */
@@ -594,34 +602,25 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(uint32_t fn, uint8_t ts,
 /* depending on the current TBF, we assign on PACCH or AGCH */
 void gprs_rlcmac_dl_tbf::trigger_ass(struct gprs_rlcmac_tbf *old_tbf)
 {
-	uint16_t pgroup;
 	/* stop pending timer */
 	stop_timers("assignment (DL-TBF)");
 
 	/* check for downlink tbf:  */
 	if (old_tbf) {
-		LOGPTBFDL(this, LOGL_DEBUG, "Send dowlink assignment on PACCH, because %s exists\n", old_tbf->name());
-		TBF_SET_ASS_STATE_DL(old_tbf, GPRS_RLCMAC_DL_ASS_SEND_ASS);
-		old_tbf->was_releasing = old_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE);
+		LOGPTBFDL(this, LOGL_DEBUG, "Send downlink assignment on PACCH, because %s exists\n", old_tbf->name());
+		osmo_fsm_inst_dispatch(old_tbf->dl_ass_fsm.fi, TBF_DL_ASS_EV_SCHED_ASS, NULL);
 
 		/* change state */
-		TBF_SET_ASS_ON(this, GPRS_RLCMAC_FLAG_PACCH, true);
-
-		/* start timer */
-		T_START(this, T0, -2001, "assignment (PACCH)", true);
+		osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_ASSIGN_ADD_PACCH, NULL);
 	} else {
-		LOGPTBFDL(this, LOGL_DEBUG, "Send dowlink assignment on PCH, no TBF exist (IMSI=%s)\n",
+		LOGPTBFDL(this, LOGL_DEBUG, "Send downlink assignment on PCH, no TBF exist (IMSI=%s)\n",
 			  imsi());
-		was_releasing = state_is(GPRS_RLCMAC_WAIT_RELEASE);
 
 		/* change state */
-		TBF_SET_ASS_ON(this, GPRS_RLCMAC_FLAG_CCCH, false);
+		osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_ASSIGN_ADD_CCCH, NULL);
 
 		/* send immediate assignment */
-		if ((pgroup = imsi2paging_group(imsi())) > 999)
-			LOGPTBFDL(this, LOGL_ERROR, "IMSI to paging group failed! (%s)\n", imsi());
-		bts_snd_dl_ass(bts, this, false, pgroup);
-		m_wait_confirm = 1;
+		bts_snd_dl_ass(bts, this);
 	}
 }
 
@@ -688,6 +687,10 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, enum CodingScheme cs)
 		int payload_written = 0;
 
 		if (llc_frame_length(&m_llc) == 0) {
+			/* The data just drained, store the current fn */
+			if (m_last_dl_drained_fn < 0)
+				m_last_dl_drained_fn = fn;
+
 			/* It is not clear, when the next real data will
 			 * arrive, so request a DL ack/nack now */
 			request_dl_ack();
@@ -700,6 +703,16 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, enum CodingScheme cs)
 				LOGPTBFDL(this, LOGL_DEBUG,
 					  "LLC queue completely drained and there's "
 					  "still %d free bytes in rlcmac data block\n", space);
+
+				/* We may need to update fbi in header here
+				 * since m_last_dl_drained_fn was updated above
+				 * Specially important when X2031 is 0. */
+				is_final = llc_queue_size(llc_queue()) == 0 && !keep_open(fn);
+				if (is_final) {
+					rdbi->cv = 0;
+					osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_LAST_DL_DATA_SENT, NULL);
+				}
+
 				if (mcs_is_edge(cs)) {
 					/* in EGPRS there's no M bit, so we need
 					 * to flag padding with LI=127 */
@@ -719,10 +732,6 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, enum CodingScheme cs)
 			 * space-1 octets */
 			m_llc.put_dummy_frame(space - 1);
 
-			/* The data just drained, store the current fn */
-			if (m_last_dl_drained_fn < 0)
-				m_last_dl_drained_fn = fn;
-
 			LOGPTBFDL(this, LOGL_DEBUG,
 				  "Empty chunk, added LLC dummy command of size %d, drained_since=%d\n",
 				  llc_frame_length(&m_llc), frames_since_last_drain(fn));
@@ -746,7 +755,7 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, enum CodingScheme cs)
 
 		if (is_final) {
 			request_dl_ack();
-			TBF_SET_STATE(this, GPRS_RLCMAC_FINISHED);
+			osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_LAST_DL_DATA_SENT, NULL);
 		}
 
 		/* dequeue next LLC frame, if any */
@@ -763,23 +772,6 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, enum CodingScheme cs)
 	return bsn;
 }
 
-bool gprs_rlcmac_dl_tbf::handle_ack_nack()
-{
-	bool ack_recovered = false;
-
-	state_flags |= (1 << GPRS_RLCMAC_FLAG_DL_ACK);
-	if (check_n_clear(GPRS_RLCMAC_FLAG_TO_DL_ACK)) {
-		ack_recovered = true;
-	}
-
-	/* reset N3105 */
-	n_reset(N3105);
-	t_stop(T3191, "ACK/NACK received");
-	TBF_POLL_SCHED_UNSET(this);
-
-	return ack_recovered;
-}
-
 struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 				const uint32_t fn, const uint8_t ts,
 				int index, int index2)
@@ -787,7 +779,6 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 	uint8_t *msg_data;
 	struct msgb *dl_msg;
 	unsigned msg_len;
-	bool need_poll;
 	/* TODO: support MCS-7 - MCS-9, where data_block_idx can be 1 */
 	uint8_t data_block_idx = 0;
 	unsigned int rrbp;
@@ -948,35 +939,35 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 	if (m_last_dl_poll_fn < 0)
 		m_last_dl_poll_fn = fn;
 
-	need_poll = state_flags & (1 << GPRS_RLCMAC_FLAG_TO_DL_ACK);
-
-	/* poll after POLL_ACK_AFTER_FRAMES frames, or when final block is tx.
-	 */
-	if (m_tx_counter >= POLL_ACK_AFTER_FRAMES || m_dl_ack_requested ||
-			need_poll) {
+	/* poll after POLL_ACK_AFTER_FRAMES frames, or when final block is tx or
+	 * when last polled DL ACK/NACK was lost. */
+	if (need_poll_for_dl_ack_nack()) {
 		if (m_dl_ack_requested) {
 			LOGPTBFDL(this, LOGL_DEBUG,
 				  "Scheduling Ack/Nack polling, because it was requested explicitly "
 				  "(e.g. first final block sent).\n");
-		} else if (need_poll) {
+		} else if (state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_TO_DL_ACK)) {
 			LOGPTBFDL(this, LOGL_DEBUG,
 				  "Scheduling Ack/Nack polling, because polling timed out.\n");
 		} else {
 			LOGPTBFDL(this, LOGL_DEBUG,
 				  "Scheduling Ack/Nack polling, because %d blocks sent.\n",
-				POLL_ACK_AFTER_FRAMES);
+				  POLL_ACK_AFTER_FRAMES);
 		}
 
 		rc = check_polling(fn, ts, &new_poll_fn, &rrbp);
 		if (rc >= 0) {
-			set_polling(new_poll_fn, ts, GPRS_RLCMAC_POLL_DL_ACK);
+			set_polling(new_poll_fn, ts, PDCH_ULC_POLL_DL_ACK);
+			LOGPTBFDL(this, LOGL_DEBUG,
+				  "Scheduled DL Acknowledgement polling on PACCH (FN=%d, TS=%d)\n",
+				  new_poll_fn, ts);
 
 			m_tx_counter = 0;
 			/* start timer whenever we send the final block */
 			if (is_final)
 				T_START(this, T3191, 3191, "final block (DL-TBF)", true);
 
-			state_flags &= ~(1 << GPRS_RLCMAC_FLAG_TO_DL_ACK); /* clear poll timeout flag */
+			state_fsm.state_flags &= ~(1 << GPRS_RLCMAC_FLAG_TO_DL_ACK); /* clear poll timeout flag */
 
 			/* Clear request flag */
 			m_dl_ack_requested = false;
@@ -985,11 +976,11 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 			rlc.rrbp = rrbp;
 			rlc.es_p = 1; /* Polling */
 
-			m_last_dl_poll_fn = poll_fn;
+			m_last_dl_poll_fn = new_poll_fn;
 
 			LOGPTBFDL(this, LOGL_INFO,
 				  "Scheduled Ack/Nack polling on FN=%d, TS=%d\n",
-				  poll_fn, poll_ts);
+				  new_poll_fn, ts);
 		}
 	}
 
@@ -1128,7 +1119,7 @@ int gprs_rlcmac_dl_tbf::update_window(unsigned first_bsn,
 		ms_update_error_rate(ms(), this, error_rate);
 
 	m_window.update(bts, rbb, first_bsn, &lost, &received);
-	rate_ctr_add(&m_ctrs->ctr[TBF_CTR_RLC_NACKED], lost);
+	rate_ctr_add(rate_ctr_group_get_ctr(m_ctrs, TBF_CTR_RLC_NACKED), lost);
 
 	/* report lost and received packets */
 	gprs_rlcmac_received_lost(this, received, lost);
@@ -1148,71 +1139,18 @@ int gprs_rlcmac_dl_tbf::update_window(unsigned first_bsn,
 	return 0;
 }
 
-int gprs_rlcmac_dl_tbf::update_window(const uint8_t ssn, const uint8_t *rbb)
+int gprs_rlcmac_dl_tbf::rcvd_dl_final_ack()
 {
-	int16_t dist; /* must be signed */
-	uint16_t lost = 0, received = 0;
-	char show_rbb[65];
-	char show_v_b[RLC_MAX_SNS + 1];
-	int error_rate;
-	struct ana_result ana_res;
+	uint16_t received;
 
-	Decoding::extract_rbb(rbb, show_rbb);
-	/* show received array in debug (bit 64..1) */
-	LOGPTBFDL(this, LOGL_DEBUG,
-		  "ack:  (BSN=%d)\"%s\"(BSN=%d)  R=ACK I=NACK\n",
-		  m_window.mod_sns(ssn - 64), show_rbb, m_window.mod_sns(ssn - 1));
+	osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_FINAL_ACK_RECVD, NULL);
 
-	/* apply received array to receive state (SSN-64..SSN-1) */
-	/* calculate distance of ssn from V(S) */
-	dist = m_window.mod_sns(m_window.v_s() - ssn);
-	/* check if distance is less than distance V(A)..V(S) */
-	if (dist >= m_window.distance()) {
-		/* this might happpen, if the downlink assignment
-		 * was not received by ms and the ack refers
-		 * to previous TBF
-		 * FIXME: we should implement polling for
-		 * control ack!*/
-		LOGPTBFDL(this, LOGL_NOTICE, "ack range is out of V(A)..V(S) range - Free TBF!\n");
-		return 1; /* indicate to free TBF */
-	}
-
-	error_rate = analyse_errors(show_rbb, ssn, &ana_res);
-
-	if (the_pcu->vty.cs_adj_enabled && ms())
-		ms_update_error_rate(ms(), this, error_rate);
-
-	m_window.update(bts, show_rbb, ssn,
-			&lost, &received);
-	rate_ctr_add(&m_ctrs->ctr[TBF_CTR_RLC_NACKED], lost);
-
-	/* report lost and received packets */
-	gprs_rlcmac_received_lost(this, received, lost);
-
-	/* Used to measure the leak rate */
-	gprs_bssgp_update_bytes_received(ana_res.received_bytes,
-		ana_res.received_packets + ana_res.lost_packets);
-
-	/* raise V(A), if possible */
-	m_window.raise(m_window.move_window());
-
-	/* show receive state array in debug (V(A)..V(S)-1) */
-	m_window.show_state(show_v_b);
-	LOGPTBFDL(this, LOGL_DEBUG,
-		  "V(B): (V(A)=%d)\"%s\"(V(S)-1=%d)  A=Acked N=Nacked U=Unacked X=Resend-Unacked I=Invalid\n",
-		  m_window.v_a(), show_v_b, m_window.v_s_mod(-1));
-
-	if (state_is(GPRS_RLCMAC_FINISHED) && m_window.window_empty()) {
-		LOGPTBFDL(this, LOGL_NOTICE,
-			  "Received acknowledge of all blocks, but without final ack inidcation (don't worry)\n");
-	}
-	return 0;
-}
-
-
-int gprs_rlcmac_dl_tbf::maybe_start_new_window()
-{
-	release();
+	/* range V(A)..V(S)-1 */
+	received = m_window.count_unacked();
+	/* report all outstanding packets as received */
+	gprs_rlcmac_received_lost(this, received, 0);
+	m_tx_counter = 0;
+	m_window.reset();
 
 	/* check for LLC PDU in the LLC Queue */
 	if (llc_queue_size(llc_queue()) > 0)
@@ -1222,69 +1160,25 @@ int gprs_rlcmac_dl_tbf::maybe_start_new_window()
 	return 0;
 }
 
-int gprs_rlcmac_dl_tbf::release()
-{
-	uint16_t received;
-
-	/* range V(A)..V(S)-1 */
-	received = m_window.count_unacked();
-
-	/* report all outstanding packets as received */
-	gprs_rlcmac_received_lost(this, received, 0);
-
-	TBF_SET_STATE(this, GPRS_RLCMAC_WAIT_RELEASE);
-
-	/* start T3193 */
-	T_START(this, T3193, 3193, "release (DL-TBF)", true);
-
-	/* reset rlc states */
-	m_tx_counter = 0;
-	m_wait_confirm = 0;
-	m_window.reset();
-
-	TBF_ASS_TYPE_UNSET(this, GPRS_RLCMAC_FLAG_CCCH);
-
-	return 0;
-}
-
-int gprs_rlcmac_dl_tbf::abort()
-{
-	uint16_t lost;
-
-	if (state_is(GPRS_RLCMAC_FLOW)) {
-		/* range V(A)..V(S)-1 */
-		lost = m_window.count_unacked();
-
-		/* report all outstanding packets as lost */
-		gprs_rlcmac_received_lost(this, 0, lost);
-		gprs_rlcmac_lost_rep(this);
-
-		/* TODO: Reschedule all LLC frames starting with the one that is
-		 * (partly) encoded in chunk 1 of block V(A). (optional) */
-	}
-
-	TBF_SET_STATE(this, GPRS_RLCMAC_RELEASING);
-
-	/* reset rlc states */
-	m_window.reset();
-
-	TBF_ASS_TYPE_UNSET(this, GPRS_RLCMAC_FLAG_CCCH);
-
-	return 0;
-}
-
 int gprs_rlcmac_dl_tbf::rcvd_dl_ack(bool final_ack, unsigned first_bsn,
 	struct bitvec *rbb)
 {
 	int rc;
 	LOGPTBFDL(this, LOGL_DEBUG, "downlink acknowledge\n");
 
+	state_fsm.state_flags |= (1 << GPRS_RLCMAC_FLAG_DL_ACK);
+	state_fsm.state_flags &= ~(1 << GPRS_RLCMAC_FLAG_TO_DL_ACK);
+
+	/* reset N3105 */
+	n_reset(N3105);
+	t_stop(T3191, "ACK/NACK received");
+
 	rc = update_window(first_bsn, rbb);
 
 	if (final_ack) {
 		LOGPTBFDL(this, LOGL_DEBUG, "Final ACK received.\n");
-		rc = maybe_start_new_window();
-	} else if (state_is(GPRS_RLCMAC_FINISHED) && m_window.window_empty()) {
+		rc = rcvd_dl_final_ack();
+	} else if (state_is(TBF_ST_FINISHED) && m_window.window_empty()) {
 		LOGPTBFDL(this, LOGL_NOTICE,
 			  "Received acknowledge of all blocks, but without final ack indication (don't worry)\n");
 	}
@@ -1292,17 +1186,6 @@ int gprs_rlcmac_dl_tbf::rcvd_dl_ack(bool final_ack, unsigned first_bsn,
 	return rc;
 }
 
-int gprs_rlcmac_dl_tbf::rcvd_dl_ack(bool final_ack, uint8_t ssn, uint8_t *rbb)
-{
-	LOGPTBFDL(this, LOGL_DEBUG, "downlink acknowledge\n");
-
-	if (!final_ack)
-		return update_window(ssn, rbb);
-
-	LOGPTBFDL(this, LOGL_DEBUG, "Final ACK received.\n");
-	return maybe_start_new_window();
-}
-
 bool gprs_rlcmac_dl_tbf::dl_window_stalled() const
 {
 	return m_window.window_stalled();
@@ -1313,12 +1196,12 @@ void gprs_rlcmac_dl_tbf::request_dl_ack()
 	m_dl_ack_requested = true;
 }
 
-bool gprs_rlcmac_dl_tbf::need_control_ts() const
+/* Does this DL TBF require to poll the MS for DL ACK/NACK? */
+bool gprs_rlcmac_dl_tbf::need_poll_for_dl_ack_nack() const
 {
-	if (poll_scheduled())
-		return false;
-
-	return state_flags & (1 << GPRS_RLCMAC_FLAG_TO_DL_ACK) ||
+	/* poll after POLL_ACK_AFTER_FRAMES frames, or when final block is tx or
+	 * when last polled DL ACK/NACK was lost. */
+	return state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_TO_DL_ACK) ||
 		m_tx_counter >= POLL_ACK_AFTER_FRAMES ||
 		m_dl_ack_requested;
 }
@@ -1356,13 +1239,21 @@ bool gprs_rlcmac_dl_tbf::keep_open(unsigned fn) const
 {
 	int keep_time_frames;
 	unsigned long dl_tbf_idle_msec;
+	int since_last_drain;
+	bool keep;
 
 	dl_tbf_idle_msec = osmo_tdef_get(the_pcu->T_defs, -2031, OSMO_TDEF_MS, -1);
 	if (dl_tbf_idle_msec == 0)
 		return false;
 
 	keep_time_frames = msecs_to_frames(dl_tbf_idle_msec);
-	return frames_since_last_drain(fn) <= keep_time_frames;
+	since_last_drain = frames_since_last_drain(fn);
+	keep = since_last_drain <= keep_time_frames;
+
+	if (since_last_drain >= 0)
+		LOGPTBFDL(this, LOGL_DEBUG, "Keep idle TBF open: %d/%d -> %s\n",
+			  since_last_drain, keep_time_frames, keep ? "yes" : "no");
+	return keep;
 }
 
 /*
@@ -1495,55 +1386,55 @@ void gprs_rlcmac_dl_tbf::update_coding_scheme_counter_dl(enum CodingScheme cs)
 	switch (cs) {
 	case CS1:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS1);
-		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS1]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_gprs_ctrs, TBF_CTR_GPRS_DL_CS1));
 		break;
 	case CS2:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS2);
-		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS2]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_gprs_ctrs, TBF_CTR_GPRS_DL_CS2));
 		break;
 	case CS3:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS3);
-		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS3]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_gprs_ctrs, TBF_CTR_GPRS_DL_CS3));
 		break;
 	case CS4:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS4);
-		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS4]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_gprs_ctrs, TBF_CTR_GPRS_DL_CS4));
 		break;
 	case MCS1:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS1);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS1]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS1));
 		break;
 	case MCS2:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS2);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS2]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS2));
 		break;
 	case MCS3:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS3);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS3]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS3));
 		break;
 	case MCS4:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS4);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS4]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS4));
 		break;
 	case MCS5:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS5);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS5]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS5));
 		break;
 	case MCS6:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS6);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS6]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS6));
 		break;
 	case MCS7:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS7);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS7]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS7));
 		break;
 	case MCS8:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS8);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS8]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS8));
 		break;
 	case MCS9:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS9);
-		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS9]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_dl_egprs_ctrs, TBF_CTR_EGPRS_DL_MCS9));
 		break;
 	default:
 		LOGPTBFDL(this, LOGL_ERROR, "attempting to update rate counters for unsupported (M)CS %s\n",
@@ -1558,3 +1449,8 @@ struct gprs_rlcmac_dl_tbf *as_dl_tbf(struct gprs_rlcmac_tbf *tbf)
 	else
 		return NULL;
 }
+
+void tbf_dl_trigger_ass(struct gprs_rlcmac_dl_tbf *tbf, struct gprs_rlcmac_tbf *old_tbf)
+{
+	return tbf->trigger_ass(old_tbf);
+}
