@@ -38,30 +38,25 @@ enum tbf_dl_prio {
 	DL_PRIO_CONTROL,   /* a control block needs to be sent */
 };
 
-#define LOGPTBFDL(tbf, level, fmt, args...) LOGP(DTBFDL, level, "%s " fmt, tbf_name(tbf), ## args)
-
 struct gprs_rlcmac_dl_tbf : public gprs_rlcmac_tbf {
 	gprs_rlcmac_dl_tbf(struct gprs_rlcmac_bts *bts, GprsMs *ms);
+	~gprs_rlcmac_dl_tbf();
 	gprs_rlc_window *window();
-	void cleanup();
 
 	int append_data(uint16_t pdu_delay_csec,
 			const uint8_t *data, uint16_t len);
 
-	int rcvd_dl_ack(bool final, uint8_t ssn, uint8_t *rbb);
 	int rcvd_dl_ack(bool final_ack, unsigned first_bsn, struct bitvec *rbb);
 	struct msgb *create_dl_acked_block(uint32_t fn, uint8_t ts, enum mcs_kind req_mcs_kind = EGPRS);
 	void trigger_ass(struct gprs_rlcmac_tbf *old_tbf);
 
-	bool handle_ack_nack();
 	void request_dl_ack();
-	bool need_control_ts() const;
+	bool need_poll_for_dl_ack_nack() const;
 	bool have_data() const;
 	int frames_since_last_poll(unsigned fn) const;
 	int frames_since_last_drain(unsigned fn) const;
 	bool keep_open(unsigned fn) const;
 	int release();
-	int abort();
 	uint16_t window_size() const;
 	void set_window_size();
 	void update_coding_scheme_counter_dl(enum CodingScheme cs);
@@ -74,7 +69,6 @@ struct gprs_rlcmac_dl_tbf : public gprs_rlcmac_tbf {
 	 * variables are in both (dl and ul) structs and not outside union.
 	 */
 	int32_t m_tx_counter; /* count all transmitted blocks */
-	uint8_t m_wait_confirm; /* wait for CCCH IMM.ASS cnf */
 	bool m_dl_ack_requested;
 	int32_t m_last_dl_poll_fn;
 	int32_t m_last_dl_drained_fn;
@@ -108,9 +102,8 @@ protected:
 	int create_new_bsn(const uint32_t fn, enum CodingScheme cs);
 	struct msgb *create_dl_acked_block(const uint32_t fn, const uint8_t ts,
 					int index, int index2 = -1);
-	int update_window(const uint8_t ssn, const uint8_t *rbb);
 	int update_window(unsigned first_bsn, const struct bitvec *rbb);
-	int maybe_start_new_window();
+	int rcvd_dl_final_ack();
 	bool dl_window_stalled() const;
 	void reuse_tbf();
 	void start_llc_timer();
@@ -157,6 +150,9 @@ int dl_tbf_handle(struct gprs_rlcmac_bts *bts,
 		  const uint8_t egprs_ms_class, const uint16_t delay_csec,
 		  const uint8_t *data, const uint16_t len);
 
+void tbf_dl_trigger_ass(struct gprs_rlcmac_dl_tbf *tbf, struct gprs_rlcmac_tbf *old_tbf);
+
+#define LOGPTBFDL(tbf, level, fmt, args...) LOGP(DTBFDL, level, "%s " fmt, tbf_name(tbf), ## args)
 #ifdef __cplusplus
 }
 #endif
