@@ -20,6 +20,7 @@
  */
 
 #include <bts.h>
+#include <bts_pch_timer.h>
 #include <tbf.h>
 #include <tbf_ul.h>
 #include <rlc.h>
@@ -90,12 +91,23 @@ static const struct rate_ctr_group_desc tbf_ul_egprs_ctrg_desc = {
 	tbf_ul_egprs_ctr_description,
 };
 
+gprs_rlcmac_ul_tbf::~gprs_rlcmac_ul_tbf()
+{
+	osmo_fsm_inst_free(ul_ack_fsm.fi);
+	ul_ack_fsm.fi = NULL;
+
+	rate_ctr_group_free(m_ul_egprs_ctrs);
+	rate_ctr_group_free(m_ul_gprs_ctrs);
+	/* ~gprs_rlcmac_tbf() is called automatically upon return */
+}
+
 static int ul_tbf_dtor(struct gprs_rlcmac_ul_tbf *tbf)
 {
 	tbf->~gprs_rlcmac_ul_tbf();
 	return 0;
 }
 
+/* Generic function to alloc a UL TBF, later configured to be assigned either over CCCH or PACCH */
 struct gprs_rlcmac_ul_tbf *tbf_alloc_ul_tbf(struct gprs_rlcmac_bts *bts, GprsMs *ms, int8_t use_trx, bool single_slot)
 {
 	struct gprs_rlcmac_ul_tbf *tbf;
@@ -103,9 +115,8 @@ struct gprs_rlcmac_ul_tbf *tbf_alloc_ul_tbf(struct gprs_rlcmac_bts *bts, GprsMs
 
 	OSMO_ASSERT(ms != NULL);
 
-	LOGP(DTBF, LOGL_DEBUG, "********** UL-TBF starts here **********\n");
-	LOGP(DTBF, LOGL_INFO, "Allocating UL TBF: MS_CLASS=%d/%d\n",
-	     ms_ms_class(ms), ms_egprs_ms_class(ms));
+	LOGPMS(ms, DTBF, LOGL_DEBUG, "********** UL-TBF starts here **********\n");
+	LOGPMS(ms, DTBF, LOGL_INFO, "Allocating UL TBF\n");
 
 	tbf = talloc(tall_pcu_ctx, struct gprs_rlcmac_ul_tbf);
 	if (!tbf)
@@ -134,62 +145,76 @@ struct gprs_rlcmac_ul_tbf *tbf_alloc_ul_tbf(struct gprs_rlcmac_bts *bts, GprsMs
 		return NULL;
 	}
 
-	llist_add_tail(tbf_bts_list(tbf), &bts->ul_tbfs);
+	llist_add_tail(tbf_trx_list(tbf), &tbf->trx->ul_tbfs);
 	bts_do_rate_ctr_inc(tbf->bts, CTR_TBF_UL_ALLOCATED);
 
 	return tbf;
 }
 
-
-gprs_rlcmac_ul_tbf *tbf_alloc_ul(struct gprs_rlcmac_bts *bts, GprsMs *ms, int8_t use_trx,
-				 uint32_t tlli)
+/* Alloc a UL TBF to be assigned over PACCH. Called when an MS requests to
+ * create a new UL TBF during the end of life of a previous UL TBF (or an SBA).
+ * In summary, this TBF is allocated as a consequence of receiving a "Pkt
+ * Resource Req" or "Pkt Ctrl Ack" from the MS.
+ * See TS 44.060 9.3.2.4.2 "Non-extended uplink TBF mode".
+ */
+gprs_rlcmac_ul_tbf *tbf_alloc_ul_pacch(struct gprs_rlcmac_bts *bts, GprsMs *ms, int8_t use_trx)
 {
 	struct gprs_rlcmac_ul_tbf *tbf;
 
-/* FIXME: Copy and paste with tbf_new_dl_assignment */
-	/* create new TBF, use same TRX as DL TBF */
-	/* use multislot class of downlink TBF */
 	tbf = tbf_alloc_ul_tbf(bts, ms, use_trx, false);
 	if (!tbf) {
-		LOGP(DTBF, LOGL_NOTICE, "No PDCH resource\n");
-		/* FIXME: send reject */
+		LOGPMS(ms, DTBF, LOGL_NOTICE, "No PDCH resource\n");
+		/* Caller will most probably send a Imm Ass Reject after return */
 		return NULL;
 	}
+	/* Contention resolution is considered to be done since TLLI is known in MS: */
 	tbf->m_contention_resolution_done = 1;
-	TBF_SET_ASS_ON(tbf, GPRS_RLCMAC_FLAG_PACCH, false);
-	T_START(tbf, T3169, 3169, "allocation (UL-TBF)", true);
-	tbf->update_ms(tlli, GPRS_RLCMAC_UL_TBF);
+	osmo_fsm_inst_dispatch(tbf->state_fsm.fi, TBF_EV_ASSIGN_ADD_PACCH, NULL);
+
+	return tbf;
+}
+
+/* Alloc a UL TBF to be assigned over CCCH. Used by request of a "One phase
+ * packet access", where MS requested only 1 PDCH TS (TS 44.018 Table 9.1.8.1). */
+struct gprs_rlcmac_ul_tbf *tbf_alloc_ul_ccch(struct gprs_rlcmac_bts *bts, struct GprsMs *ms)
+{
+	struct gprs_rlcmac_ul_tbf *tbf;
+
+	tbf = tbf_alloc_ul_tbf(bts, ms, -1, true);
+	if (!tbf) {
+		LOGP(DTBF, LOGL_NOTICE, "No PDCH resource for Uplink TBF\n");
+		/* Caller will most probably send a Imm Ass Reject after return */
+		return NULL;
+	}
+	osmo_fsm_inst_dispatch(tbf->state_fsm.fi, TBF_EV_ASSIGN_ADD_CCCH, NULL);
+	tbf->contention_resolution_start();
 	OSMO_ASSERT(tbf->ms());
 
 	return tbf;
 }
 
+/* Create a temporary dummy TBF to Tx a ImmAssReject if allocating a new one during
+ * packet resource Request failed. This is similar as tbf_alloc_ul() but without
+ * calling tbf->setup() (in charge of TFI/USF allocation), and reusing resources
+ * from Packet Resource Request we received. See TS 44.060 sec 7.1.3.2.1  */
 struct gprs_rlcmac_ul_tbf *handle_tbf_reject(struct gprs_rlcmac_bts *bts,
-			GprsMs *ms, uint32_t tlli, uint8_t trx_no, uint8_t ts)
+			GprsMs *ms, uint8_t trx_no, uint8_t ts)
 {
 	struct gprs_rlcmac_ul_tbf *ul_tbf = NULL;
 	struct gprs_rlcmac_trx *trx = &bts->trx[trx_no];
-
-	if (!ms)
-		ms = bts_alloc_ms(bts, 0, 0);
-	ms_set_tlli(ms, tlli);
+	OSMO_ASSERT(ms);
 
 	ul_tbf = talloc(tall_pcu_ctx, struct gprs_rlcmac_ul_tbf);
 	if (!ul_tbf)
 		return ul_tbf;
-
 	talloc_set_destructor(ul_tbf, ul_tbf_dtor);
 	new (ul_tbf) gprs_rlcmac_ul_tbf(bts, ms);
 
-	llist_add(tbf_bts_list((struct gprs_rlcmac_tbf *)ul_tbf), &bts->ul_tbfs);
-	bts_do_rate_ctr_inc(ul_tbf->bts, CTR_TBF_UL_ALLOCATED);
-	TBF_SET_ASS_ON(ul_tbf, GPRS_RLCMAC_FLAG_PACCH, false);
-
-	ms_attach_tbf(ms, ul_tbf);
-	ul_tbf->update_ms(tlli, GPRS_RLCMAC_UL_TBF);
-	TBF_SET_ASS_STATE_UL(ul_tbf, GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ);
-	ul_tbf->control_ts = ts;
 	ul_tbf->trx = trx;
+	/* The only one TS is the common, control TS */
+	ul_tbf->first_ts = ts;
+	ul_tbf->first_common_ts = ts;
+	tbf_assign_control_ts(ul_tbf);
 	ul_tbf->m_ctrs = rate_ctr_group_alloc(ul_tbf, &tbf_ctrg_desc, next_tbf_ctr_group_id++);
 	ul_tbf->m_ul_egprs_ctrs = rate_ctr_group_alloc(ul_tbf,
 						       &tbf_ul_egprs_ctrg_desc,
@@ -202,6 +227,13 @@ struct gprs_rlcmac_ul_tbf *handle_tbf_reject(struct gprs_rlcmac_bts *bts,
 		talloc_free(ul_tbf);
 		return NULL;
 	}
+	tbf_update_state_fsm_name(ul_tbf);
+
+	ms_attach_tbf(ms, ul_tbf);
+	llist_add(tbf_trx_list((struct gprs_rlcmac_tbf *)ul_tbf), &trx->ul_tbfs);
+	bts_do_rate_ctr_inc(ul_tbf->bts, CTR_TBF_UL_ALLOCATED);
+	osmo_fsm_inst_dispatch(ul_tbf->state_fsm.fi, TBF_EV_ASSIGN_ADD_PACCH, NULL);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ass_fsm.fi, TBF_UL_ASS_EV_SCHED_ASS_REJ, NULL);
 
 	return ul_tbf;
 }
@@ -210,11 +242,15 @@ gprs_rlcmac_ul_tbf::gprs_rlcmac_ul_tbf(struct gprs_rlcmac_bts *bts_, GprsMs *ms)
 	gprs_rlcmac_tbf(bts_, ms, GPRS_RLCMAC_UL_TBF),
 	m_rx_counter(0),
 	m_contention_resolution_done(0),
-	m_final_ack_sent(0),
 	m_ul_gprs_ctrs(NULL),
 	m_ul_egprs_ctrs(NULL)
 {
-	memset(&m_usf, 0, sizeof(m_usf));
+	memset(&m_usf, USF_INVALID, sizeof(m_usf));
+
+	memset(&ul_ack_fsm, 0, sizeof(ul_ack_fsm));
+	ul_ack_fsm.tbf = this;
+	ul_ack_fsm.fi = osmo_fsm_inst_alloc(&tbf_ul_ack_fsm, this, &ul_ack_fsm, LOGL_INFO, NULL);
+
 }
 
 /*
@@ -267,71 +303,37 @@ int gprs_rlcmac_ul_tbf::assemble_forward_llc(const gprs_rlc_data *_data)
 	return 0;
 }
 
-bool gprs_rlcmac_ul_tbf::ctrl_ack_to_toggle()
+void gprs_rlcmac_ul_tbf::contention_resolution_start()
 {
-	if (check_n_clear(GPRS_RLCMAC_FLAG_TO_UL_ACK))
-		return true; /* GPRS_RLCMAC_FLAG_TO_UL_ACK was set, now cleared */
-
-	state_flags |= (1 << GPRS_RLCMAC_FLAG_TO_UL_ACK);
-	return false; /* GPRS_RLCMAC_FLAG_TO_UL_ACK was unset, now set */
+	/* 3GPP TS 44.018 sec 11.1.2 Timers on the network side: "This timer is
+	 * started when a temporary block flow is allocated with an IMMEDIATE
+	 * ASSIGNMENT or an IMMEDIATE PACKET ASSIGNMENT or an EC IMMEDIATE
+	 * ASSIGNMENT TYPE 1 message during a packet access procedure. It is
+	 * stopped when the mobile station has correctly seized the temporary
+	 * block flow."
+	 * In our code base, it means we want to do contention resolution
+	 * timeout only for one-phase packet access, since two-phase is handled
+	 * through SBA structs, which are freed by the PDCH UL Controller if the
+	 * single allocated block is lost. */
+	T_START(this, T3141, 3141, "Contention resolution (UL-TBF, CCCH)", true);
 }
-
-bool gprs_rlcmac_ul_tbf::handle_ctrl_ack()
+void gprs_rlcmac_ul_tbf::contention_resolution_success()
 {
-	/* check if this control ack belongs to packet uplink ack */
-	if (ul_ack_state_is(GPRS_RLCMAC_UL_ACK_WAIT_ACK)) {
-		TBF_SET_ACK_STATE(this, GPRS_RLCMAC_UL_ACK_NONE);
-		return true;
-	}
+	if (m_contention_resolution_done)
+		return;
 
-	return false;
-}
-
-struct msgb *gprs_rlcmac_ul_tbf::create_ul_ack(uint32_t fn, uint8_t ts)
-{
-	int final = (state_is(GPRS_RLCMAC_FINISHED));
-	struct msgb *msg;
-	int rc;
-	unsigned int rrbp = 0;
-	uint32_t new_poll_fn = 0;
-
-	if (final) {
-		if (poll_scheduled() && ul_ack_state_is(GPRS_RLCMAC_UL_ACK_WAIT_ACK)) {
-			LOGPTBFUL(this, LOGL_DEBUG,
-				  "Polling is already scheduled, so we must wait for the final uplink ack...\n");
-			return NULL;
-		}
-
-		rc = check_polling(fn, ts, &new_poll_fn, &rrbp);
-		if (rc < 0)
-			return NULL;
-	}
-
-	msg = msgb_alloc(23, "rlcmac_ul_ack");
-	if (!msg)
-		return NULL;
-	bitvec *ack_vec = bitvec_alloc(23, tall_pcu_ctx);
-	if (!ack_vec) {
-		msgb_free(msg);
-		return NULL;
-	}
-	bitvec_unhex(ack_vec, DUMMY_VEC);
-	Encoding::write_packet_uplink_ack(ack_vec, this, final, rrbp);
-	bitvec_pack(ack_vec, msgb_put(msg, 23));
-	bitvec_free(ack_vec);
+	/* 3GPP TS 44.060 sec 7a.2.1 Contention Resolution */
+	/* 3GPP TS 44.018 3.5.2.1.4 Packet access completion: The one phase
+	   packet access procedure is completed at a successful contention
+	   resolution. The mobile station has entered the packet transfer mode.
+	   Timer T3141 is stopped on the network side */
+	t_stop(T3141, "Contention resolution success (UL-TBF, CCCH)");
 
 	/* now we must set this flag, so we are allowed to assign downlink
 	 * TBF on PACCH. it is only allowed when TLLI is acknowledged. */
 	m_contention_resolution_done = 1;
 
-	if (final) {
-		set_polling(new_poll_fn, ts, GPRS_RLCMAC_POLL_UL_ACK);
-		/* waiting for final acknowledge */
-		m_final_ack_sent = 1;
-	} else
-		TBF_SET_ACK_STATE(this, GPRS_RLCMAC_UL_ACK_NONE);
-
-	return msg;
+	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF_CONTENTION_RESOLUTION_SUCCESS);
 }
 
 /*! \brief receive data from PDCH/L1 */
@@ -346,8 +348,6 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 
 	const uint16_t ws = m_window.ws();
 
-	this->state_flags |= (1 << GPRS_RLCMAC_FLAG_UL_DATA);
-
 	LOGPTBFUL(this, LOGL_DEBUG, "UL DATA TFI=%d received (V(Q)=%d .. "
 		"V(R)=%d)\n", rlc->tfi, this->m_window.v_q(),
 		this->m_window.v_r());
@@ -356,15 +356,11 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 	gprs_rlcmac_rssi(this, rssi);
 
 	/* store measurement values */
-	if (ms())
-		ms_update_l1_meas(ms(), meas);
+	ms_update_l1_meas(ms(), meas);
 
 	uint32_t new_tlli = GSM_RESERVED_TMSI;
 	unsigned int block_idx;
 
-	/* restart T3169 */
-	T_START(this, T3169, 3169, "acked (data)", true);
-
 	/* Increment RX-counter */
 	this->m_rx_counter++;
 	update_coding_scheme_counter_ul(rlc->cs);
@@ -373,7 +369,6 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 		int num_chunks;
 		uint8_t *rlc_data;
 		rdbi = &rlc->block_info[block_idx];
-		bool need_rlc_data = false;
 
 		LOGPTBFUL(this, LOGL_DEBUG,
 			  "Got %s RLC data block: CV=%d, BSN=%d, SPB=%d, PI=%d, E=%d, TI=%d, bitoffs=%d\n",
@@ -388,23 +383,12 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 			LOGPTBFUL(this, LOGL_DEBUG, "BSN %d out of window %d..%d (it's normal)\n",
 				  rdbi->bsn,
 				  m_window.v_q(), m_window.mod_sns(m_window.v_q() + ws - 1));
+			continue;
 		} else if (m_window.is_received(rdbi->bsn)) {
 			LOGPTBFUL(this, LOGL_DEBUG,
 				  "BSN %d already received\n", rdbi->bsn);
-		} else {
-			need_rlc_data = true;
-		}
-
-		if (!is_tlli_valid()) {
-			if (!rdbi->ti) {
-				LOGPTBFUL(this, LOGL_NOTICE, "Missing TLLI within UL DATA.\n");
-				continue;
-			}
-			need_rlc_data = true;
-		}
-
-		if (!need_rlc_data)
 			continue;
+		}
 
 		/* Store block and meta info to BSN buffer */
 
@@ -459,6 +443,7 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 					  "Decoded premier TLLI=0x%08x of UL DATA TFI=%d.\n",
 					  new_tlli, rlc->tfi);
 				update_ms(new_tlli, GPRS_RLCMAC_UL_TBF);
+				bts_pch_timer_stop(bts, ms());
 			} else if (new_tlli != GSM_RESERVED_TMSI && new_tlli != tlli()) {
 				LOGPTBFUL(this, LOGL_NOTICE,
 					  "Decoded TLLI=%08x mismatch on UL DATA TFI=%d. (Ignoring due to contention resolution)\n",
@@ -466,6 +451,10 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 				m_window.invalidate_bsn(rdbi->bsn);
 				continue;
 			}
+		} else if (!is_tlli_valid()) {
+			LOGPTBFUL(this, LOGL_NOTICE, "Missing TLLI within UL DATA.\n");
+			m_window.invalidate_bsn(rdbi->bsn);
+			continue;
 		}
 
 		m_window.receive_bsn(rdbi->bsn);
@@ -488,14 +477,15 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 	rdbi = &block->block_info;
 
 	/* Check if we already received all data TBF had to send: */
-	if (this->state_is(GPRS_RLCMAC_FLOW) /* still in flow state */
-	 && this->m_window.v_q() == this->m_window.v_r()) { /* if complete */
+	if (this->state_is(TBF_ST_FLOW) /* still in flow state */
+	 && this->m_window.v_q() == this->m_window.v_r() /* if complete */
+	 && block->len) { /* if there was ever a last block received */
 		LOGPTBFUL(this, LOGL_DEBUG,
 			  "No gaps in received block, last block: BSN=%d CV=%d\n",
 			  rdbi->bsn, rdbi->cv);
 		if (rdbi->cv == 0) {
 			LOGPTBFUL(this, LOGL_DEBUG, "Finished with UL TBF\n");
-			TBF_SET_STATE(this, GPRS_RLCMAC_FINISHED);
+			osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_LAST_UL_DATA_RECVD, NULL);
 			/* Reset N3103 counter. */
 			this->n_reset(N3103);
 		}
@@ -503,7 +493,7 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 
 	/* If TLLI is included or if we received half of the window, we send
 	 * an ack/nack */
-	maybe_schedule_uplink_acknack(rlc, rdbi->cv == 0);
+	maybe_schedule_uplink_acknack(rlc, block->len && rdbi->cv == 0);
 
 	return 0;
 }
@@ -527,10 +517,10 @@ void gprs_rlcmac_ul_tbf::maybe_schedule_uplink_acknack(
 	}
 	if (countdown_finished) {
 		require_ack = true;
-		if (state_is(GPRS_RLCMAC_FLOW))
+		if (state_is(TBF_ST_FLOW))
 			LOGPTBFUL(this, LOGL_DEBUG,
 				  "Scheduling Ack/Nack, because some data is missing and last block has CV==0.\n");
-		else if (state_is(GPRS_RLCMAC_FINISHED))
+		else if (state_is(TBF_ST_FINISHED))
 			LOGPTBFUL(this, LOGL_DEBUG,
 				  "Scheduling final Ack/Nack, because all data was received and last block has CV==0.\n");
 	}
@@ -544,14 +534,7 @@ void gprs_rlcmac_ul_tbf::maybe_schedule_uplink_acknack(
 	if (!require_ack)
 		return;
 
-	if (ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE)) {
-		/* trigger sending at next RTS */
-		TBF_SET_ACK_STATE(this, GPRS_RLCMAC_UL_ACK_SEND_ACK);
-	} else {
-		/* already triggered */
-		LOGPTBFUL(this, LOGL_DEBUG,
-			  "Sending Ack/Nack already scheduled, no need to re-schedule\n");
-	}
+	osmo_fsm_inst_dispatch(this->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
 }
 
 /* Send Uplink unit-data to SGSN. */
@@ -703,55 +686,55 @@ void gprs_rlcmac_ul_tbf::update_coding_scheme_counter_ul(enum CodingScheme cs)
 	switch (cs) {
 	case CS1:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_UL_CS1);
-		rate_ctr_inc(&m_ul_gprs_ctrs->ctr[TBF_CTR_GPRS_UL_CS1]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_gprs_ctrs, TBF_CTR_GPRS_UL_CS1));
 		break;
 	case CS2:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_UL_CS2);
-		rate_ctr_inc(&m_ul_gprs_ctrs->ctr[TBF_CTR_GPRS_UL_CS2]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_gprs_ctrs, TBF_CTR_GPRS_UL_CS2));
 		break;
 	case CS3:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_UL_CS3);
-		rate_ctr_inc(&m_ul_gprs_ctrs->ctr[TBF_CTR_GPRS_UL_CS3]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_gprs_ctrs, TBF_CTR_GPRS_UL_CS3));
 		break;
 	case CS4:
 		bts_do_rate_ctr_inc(bts, CTR_GPRS_UL_CS4);
-		rate_ctr_inc(&m_ul_gprs_ctrs->ctr[TBF_CTR_GPRS_UL_CS4]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_gprs_ctrs, TBF_CTR_GPRS_UL_CS4));
 		break;
 	case MCS1:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS1);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS1]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS1));
 		break;
 	case MCS2:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS2);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS2]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS2));
 		break;
 	case MCS3:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS3);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS3]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS3));
 		break;
 	case MCS4:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS4);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS4]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS4));
 		break;
 	case MCS5:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS5);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS5]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS5));
 		break;
 	case MCS6:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS6);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS6]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS6));
 		break;
 	case MCS7:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS7);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS7]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS7));
 		break;
 	case MCS8:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS8);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS8]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS8));
 		break;
 	case MCS9:
 		bts_do_rate_ctr_inc(bts, CTR_EGPRS_UL_MCS9);
-		rate_ctr_inc(&m_ul_egprs_ctrs->ctr[TBF_CTR_EGPRS_UL_MCS9]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(m_ul_egprs_ctrs, TBF_CTR_EGPRS_UL_MCS9));
 		break;
 	default:
 		LOGPTBFUL(this, LOGL_ERROR, "attempting to update rate counters for unsupported (M)CS %s\n",
@@ -773,6 +756,12 @@ gprs_rlc_window *gprs_rlcmac_ul_tbf::window()
 	return &m_window;
 }
 
+void gprs_rlcmac_ul_tbf::usf_timeout()
+{
+	if (n_inc(N3101))
+		osmo_fsm_inst_dispatch(this->state_fsm.fi, TBF_EV_MAX_N3101, NULL);
+}
+
 struct gprs_rlcmac_ul_tbf *as_ul_tbf(struct gprs_rlcmac_tbf *tbf)
 {
 	if (tbf && tbf->direction == GPRS_RLCMAC_UL_TBF)
@@ -780,3 +769,24 @@ struct gprs_rlcmac_ul_tbf *as_ul_tbf(struct gprs_rlcmac_tbf *tbf)
 	else
 		return NULL;
 }
+
+void tbf_usf_timeout(struct gprs_rlcmac_ul_tbf *tbf)
+{
+	tbf->usf_timeout();
+}
+
+
+bool ul_tbf_contention_resolution_done(const struct gprs_rlcmac_ul_tbf *tbf)
+{
+	return tbf->m_contention_resolution_done;
+}
+
+struct osmo_fsm_inst *tbf_ul_ack_fi(const struct gprs_rlcmac_ul_tbf *tbf)
+{
+	return tbf->ul_ack_fsm.fi;
+}
+
+void ul_tbf_contention_resolution_success(struct gprs_rlcmac_ul_tbf *tbf)
+{
+	return tbf->contention_resolution_success();
+}
