@@ -24,6 +24,7 @@
 #include "tbf_dl.h"
 #include "bts.h"
 #include "gprs_ms.h"
+#include "bts_pch_timer.h"
 
 #include <string.h>
 #include <stdio.h>
@@ -34,6 +35,7 @@ extern "C" {
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/fsm.h>
 }
 
 /* globals used by the code */
@@ -59,8 +61,8 @@ static void check_tfi_usage(struct gprs_rlcmac_bts *bts)
 
 	struct gprs_rlcmac_tbf *tfi_usage[8][8][2][32] = {{{{NULL}}}};
 	struct llist_head *tbf_lists[2] = {
-		&bts->ul_tbfs,
-		&bts->dl_tbfs
+		&bts->trx[0].ul_tbfs,
+		&bts->trx[0].dl_tbfs
 	};
 
 	struct llist_item *pos;
@@ -150,6 +152,8 @@ static void test_alloc_a(gprs_rlcmac_tbf_direction dir,
 
 	OSMO_ASSERT(i == count);
 
+	OSMO_ASSERT(bts_all_pdch_allocated(bts));
+
 	for (i = 0; i < count; ++i)
 		if (tbfs[i])
 			tbf_free(tbfs[i]);
@@ -423,9 +427,6 @@ static void test_alloc_b()
 	test_all_alloc_b();
 }
 
-typedef int (*algo_t)(struct gprs_rlcmac_bts *bts, struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf, bool single,
-		      int8_t use_trx);
-
 static char get_dir_char(uint8_t mask, uint8_t tx, uint8_t rx, uint8_t busy)
 {
 	int offs = busy ? 32 : 0;
@@ -480,6 +481,7 @@ static GprsMs *alloc_tbfs(struct gprs_rlcmac_bts *bts, struct GprsMs *old_ms, en
 			tbf_free(ms_ul_tbf(old_ms));
 		tbf = tbf_alloc_ul_tbf(bts, old_ms, trx_no, false);
 		if (tbf == NULL) {
+			OSMO_ASSERT(trx_no != -1 || bts_all_pdch_allocated(bts));
 			ms_unref(old_ms);
 			return NULL;
 		}
@@ -491,6 +493,7 @@ static GprsMs *alloc_tbfs(struct gprs_rlcmac_bts *bts, struct GprsMs *old_ms, en
 			tbf_free(ms_dl_tbf(old_ms));
 		tbf = tbf_alloc_dl_tbf(bts, old_ms, trx_no, false);
 		if (tbf == NULL) {
+			OSMO_ASSERT(trx_no != -1 || bts_all_pdch_allocated(bts));
 			ms_unref(old_ms);
 			return NULL;
 		}
@@ -640,7 +643,7 @@ static unsigned alloc_many_tbfs(struct gprs_rlcmac_bts *bts, unsigned min_class,
 	return counter;
 }
 
-static void test_successive_allocation(algo_t algo, unsigned min_class,
+static void test_successive_allocation(alloc_algorithm_func_t algo, unsigned min_class,
 	unsigned max_class, enum test_mode mode,
 	unsigned expect_num, const char *text)
 {
@@ -674,7 +677,7 @@ static void test_successive_allocation(algo_t algo, unsigned min_class,
 	talloc_free(bts);
 }
 
-static void test_many_connections(algo_t algo, unsigned expect_num,
+static void test_many_connections(alloc_algorithm_func_t algo, unsigned expect_num,
 	const char *text)
 {
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
@@ -738,17 +741,17 @@ static inline void test_a_b_dyn(enum test_mode mode, uint8_t exp_A, uint8_t exp_
 static void test_successive_allocations()
 {
 	test_successive_allocation(alloc_algorithm_a,       1,  1, TEST_MODE_UL_AND_DL, 35, "A");
-	test_successive_allocation(alloc_algorithm_b,      10, 10, TEST_MODE_UL_AND_DL, 32, "B");
-	test_successive_allocation(alloc_algorithm_b,      12, 12, TEST_MODE_UL_AND_DL, 32, "B");
+	test_successive_allocation(alloc_algorithm_b,      10, 10, TEST_MODE_UL_AND_DL, 15, "B");
+	test_successive_allocation(alloc_algorithm_b,      12, 12, TEST_MODE_UL_AND_DL, 15, "B");
 
-	test_successive_allocation(alloc_algorithm_b,       1,                12, TEST_MODE_UL_AND_DL, 32, "B");
-	test_successive_allocation(alloc_algorithm_b,       1, mslot_class_max(), TEST_MODE_UL_AND_DL, 32, "B");
-	test_successive_allocation(alloc_algorithm_dynamic, 1, mslot_class_max(), TEST_MODE_UL_AND_DL, 35, "dynamic");
+	test_successive_allocation(alloc_algorithm_b,       1,                12, TEST_MODE_UL_AND_DL, 23, "B");
+	test_successive_allocation(alloc_algorithm_b,       1, mslot_class_max(), TEST_MODE_UL_AND_DL, 17, "B");
+	test_successive_allocation(alloc_algorithm_dynamic, 1, mslot_class_max(), TEST_MODE_UL_AND_DL, 17, "dynamic");
 
-	test_a_b_dyn(TEST_MODE_DL_AND_UL,    35, 32,  32);
-	test_a_b_dyn(TEST_MODE_DL_AFTER_UL, 160, 32,  95);
-	test_a_b_dyn(TEST_MODE_UL_AFTER_DL,  35, 32,  35);
-	test_a_b_dyn(TEST_MODE_UL_ONLY,      35, 32,  35);
+	test_a_b_dyn(TEST_MODE_DL_AND_UL,    35, 15,  15);
+	test_a_b_dyn(TEST_MODE_DL_AFTER_UL, 160, 32,  101);
+	test_a_b_dyn(TEST_MODE_UL_AFTER_DL,  35, 15,  15);
+	test_a_b_dyn(TEST_MODE_UL_ONLY,      35, 15,  21);
 	test_a_b_dyn(TEST_MODE_DL_ONLY,     160, 32, 101);
 }
 
@@ -804,6 +807,30 @@ static void test_2_consecutive_dl_tbfs()
 	talloc_free(bts);
 }
 
+static void test_bts_pch_timer(void)
+{
+	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
+	struct osmo_mobile_identity mi_imsi1, mi_imsi2;
+	struct osmo_mobile_identity  mi_tmsi1;
+	mi_imsi1.type = mi_imsi2.type = GSM_MI_TYPE_IMSI;
+	mi_tmsi1.type = GSM_MI_TYPE_TMSI;
+	OSMO_STRLCPY_ARRAY(mi_imsi1.imsi, "1234");
+	OSMO_STRLCPY_ARRAY(mi_imsi2.imsi, "5678");
+	mi_tmsi1.tmsi = 987654321;
+
+	fprintf(stderr, "Testing bts_pch_timer dealloc on bts dealloc\n");
+	log_set_category_filter(osmo_stderr_target, DPCU, 1, LOGL_DEBUG);
+
+	fprintf(stderr, "Starting PCH timer for 2 IMSI\n");
+	bts_pch_timer_start(bts, &mi_imsi1, mi_imsi1.imsi);
+	bts_pch_timer_start(bts, &mi_imsi2, mi_imsi2.imsi);
+	fprintf(stderr, "Starting PCH timer for 1 TMSI\n");
+	bts_pch_timer_start(bts, &mi_tmsi1, "6666");
+
+	fprintf(stderr, "Deallocating BTS, expecting the PCH timer to be stopped and deallocated\n");
+	talloc_free(bts);
+}
+
 int main(int argc, char **argv)
 {
 	tall_pcu_ctx = talloc_named_const(NULL, 1, "moiji-mobile AllocTest context");
@@ -819,6 +846,7 @@ int main(int argc, char **argv)
 	log_set_category_filter(osmo_stderr_target, DTBF, 1, LOGL_INFO);
 	if (getenv("LOGL_DEBUG"))
 		log_set_log_level(osmo_stderr_target, LOGL_DEBUG);
+	osmo_fsm_log_addr(false);
 
 	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
 
@@ -829,6 +857,7 @@ int main(int argc, char **argv)
 	test_many_connections(alloc_algorithm_b, 32, "B");
 	test_many_connections(alloc_algorithm_dynamic, 160, "dynamic");
 	test_2_consecutive_dl_tbfs();
+	test_bts_pch_timer();
 
 	talloc_free(the_pcu);
 	return EXIT_SUCCESS;
