@@ -44,6 +44,7 @@ extern "C" {
 #include <osmocom/vty/vty.h>
 #include <osmocom/gprs/protocol/gsm_04_60.h>
 #include <osmocom/gsm/l1sap.h>
+#include <osmocom/core/fsm.h>
 }
 
 #include <errno.h>
@@ -57,6 +58,18 @@ bool spoof_mnc_3_digits = false;
 /* Measurements shared by all unit tests */
 static struct pcu_l1_meas meas;
 
+int gprs_gp_send_test_cb(void *ctx, struct msgb *msg)
+{
+	return 0;
+}
+
+static gprs_pcu *prepare_pcu(void)
+{
+	struct gprs_pcu *pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	bssgp_set_bssgp_callback(gprs_gp_send_test_cb, NULL);
+	return pcu;
+}
+
 static int bts_handle_rach(struct gprs_rlcmac_bts *bts, uint16_t ra, uint32_t Fn, int16_t qta)
 {
 	struct rach_ind_params rip = {
@@ -75,9 +88,9 @@ static int bts_handle_rach(struct gprs_rlcmac_bts *bts, uint16_t ra, uint32_t Fn
 static void check_tbf(gprs_rlcmac_tbf *tbf)
 {
 	OSMO_ASSERT(tbf);
-	if (tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE))
-		OSMO_ASSERT(tbf->timers_pending(T3191) || tbf->timers_pending(T3193));
-	if (tbf->state_is(GPRS_RLCMAC_RELEASING))
+	if (tbf->state_is(TBF_ST_WAIT_RELEASE))
+		OSMO_ASSERT(tbf->timers_pending(T3191) || osmo_timer_pending(&tbf->state_fsm.fi->timer));
+	if (tbf->state_is(TBF_ST_RELEASING))
 		OSMO_ASSERT(tbf->timers_pending(T_MAX));
 }
 
@@ -94,7 +107,7 @@ static void test_tbf_base()
 
 static void test_tbf_tlli_update()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	GprsMs *ms, *ms_new;
 
@@ -203,9 +216,9 @@ static gprs_rlcmac_dl_tbf *create_dl_tbf(struct gprs_rlcmac_bts *bts, uint8_t ms
 	check_tbf(dl_tbf);
 
 	/* "Establish" the DL TBF */
-	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_SEND_ASS);
-	TBF_SET_STATE(dl_tbf, GPRS_RLCMAC_FLOW);
-	dl_tbf->m_wait_confirm = 0;
+	osmo_fsm_inst_dispatch(dl_tbf->dl_ass_fsm.fi, TBF_DL_ASS_EV_SCHED_ASS, NULL);
+	osmo_fsm_inst_dispatch(dl_tbf->state_fsm.fi, TBF_EV_ASSIGN_ADD_CCCH, NULL);
+	osmo_fsm_inst_dispatch(dl_tbf->state_fsm.fi, TBF_EV_ASSIGN_ACK_PACCH, NULL);
 	check_tbf(dl_tbf);
 
 	*trx_no_ = trx_no;
@@ -213,11 +226,6 @@ static gprs_rlcmac_dl_tbf *create_dl_tbf(struct gprs_rlcmac_bts *bts, uint8_t ms
 	return dl_tbf;
 }
 
-static unsigned fn2bn(unsigned fn)
-{
-	return (fn % 52) / 4;
-}
-
 static unsigned fn_add_blocks(unsigned fn, unsigned blocks)
 {
 	unsigned bn = fn2bn(fn) + blocks;
@@ -250,9 +258,35 @@ enum test_tbf_final_ack_mode {
 	TEST_MODE_REVERSE_FREE
 };
 
+/* Receive an ACK */
+static void _rcv_ack(bool fin, gprs_rlcmac_dl_tbf *tbf, uint8_t *rbb)
+{
+	gprs_rlc_dl_window *w = static_cast<gprs_rlc_dl_window *>(tbf->window());
+	uint8_t bits_data[RLC_GPRS_WS/8];
+	bitvec bits;
+	Ack_Nack_Description_t ack_nack;
+	int bsn_begin, bsn_end;
+	uint8_t ssn = w->v_s();
+
+	bits.data = bits_data;
+	bits.data_len = sizeof(bits_data);
+	bits.cur_bit = 0;
+	ack_nack.FINAL_ACK_INDICATION = fin;
+	ack_nack.STARTING_SEQUENCE_NUMBER = ssn;
+	memcpy(ack_nack.RECEIVED_BLOCK_BITMAP, rbb, RLC_GPRS_WS/8);
+
+	Decoding::decode_gprs_acknack_bits(
+		&ack_nack, &bits,
+		&bsn_begin, &bsn_end, w);
+
+	tbf->rcvd_dl_ack(fin, bsn_begin, &bits);
+	if (!fin)
+		OSMO_ASSERT(w->window_empty());
+}
+
 static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 	unsigned i;
@@ -291,21 +325,20 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 	} while (block_nr < 3);
 
 	OSMO_ASSERT(dl_tbf->have_data());
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Queue a final ACK */
 	memset(rbb, 0, sizeof(rbb));
 	/* Receive a final ACK */
-	dl_tbf->rcvd_dl_ack(true, 1, rbb);
+	_rcv_ack(true, dl_tbf, rbb);
 
 	/* Clean up and ensure tbfs are in the correct state */
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_WAIT_RELEASE));
 	new_tbf = ms_dl_tbf(ms);
 	check_tbf(new_tbf);
 	OSMO_ASSERT(new_tbf != dl_tbf);
 	OSMO_ASSERT(new_tbf->tfi() == 1);
 	check_tbf(dl_tbf);
-	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
 	if (test_mode == TEST_MODE_REVERSE_FREE) {
 		ms_ref(ms);
 		tbf_free(new_tbf);
@@ -327,17 +360,9 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 	fprintf(stderr, "=== end %s ===\n", __func__);
 }
 
-/* Receive an ACK */
-#define RCV_ACK(fin, tbf, rbb) do { \
-		gprs_rlc_dl_window *w = static_cast<gprs_rlc_dl_window *>(tbf->window());	\
-		tbf->rcvd_dl_ack(fin, w->v_s(), rbb);	\
-		if (!fin)						\
-			OSMO_ASSERT(w->window_empty());	\
-	} while(0)
-
 static void test_tbf_delayed_release()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 	unsigned i;
@@ -362,43 +387,42 @@ static void test_tbf_delayed_release()
 	for (i = 0; i < sizeof(llc_data); i++)
 		llc_data[i] = i%256;
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Schedule two LLC frames */
 	dl_tbf->append_data(1000, llc_data, sizeof(llc_data));
 	dl_tbf->append_data(1000, llc_data, sizeof(llc_data));
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Drain the queue */
 	while (dl_tbf->have_data())
 		/* Request to send one RLC/MAC block */
 		request_dl_rlc_block(dl_tbf, &fn);
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* ACK all blocks */
 	memset(rbb, 0xff, sizeof(rbb));
 
-	RCV_ACK(false, dl_tbf, rbb); /* Receive an ACK */
+	_rcv_ack(false, dl_tbf, rbb); /* Receive an ACK */
 
 	/* Force sending of a single block containing an LLC dummy command */
 	request_dl_rlc_block(dl_tbf, &fn);
 
-	RCV_ACK(false, dl_tbf, rbb); /* Receive an ACK */
+	_rcv_ack(false, dl_tbf, rbb); /* Receive an ACK */
 
 	/* Timeout (make sure fn % 52 remains valid) */
 	dl_tbf_idle_msec = osmo_tdef_get(the_pcu->T_defs, -2031, OSMO_TDEF_MS, -1);
 	fn += 52 * ((msecs_to_frames(dl_tbf_idle_msec + 100) + 51)/ 52);
 	request_dl_rlc_block(dl_tbf, &fn);
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FINISHED));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FINISHED));
 
-	RCV_ACK(true, dl_tbf, rbb); /* Receive a final ACK */
+	_rcv_ack(true, dl_tbf, rbb); /* Receive a final ACK */
 
 	/* Clean up and ensure tbfs are in the correct state */
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
-	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_WAIT_RELEASE));
 	check_tbf(dl_tbf);
 	tbf_free(dl_tbf);
 	TALLOC_FREE(the_pcu);
@@ -407,7 +431,7 @@ static void test_tbf_delayed_release()
 
 static void test_tbf_imsi()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 	uint8_t ms_class = 45;
@@ -468,7 +492,7 @@ static void test_tbf_imsi()
 
 static void test_tbf_exhaustion()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	unsigned i;
 	uint8_t ts_no = 4;
@@ -511,7 +535,7 @@ static void test_tbf_exhaustion()
 
 static void test_tbf_dl_llc_loss()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 	uint8_t ms_class = 45;
@@ -565,6 +589,16 @@ static void test_tbf_dl_llc_loss()
 
 	OSMO_ASSERT(ms_dl_tbf(ms) != NULL);
 
+	/* Here PCU would answer with data_cnf and trigger
+	 * bts_rcv_imm_ass_cnf(), which would trigger TBF_EV_ASSIGN_PCUIF_CNF.
+	 * That in turn would set up timer X2002. Finally, X2002 timeout
+	 * moves it to FLOW state. We set X2002 timeout to 0 here to get
+	 * immediate trigger through osmo_select_main() */
+	OSMO_ASSERT(osmo_tdef_set(the_pcu->T_defs, -2002, 0, OSMO_TDEF_MS) == 0);
+	osmo_fsm_inst_dispatch(ms_dl_tbf(ms)->state_fsm.fi, TBF_EV_ASSIGN_PCUIF_CNF, NULL);
+	osmo_select_main(0);
+	OSMO_ASSERT(ms_dl_tbf(ms)->state_is(TBF_ST_FLOW));
+
 	/* Get first BSN */
 	struct msgb *msg;
 	int fn = 0;
@@ -629,6 +663,14 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_single_phase(struct gprs_rlcmac_bts
 	return ul_tbf;
 }
 
+static void send_ul_mac_block_buf(struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_pdch *pdch,
+				  unsigned fn, uint8_t *buf, int num_bytes)
+{
+	bts_set_current_block_frame_number(bts, fn);
+	pdch->rcv_block(buf, num_bytes, fn, &meas);
+	pdch_ulc_expire_fn(pdch->ulc, fn);
+}
+
 static void send_ul_mac_block(struct gprs_rlcmac_bts *bts, unsigned trx_no, unsigned ts_no,
 	RlcMacUplink_t *ulreq, unsigned fn)
 {
@@ -644,27 +686,45 @@ static void send_ul_mac_block(struct gprs_rlcmac_bts *bts, unsigned trx_no, unsi
 	OSMO_ASSERT(size_t(num_bytes) < sizeof(buf));
 	bitvec_free(rlc_block);
 
-	bts_set_current_block_frame_number(bts, fn, 0);
-
 	pdch = &bts->trx[trx_no].pdch[ts_no];
-	pdch->rcv_block(&buf[0], num_bytes, fn, &meas);
+	send_ul_mac_block_buf(bts, pdch, fn, &buf[0], num_bytes);
+}
+
+
+static uint32_t get_poll_fn(struct gprs_rlcmac_tbf *tbf, uint8_t poll_ts)
+{
+	struct gprs_rlcmac_pdch *pdch = &tbf->trx->pdch[poll_ts];
+	struct pdch_ulc *ulc = pdch->ulc;
+	struct rb_node *node;
+	struct pdch_ulc_node *item;
+
+	for (node = rb_first(&ulc->tree_root); node; node = rb_next(node)) {
+		item = container_of(node, struct pdch_ulc_node, node);
+		if (item->type == PDCH_ULC_NODE_TBF_POLL && item->tbf_poll.poll_tbf == tbf)
+			return item->fn;
+	}
+	OSMO_ASSERT(0);
 }
 
 static void send_control_ack(gprs_rlcmac_tbf *tbf)
 {
 	RlcMacUplink_t ulreq = {0};
 
-	OSMO_ASSERT(tbf->poll_fn != 0);
-	OSMO_ASSERT(tbf->is_control_ts(tbf->poll_ts));
-
 	ulreq.u.MESSAGE_TYPE = MT_PACKET_CONTROL_ACK;
 	Packet_Control_Acknowledgement_t *ctrl_ack =
 		&ulreq.u.Packet_Control_Acknowledgement;
 
 	ctrl_ack->PayloadType = GPRS_RLCMAC_CONTROL_BLOCK;
 	ctrl_ack->TLLI = tbf->tlli();
-	send_ul_mac_block(tbf->bts, tbf->trx->trx_no, tbf->poll_ts,
-		&ulreq, tbf->poll_fn);
+	send_ul_mac_block(tbf->bts, tbf->trx->trx_no, tbf->control_ts,
+		&ulreq, get_poll_fn(tbf, tbf->control_ts));
+}
+
+static void send_empty_block(gprs_rlcmac_tbf *tbf, unsigned ts_no, unsigned fn)
+{
+	struct gprs_rlcmac_pdch *pdch;
+	pdch = &tbf->bts->trx[tbf->trx->trx_no].pdch[ts_no];
+	send_ul_mac_block_buf(tbf->bts, pdch, fn, NULL, 0);
 }
 
 static gprs_rlcmac_ul_tbf *puan_urbb_len_issue(struct gprs_rlcmac_bts *bts,
@@ -771,10 +831,11 @@ static gprs_rlcmac_ul_tbf *puan_urbb_len_issue(struct gprs_rlcmac_bts *bts,
 
 	pdch->rcv_block(data_msg, 42, *fn, &meas);
 
-	struct msgb *msg1 = ul_tbf->create_ul_ack(*fn, ts_no);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
+	struct msgb *msg1 = tbf_ul_ack_create_rlcmac_msg(ul_tbf, *fn, ts_no);
 
-	static uint8_t exp1[] = { 0x40, 0x24, 0x01, 0x0b, 0x3e, 0x24, 0x46, 0x68, 0x90, 0x87, 0xb0, 0x06,
-				  0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b
+	static uint8_t exp1[] = { 0x40, 0x24, 0x01, 0x0b, 0x3e, 0x24, 0x46, 0x68, 0x9c, 0x70, 0x87, 0xb0,
+				  0x06, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b
 	};
 
 	if (!msgb_eq_data_print(msg1, exp1, GSM_MACBLOCK_LEN)) {
@@ -798,10 +859,11 @@ static gprs_rlcmac_ul_tbf *puan_urbb_len_issue(struct gprs_rlcmac_bts *bts,
 
 	pdch->rcv_block(data_msg, 42, *fn, &meas);
 
-	msg1 = ul_tbf->create_ul_ack(*fn, ts_no);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
+	msg1 = tbf_ul_ack_create_rlcmac_msg(ul_tbf, *fn, ts_no);
 
-	static uint8_t exp2[] = { 0x40, 0x24, 0x01, 0x0b, 0x3e, 0x24, 0x46, 0x68, 0x90, 0x88, 0xb0, 0x06, 0x8b,
-				  0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b
+	static uint8_t exp2[] = { 0x40, 0x24, 0x01, 0x0b, 0x3e, 0x24, 0x46, 0x68, 0x9c, 0x70, 0x88, 0xb0,
+				  0x06, 0x8b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b
 	};
 
 	if (!msgb_eq_data_print(msg1, exp2, GSM_MACBLOCK_LEN)) {
@@ -1352,7 +1414,8 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_no_length(struct
 		pdch = &bts->trx[trx_no].pdch[ts_no];
 		pdch->rcv_block(&data[0], sizeof(data), *fn, &meas);
 	}
-	ul_tbf->create_ul_ack(*fn, ts_no);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
+	tbf_ul_ack_create_rlcmac_msg(ul_tbf, *fn, ts_no);
 	memset(data, 0x2b, sizeof(data));
 	hdr3 = (struct gprs_rlc_ul_header_egprs_3 *)data;
 	hdr3->r = 0;
@@ -1380,7 +1443,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_no_length(struct
 	request_dl_rlc_block(ul_tbf, fn);
 
 	check_tbf(ul_tbf);
-	OSMO_ASSERT(ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE));
+	OSMO_ASSERT(tbf_ul_ack_fi(ul_tbf)->state == TBF_UL_ACK_ST_NONE);
 
 	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
@@ -1434,7 +1497,8 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_with_length(stru
 		pdch = &bts->trx[trx_no].pdch[ts_no];
 		pdch->rcv_block(&data[0], sizeof(data), *fn, &meas);
 	}
-	ul_tbf->create_ul_ack(*fn, ts_no);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
+	tbf_ul_ack_create_rlcmac_msg(ul_tbf, *fn, ts_no);
 	memset(data, 0x2b, sizeof(data));
 	hdr3 = (struct gprs_rlc_ul_header_egprs_3 *)data;
 	hdr3->r = 0;
@@ -1458,12 +1522,13 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_with_length(stru
 
 	pdch = &bts->trx[trx_no].pdch[ts_no];
 	pdch->rcv_block(&data[0], sizeof(data), *fn, &meas);
-	ul_tbf->create_ul_ack(*fn, ts_no);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
+	tbf_ul_ack_create_rlcmac_msg(ul_tbf, *fn, ts_no);
 
 	request_dl_rlc_block(ul_tbf, fn);
 
 	check_tbf(ul_tbf);
-	OSMO_ASSERT(ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE));
+	OSMO_ASSERT(tbf_ul_ack_fi(ul_tbf)->state == TBF_UL_ACK_ST_NONE);
 
 	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
@@ -1520,7 +1585,8 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_CRBB(struct gprs_rlcm
 		pdch = &bts->trx[trx_no].pdch[ts_no];
 		pdch->rcv_block(&data[0], sizeof(data), *fn, &meas);
 	}
-	ul_tbf->create_ul_ack(*fn, ts_no);
+	osmo_fsm_inst_dispatch(ul_tbf->ul_ack_fsm.fi, TBF_UL_ACK_EV_SCHED_ACK, NULL);
+	tbf_ul_ack_create_rlcmac_msg(ul_tbf, *fn, ts_no);
 	memset(data, 0x2b, sizeof(data));
 	hdr3 = (struct gprs_rlc_ul_header_egprs_3 *)data;
 	hdr3->r = 0;
@@ -1548,7 +1614,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_CRBB(struct gprs_rlcm
 	request_dl_rlc_block(ul_tbf, fn);
 
 	check_tbf(ul_tbf);
-	OSMO_ASSERT(ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE));
+	OSMO_ASSERT(tbf_ul_ack_fi(ul_tbf)->state == TBF_UL_ACK_ST_NONE);
 
 	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
@@ -1652,7 +1718,7 @@ static void send_dl_data(struct gprs_rlcmac_bts *bts, uint32_t tlli, const char
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_dl_tbf(ms) != NULL);
 
-	if (imsi[0] && strcmp(imsi, "000") != 0) {
+	if (imsi[0] != '\0') {
 		ms2 = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 		OSMO_ASSERT(ms == ms2);
 	}
@@ -1692,7 +1758,7 @@ static inline void print_ta_tlli(const gprs_rlcmac_ul_tbf *ul_tbf, bool print_ms
 
 static void test_tbf_single_phase()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = DUMMY_FN; /* 17,25,9 */
@@ -1716,7 +1782,7 @@ static void test_tbf_single_phase()
 
 static void test_tbf_egprs_two_phase_puan(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -1771,7 +1837,7 @@ static void test_tbf_egprs_two_phase_puan(void)
  */
 static void test_immediate_assign_rej_single_block()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint32_t fn = 2654218;
 	uint16_t qta = 31;
@@ -1793,7 +1859,7 @@ static void test_immediate_assign_rej_single_block()
 	 */
 	rc = bts_handle_rach(bts, 0x70, rach_fn, qta);
 
-	OSMO_ASSERT(rc == -EINVAL);
+	OSMO_ASSERT(rc == -EBUSY);
 
 	TALLOC_FREE(the_pcu);
 	fprintf(stderr, "=== end %s ===\n", __func__);
@@ -1804,7 +1870,7 @@ static void test_immediate_assign_rej_single_block()
  */
 static void test_immediate_assign_rej_multi_block()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint32_t fn = 2654218;
 	uint16_t qta = 31;
@@ -1845,7 +1911,7 @@ static void test_immediate_assign_rej()
 
 static void test_tbf_two_phase()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -1877,7 +1943,7 @@ static inline void print_ms(GprsMs *ms, bool old)
 
 static void test_tbf_ra_update_rach()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -1887,6 +1953,7 @@ static void test_tbf_ra_update_rach()
 	const char *imsi = "0011223344";
 	uint8_t ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
+	gprs_rlcmac_dl_tbf *dl_tbf;
 	GprsMs *ms, *ms1, *ms2;
 
 	fprintf(stderr, "=== start %s ===\n", __func__);
@@ -1913,6 +1980,12 @@ static void test_tbf_ra_update_rach()
 	transmit_dl_data(bts, tlli1, &fn);
 	OSMO_ASSERT(llc_queue_size(ms_llc_queue(ms1)) == 0);
 
+	dl_tbf = ms_dl_tbf(ms1);
+	OSMO_ASSERT(dl_tbf);
+	fn = get_poll_fn(dl_tbf, dl_tbf->control_ts);
+	send_empty_block(dl_tbf, dl_tbf->control_ts, fn);
+	fn = fn_add_blocks(fn, 1);
+
 	/* Now establish a new TBF for the RA UPDATE COMPLETE (new TLLI) */
 	ul_tbf = establish_ul_tbf_two_phase(bts, ts_no, tlli2, &fn, qta,
 		ms_class, 0);
@@ -1943,7 +2016,7 @@ static void test_tbf_ra_update_rach()
 
 static void test_tbf_dl_flow_and_rach_two_phase()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2004,7 +2077,7 @@ static void test_tbf_dl_flow_and_rach_two_phase()
 
 static void test_tbf_dl_flow_and_rach_single_phase()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2064,7 +2137,7 @@ static void test_tbf_dl_flow_and_rach_single_phase()
 
 static void test_tbf_dl_reuse()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2111,7 +2184,7 @@ static void test_tbf_dl_reuse()
 	transmit_dl_data(bts, tlli1, &fn);
 	OSMO_ASSERT(llc_queue_size(ms_llc_queue(ms1)) == 0);
 	OSMO_ASSERT(ms_dl_tbf(ms1));
-	OSMO_ASSERT(ms_dl_tbf(ms1)->state_is(GPRS_RLCMAC_FINISHED));
+	OSMO_ASSERT(ms_dl_tbf(ms1)->state_is(TBF_ST_FINISHED));
 
 	dl_tbf1 = ms_dl_tbf(ms1);
 
@@ -2126,6 +2199,9 @@ static void test_tbf_dl_reuse()
 		send_dl_data(bts, tlli1, imsi, (const uint8_t *)buf, rc);
 	}
 
+	/* Drop first DL_ACK poll queued */
+	send_empty_block(dl_tbf1, dl_tbf1->control_ts, get_poll_fn(dl_tbf1, dl_tbf1->control_ts));
+
 	/* Fake Final DL Ack/Nack */
 	ulreq.u.MESSAGE_TYPE = MT_PACKET_DOWNLINK_ACK_NACK;
 	Packet_Downlink_Ack_Nack_t *ack = &ulreq.u.Packet_Downlink_Ack_Nack;
@@ -2134,29 +2210,29 @@ static void test_tbf_dl_reuse()
 	ack->DOWNLINK_TFI = dl_tbf1->tfi();
 	ack->Ack_Nack_Description.FINAL_ACK_INDICATION = 1;
 
-	send_ul_mac_block(bts, 0, dl_tbf1->poll_ts, &ulreq, dl_tbf1->poll_fn);
+	send_ul_mac_block(bts, 0, dl_tbf1->control_ts, &ulreq, get_poll_fn(dl_tbf1, dl_tbf1->control_ts));
 
-	OSMO_ASSERT(dl_tbf1->state_is(GPRS_RLCMAC_WAIT_RELEASE));
+	OSMO_ASSERT(dl_tbf1->state_is(TBF_ST_WAIT_RELEASE));
 
 	request_dl_rlc_block(dl_tbf1, &fn);
 
 	ms2 = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms2 == ms1);
 	OSMO_ASSERT(ms_dl_tbf(ms2));
-	OSMO_ASSERT(ms_dl_tbf(ms2)->state_is(GPRS_RLCMAC_ASSIGN));
+	OSMO_ASSERT(ms_dl_tbf(ms2)->state_is(TBF_ST_ASSIGN));
 
 	dl_tbf2 = ms_dl_tbf(ms2);
 
 	OSMO_ASSERT(dl_tbf1 != dl_tbf2);
 
 	send_control_ack(dl_tbf1);
-	OSMO_ASSERT(dl_tbf2->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf2->state_is(TBF_ST_FLOW));
 
 	/* Transmit all data */
 	transmit_dl_data(bts, tlli1, &fn);
 	OSMO_ASSERT(llc_queue_size(ms_llc_queue(ms2)) == 0);
 	OSMO_ASSERT(ms_dl_tbf(ms2));
-	OSMO_ASSERT(ms_dl_tbf(ms2)->state_is(GPRS_RLCMAC_FINISHED));
+	OSMO_ASSERT(ms_dl_tbf(ms2)->state_is(TBF_ST_FINISHED));
 
 	TALLOC_FREE(the_pcu);
 	fprintf(stderr, "=== end %s ===\n", __func__);
@@ -2164,7 +2240,7 @@ static void test_tbf_dl_reuse()
 
 static void test_tbf_gprs_egprs()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 	uint8_t ms_class = 45;
@@ -2231,7 +2307,7 @@ static inline void ws_check(gprs_rlcmac_dl_tbf *dl_tbf, const char *test, uint8_
 
 static void test_tbf_ws()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	GprsMs *ms;
 	uint8_t ts_no = 4;
@@ -2253,7 +2329,7 @@ static void test_tbf_ws()
 	the_pcu->alloc_algorithm = alloc_algorithm_b;
 	bts->trx[0].pdch[2].enable();
 	bts->trx[0].pdch[3].enable();
-	bts->trx[0].pdch[4].enable();
+	/* bts->trx[0].pdch[4].enable(); Already enabled during setup_bts() */
 	bts->trx[0].pdch[5].enable();
 
 	gprs_bssgp_init(bts, 4234, 4234, 1, 1, false, 0, 0, 0);
@@ -2276,7 +2352,7 @@ static void test_tbf_ws()
 
 static void test_tbf_update_ws(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	GprsMs *ms;
 	uint8_t ts_no = 4;
@@ -2298,7 +2374,7 @@ static void test_tbf_update_ws(void)
 	the_pcu->alloc_algorithm = alloc_algorithm_b;
 	bts->trx[0].pdch[2].enable();
 	bts->trx[0].pdch[3].enable();
-	bts->trx[0].pdch[4].enable();
+	/* bts->trx[0].pdch[4].enable(); Already enabled during setup_bts()) */
 	bts->trx[0].pdch[5].enable();
 
 	gprs_bssgp_init(bts, 5234, 5234, 1, 1, false, 0, 0, 0);
@@ -2320,7 +2396,7 @@ static void test_tbf_update_ws(void)
 
 static void test_tbf_puan_urbb_len(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2459,7 +2535,7 @@ static gprs_rlcmac_ul_tbf *tbf_li_decoding(struct gprs_rlcmac_bts *bts,
 
 static void test_tbf_li_decoding(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2495,7 +2571,7 @@ static void test_tbf_li_decoding(void)
  */
 static void test_tbf_epdan_out_of_rx_window(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ms_class = 11;
 	uint8_t egprs_ms_class = 11;
@@ -2545,7 +2621,7 @@ static void test_tbf_epdan_out_of_rx_window(void)
 	prlcmvb->mark_unacked(1286);
 	prlcmvb->mark_unacked(1287);
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	block = bitvec_alloc(23, tall_pcu_ctx);
 
@@ -2586,7 +2662,7 @@ static void test_tbf_epdan_out_of_rx_window(void)
 
 static void test_tbf_egprs_two_phase_spb(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2617,7 +2693,7 @@ static void test_tbf_egprs_two_phase_spb(void)
 
 static void test_tbf_egprs_two_phase()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	int ts_no = 7;
 	uint32_t fn = 2654218;
@@ -2671,35 +2747,38 @@ static void establish_and_use_egprs_dl_tbf(struct gprs_rlcmac_bts *bts, int mcs)
 	for (i = 0; i < sizeof(llc_data); i++)
 		llc_data[i] = i%256;
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Schedule a small LLC frame */
 	dl_tbf->append_data(1000, test_data, 10);
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Drain the queue */
-	while (dl_tbf->have_data())
+	while (dl_tbf->have_data()) {
 		/* Request to send one RLC/MAC block */
 		request_dl_rlc_block(dl_tbf, &fn);
+	}
+	send_empty_block(dl_tbf, dl_tbf->control_ts, fn);
 
 	/* Schedule a large LLC frame */
 	dl_tbf->append_data(1000, test_data, sizeof(test_data));
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Drain the queue */
-	while (dl_tbf->have_data())
+	while (dl_tbf->have_data()) {
 		/* Request to send one RLC/MAC block */
 		request_dl_rlc_block(dl_tbf, &fn);
+	}
+	send_empty_block(dl_tbf, dl_tbf->control_ts, fn);
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
-	RCV_ACK(true, dl_tbf, rbb); /* Receive a final ACK */
+	_rcv_ack(true, dl_tbf, rbb); /* Receive a final ACK */
 
 	/* Clean up and ensure tbfs are in the correct state */
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
-	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_WAIT_RELEASE));
 	check_tbf(dl_tbf);
 	tbf_free(dl_tbf);
 }
@@ -2725,7 +2804,7 @@ static gprs_rlcmac_dl_tbf *tbf_init(struct gprs_rlcmac_bts *bts,
 	for (i = 0; i < sizeof(test_data); i++)
 		test_data[i] = i%256;
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	/* Schedule a LLC frame
 	 * passing only 100 bytes, since it is enough to construct
@@ -2734,7 +2813,7 @@ static gprs_rlcmac_dl_tbf *tbf_init(struct gprs_rlcmac_bts *bts,
 	 */
 	dl_tbf->append_data(1000, test_data, 100);
 
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
 	return dl_tbf;
 
@@ -2744,11 +2823,10 @@ static void tbf_cleanup(gprs_rlcmac_dl_tbf *dl_tbf)
 {
 	uint8_t rbb[64/8];
 
-	RCV_ACK(true, dl_tbf, rbb); /* Receive a final ACK */
+	_rcv_ack(true, dl_tbf, rbb); /* Receive a final ACK */
 
 	/* Clean up and ensure tbfs are in the correct state */
-	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
-	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_WAIT_RELEASE));
 	check_tbf(dl_tbf);
 	tbf_free(dl_tbf);
 
@@ -3034,13 +3112,14 @@ static void establish_and_use_egprs_dl_tbf_for_retx(struct gprs_rlcmac_bts *bts,
 		/* Send first RLC data block with demanded_mcs */
 		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 	}
-
+	/* Clean up pending items in UL controller: */
+	send_empty_block(dl_tbf, dl_tbf->control_ts, fn+50);
 	tbf_cleanup(dl_tbf);
 }
 
 static void test_tbf_egprs_retx_dl(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 
@@ -3068,7 +3147,7 @@ static void test_tbf_egprs_retx_dl(void)
 
 static void test_tbf_egprs_spb_dl(void)
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 
@@ -3098,7 +3177,7 @@ static void test_tbf_egprs_spb_dl(void)
 
 static void test_tbf_egprs_dl()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint8_t ts_no = 4;
 	int i;
@@ -3121,22 +3200,25 @@ static void test_tbf_egprs_dl()
 
 static void test_packet_access_rej_prr_no_other_tbfs()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint32_t fn = 2654218;
 	int ts_no = 7;
 	uint8_t trx_no = 0;
 	uint32_t tlli = 0xffeeddcc;
-	struct gprs_rlcmac_ul_tbf *ul_tbf = NULL;
+	struct gprs_rlcmac_ul_tbf *ul_tbf;
+	struct GprsMs *ms;
 
 	fprintf(stderr, "=== start %s ===\n", __func__);
 
 	setup_bts(bts, ts_no, 4);
+	OSMO_ASSERT(osmo_tdef_set(the_pcu->T_defs, -2000, 0, OSMO_TDEF_MS) == 0);
 
 	int rc = 0;
 
-	ul_tbf = handle_tbf_reject(bts, NULL, tlli,
-				trx_no, ts_no);
+	ms = bts_alloc_ms(bts, 0, 0);
+	ms_set_tlli(ms, tlli);
+	ul_tbf = handle_tbf_reject(bts, ms, trx_no, ts_no);
 
 	OSMO_ASSERT(ul_tbf != 0);
 
@@ -3147,8 +3229,7 @@ static void test_packet_access_rej_prr_no_other_tbfs()
 		trx_no, ts_no, fn, bn);
 
 	OSMO_ASSERT(rc == 0);
-
-	ul_tbf->handle_timeout();
+	osmo_select_main(0);
 
 	TALLOC_FREE(the_pcu);
 	fprintf(stderr, "=== end %s ===\n", __func__);
@@ -3156,7 +3237,7 @@ static void test_packet_access_rej_prr_no_other_tbfs()
 
 static void test_packet_access_rej_prr()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint32_t fn = 2654218;
 	uint16_t qta = 31;
@@ -3181,44 +3262,44 @@ static void test_packet_access_rej_prr()
 	/*
 	 * Trigger rach till resources(USF) exhaust
 	 */
-	rc = bts_handle_rach(bts, 0x78, rach_fn, qta);
-	rc = bts_handle_rach(bts, 0x79, rach_fn, qta);
-	rc = bts_handle_rach(bts, 0x7a, rach_fn, qta);
-	rc = bts_handle_rach(bts, 0x7b, rach_fn, qta);
-	rc = bts_handle_rach(bts, 0x7c, rach_fn, qta);
-	rc = bts_handle_rach(bts, 0x7d, rach_fn, qta);
-	rc = bts_handle_rach(bts, 0x7e, rach_fn, qta);
-
-	/* fake a resource request */
-	ulreq.u.MESSAGE_TYPE = MT_PACKET_RESOURCE_REQUEST;
-	presreq = &ulreq.u.Packet_Resource_Request;
-	presreq->PayloadType = GPRS_RLCMAC_CONTROL_BLOCK;
-	presreq->ID.UnionType = 1; /* != 0 */
-	presreq->ID.u.TLLI = tlli;
-	presreq->Exist_MS_Radio_Access_capability2 = 1;
-	pmsradiocap = &presreq->MS_Radio_Access_capability2;
-	pmsradiocap->Count_MS_RA_capability_value = 1;
-	pmsradiocap->MS_RA_capability_value[0].u.Content.
-		Exist_Multislot_capability = 1;
-	pmultislotcap = &pmsradiocap->MS_RA_capability_value[0].
-		u.Content.Multislot_capability;
-
-	pmultislotcap->Exist_GPRS_multislot_class = 1;
-	pmultislotcap->GPRS_multislot_class = ms_class;
-	if (egprs_ms_class) {
-		pmultislotcap->Exist_EGPRS_multislot_class = 1;
-		pmultislotcap->EGPRS_multislot_class = egprs_ms_class;
+	int i;
+	for (i = 0; i < 8; i++) {
+		rc = bts_handle_rach(bts, 0x70 + i, rach_fn, qta);
 	}
 
-	send_ul_mac_block(bts, trx_no, ts_no, &ulreq, sba_fn);
+	sba_fn = 52;
+	for (i = 0; i < 8; i++) {
+		/* fake a resource request */
+		ulreq.u.MESSAGE_TYPE = MT_PACKET_RESOURCE_REQUEST;
+		presreq = &ulreq.u.Packet_Resource_Request;
+		presreq->PayloadType = GPRS_RLCMAC_CONTROL_BLOCK;
+		presreq->ID.UnionType = 1; /* != 0 */
+		presreq->ID.u.TLLI = tlli + i;
+		presreq->Exist_MS_Radio_Access_capability2 = 1;
+		pmsradiocap = &presreq->MS_Radio_Access_capability2;
+		pmsradiocap->Count_MS_RA_capability_value = 1;
+		pmsradiocap->MS_RA_capability_value[0].u.Content.
+			Exist_Multislot_capability = 1;
+		pmultislotcap = &pmsradiocap->MS_RA_capability_value[0].
+			u.Content.Multislot_capability;
 
-	/* trigger packet access reject */
-	uint8_t bn = fn2bn(fn);
+		pmultislotcap->Exist_GPRS_multislot_class = 1;
+		pmultislotcap->GPRS_multislot_class = ms_class;
+		if (egprs_ms_class) {
+			pmultislotcap->Exist_EGPRS_multislot_class = 1;
+			pmultislotcap->EGPRS_multislot_class = egprs_ms_class;
+		}
 
-	rc = gprs_rlcmac_rcv_rts_block(bts,
-		trx_no, ts_no, fn, bn);
+		send_ul_mac_block(bts, trx_no, ts_no, &ulreq, sba_fn);
+		sba_fn = fn_next_block(sba_fn);
 
-	OSMO_ASSERT(rc == 0);
+		/* trigger packet access reject */
+		uint8_t bn = fn2bn(fn);
+
+		rc = gprs_rlcmac_rcv_rts_block(bts,
+			trx_no, ts_no, fn, bn);
+		OSMO_ASSERT(rc == 0);
+	}
 
 	TALLOC_FREE(the_pcu);
 	fprintf(stderr, "=== end %s ===\n", __func__);
@@ -3226,10 +3307,10 @@ static void test_packet_access_rej_prr()
 
 void test_packet_access_rej_epdan()
 {
-	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = prepare_pcu();
 	struct gprs_rlcmac_bts *bts = bts_alloc(the_pcu, 0);
 	uint32_t tlli = 0xffeeddcc;
-	static uint8_t exp[] = { 0x40, 0x84, 0x7f, 0xf7, 0x6e, 0xe6, 0x41, 0x4b,
+	static uint8_t exp[] = { 0x40, 0x84, 0x7f, 0xf7, 0x6e, 0xe6, 0x7e, 0xab,
 				 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b,
 				 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b, 0x2b
 	};
@@ -3240,7 +3321,8 @@ void test_packet_access_rej_epdan()
 
 	dl_tbf->update_ms(tlli, GPRS_RLCMAC_DL_TBF);
 
-	struct msgb *msg = dl_tbf->create_packet_access_reject();
+	osmo_fsm_inst_dispatch(dl_tbf->ul_ass_fsm.fi, TBF_UL_ASS_EV_SCHED_ASS_REJ, NULL);
+	struct msgb *msg = tbf_ul_ass_create_rlcmac_msg((const struct gprs_rlcmac_tbf*)dl_tbf, 0, 0);
 
 	fprintf(stderr, "packet reject: %s\n",
 			osmo_hexdump(msg->data, 23));
@@ -3270,6 +3352,7 @@ int main(int argc, char **argv)
 	log_parse_category_mask(osmo_stderr_target, "DRLCMAC,1:DRLCMACDATA,3:DRLCMACDL,3:DRLCMACUL,3:"
 				"DRLCMACSCHED,1:DRLCMACMEAS,3:DNS,3:DLBSSGP,3:DPCU,5:"
 				"DL1IF,6:DTBF,1:DTBFUL,1:DTBFDL,1:DLGLOBAL,2:");
+	osmo_fsm_log_addr(false);
 	vty_init(&pcu_vty_info);
 	pcu_vty_init();
 
