@@ -37,6 +37,7 @@ extern "C" {
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/core/bits.h>
+#include <osmocom/core/fsm.h>
 }
 
 #define OSMO_ASSERT_STR_EQ(a, b) \
@@ -204,8 +205,12 @@ static void test_rlc_dl_ul_basic()
 		int count;
 		const char *rbb;
 		char win_rbb[65];
-		uint8_t bin_rbb[8];
+		uint8_t bin_rbb[RLC_GPRS_WS/8];
+		bitvec bits;
 		win_rbb[64] = '\0';
+		bits.data = bin_rbb;
+		bits.data_len = sizeof(bin_rbb);
+		bits.cur_bit = 0;
 
 		ul_win.m_v_n.reset();
 
@@ -219,9 +224,11 @@ static void test_rlc_dl_ul_basic()
 		OSMO_ASSERT(ul_win.ssn() == 0);
 		ul_win.update_rbb(win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
-		Encoding::encode_rbb(win_rbb, bin_rbb);
+		bits.cur_bit = 0;
+		Encoding::encode_rbb(win_rbb, &bits);
 		printf("rbb: %s\n", osmo_hexdump(bin_rbb, sizeof(bin_rbb)));
-		Decoding::extract_rbb(bin_rbb, win_rbb);
+		Decoding::extract_rbb(&bits, win_rbb);
+		//printf("win_rbb: %s\n", win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
 
 		/* simulate to have received 0, 1 and 5 */
@@ -237,9 +244,10 @@ static void test_rlc_dl_ul_basic()
 		OSMO_ASSERT(ul_win.ssn() == 1);
 		ul_win.update_rbb(win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
-		Encoding::encode_rbb(win_rbb, bin_rbb);
+		bits.cur_bit = 0;
+		Encoding::encode_rbb(win_rbb, &bits);
 		printf("rbb: %s\n", osmo_hexdump(bin_rbb, sizeof(bin_rbb)));
-		Decoding::extract_rbb(bin_rbb, win_rbb);
+		Decoding::extract_rbb(&bits, win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
 
 		OSMO_ASSERT(ul_win.is_in_window(1));
@@ -254,9 +262,10 @@ static void test_rlc_dl_ul_basic()
 		OSMO_ASSERT(ul_win.ssn() == 2);
 		ul_win.update_rbb(win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
-		Encoding::encode_rbb(win_rbb, bin_rbb);
+		bits.cur_bit = 0;
+		Encoding::encode_rbb(win_rbb, &bits);
 		printf("rbb: %s\n", osmo_hexdump(bin_rbb, sizeof(bin_rbb)));
-		Decoding::extract_rbb(bin_rbb, win_rbb);
+		Decoding::extract_rbb(&bits, win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
 
 		OSMO_ASSERT(ul_win.is_in_window(5));
@@ -271,9 +280,10 @@ static void test_rlc_dl_ul_basic()
 		OSMO_ASSERT(ul_win.ssn() == 6);
 		ul_win.update_rbb(win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
-		Encoding::encode_rbb(win_rbb, bin_rbb);
+		bits.cur_bit = 0;
+		Encoding::encode_rbb(win_rbb, &bits);
 		printf("rbb: %s\n", osmo_hexdump(bin_rbb, sizeof(bin_rbb)));
-		Decoding::extract_rbb(bin_rbb, win_rbb);
+		Decoding::extract_rbb(&bits, win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
 
 		OSMO_ASSERT(ul_win.is_in_window(65));
@@ -290,9 +300,10 @@ static void test_rlc_dl_ul_basic()
 		OSMO_ASSERT(ul_win.ssn() == 66);
 		ul_win.update_rbb(win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
-		Encoding::encode_rbb(win_rbb, bin_rbb);
+		bits.cur_bit = 0;
+		Encoding::encode_rbb(win_rbb, &bits);
 		printf("rbb: %s\n", osmo_hexdump(bin_rbb, sizeof(bin_rbb)));
-		Decoding::extract_rbb(bin_rbb, win_rbb);
+		Decoding::extract_rbb(&bits, win_rbb);
 		OSMO_ASSERT_STR_EQ(win_rbb, rbb);
 
 		OSMO_ASSERT(ul_win.is_in_window(2));
@@ -687,7 +698,7 @@ static void test_egprs_ul_ack_nack()
 	win->reset_state();
 	win->set_ws(256);
 
-	Encoding::write_packet_uplink_ack(dest, tbf, false, 0);
+	write_packet_uplink_ack(dest, tbf, false, 0);
 	extract_egprs_ul_ack_nack(tbf, dest, &ssn, &crbb_test, &rbb, false);
 	check_egprs_bitmap(tbf, ssn, &crbb_test, rbb, &rbb_size);
 	free_egprs_ul_ack_nack(&rbb, &crbb_test);
@@ -698,7 +709,7 @@ static void test_egprs_ul_ack_nack()
 	win->set_ws(256);
 	win->receive_bsn(1);
 
-	Encoding::write_packet_uplink_ack(dest, tbf, false, 0);
+	write_packet_uplink_ack(dest, tbf, false, 0);
 	extract_egprs_ul_ack_nack(tbf, dest, &ssn, &crbb_test, &rbb, false);
 	check_egprs_bitmap(tbf, ssn, &crbb_test, rbb, &rbb_size);
 	free_egprs_ul_ack_nack(&rbb, &crbb_test);
@@ -709,7 +720,7 @@ static void test_egprs_ul_ack_nack()
 	win->set_ws(128);
 	win->receive_bsn(127);
 
-	Encoding::write_packet_uplink_ack(dest, tbf, false, 0);
+	write_packet_uplink_ack(dest, tbf, false, 0);
 	extract_egprs_ul_ack_nack(tbf, dest, &ssn, &crbb_test, &rbb, false);
 	check_egprs_bitmap(tbf, ssn, &crbb_test, rbb, &rbb_size);
 	free_egprs_ul_ack_nack(&rbb, &crbb_test);
@@ -720,7 +731,7 @@ static void test_egprs_ul_ack_nack()
 	for (uint16_t i=1; i<384/2; i++)
 		win->receive_bsn(i*2);
 
-	Encoding::write_packet_uplink_ack(dest, tbf, false, 0);
+	write_packet_uplink_ack(dest, tbf, false, 0);
 	extract_egprs_ul_ack_nack(tbf, dest, &ssn, &crbb_test, &rbb, false);
 	check_egprs_bitmap(tbf, ssn, &crbb_test, rbb, &rbb_size);
 	free_egprs_ul_ack_nack(&rbb, &crbb_test);
@@ -877,7 +888,7 @@ void test_immediate_assign_rej()
 	bitvec_unhex(immediate_assignment_rej, DUMMY_VEC);
 	plen = Encoding::write_immediate_assignment_reject(
 		immediate_assignment_rej, 112, 100,
-		GSM_L1_BURST_TYPE_ACCESS_1);
+		GSM_L1_BURST_TYPE_ACCESS_1, 20);
 
 	printf("assignment reject: %s\n",
 		osmo_hexdump(immediate_assignment_rej->data, 22));
@@ -892,7 +903,7 @@ void test_immediate_assign_rej()
 
 	plen = Encoding::write_immediate_assignment_reject(
 		immediate_assignment_rej, 112, 100,
-		GSM_L1_BURST_TYPE_ACCESS_0);
+		GSM_L1_BURST_TYPE_ACCESS_0, 20);
 
 	printf("assignment reject: %s\n",
 		osmo_hexdump(immediate_assignment_rej->data, 22));
@@ -932,6 +943,7 @@ int main(int argc, char **argv)
 
 	log_set_category_filter(osmo_stderr_target, DTBF, 1, LOGL_INFO);
 	log_set_category_filter(osmo_stderr_target, DTBFUL, 1, LOGL_INFO);
+	osmo_fsm_log_addr(false);
 
 	the_pcu = gprs_pcu_alloc(tall_pcu_ctx);
 
